/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/ScriptOCron/RCS/ScriptOCron.h,v 2.20 2018/02/15 20:59:43 agmsmith Exp $
 *
 * This file contains stuff that's common to both the command line version and
 * the graphical user interface version of AGMSScriptOCron.
 *
 * This is a BeOS / Haiku OS program for editing and running commands without
 * needing to know about using the command line.  There's a graphical user
 * interface for editing the commands, letting you specify each argument
 * separately.  Running of commands can be triggered by "cron" style time based
 * events, the BeOS scripting system (see the Hey command), as well as a GUI
 * start button.
 *
 * Note that this uses the AGMS coding style.  That means no tabs, indents are
 * two spaces, m_ is the prefix for member variables, g_ is the prefix for
 * global names, C style comments, constants are in all capital letters and
 * most other things are mixed case, it's word wrapped to fit in 79 characters
 * per line to make proofreading on paper easier, and functions are listed in
 * reverse dependency order so that forward declarations (function prototypes
 * with no code) aren't needed.
 *
 * The command line only (no graphical user interface) AGMSScriptOCron software
 * is licensed under the MIT license.  The commercial Fetchit! software (the
 * graphical user interface layer on top of AGMSSCriptOCron) is sold under the
 * TuneTracker Systems Sales Terms and Conditions (summary: licensed to use a
 * single copy on a single computer, no reselling, no source code).  The source
 * code for the commercial version is proprietary, owned by TuneTracker Systems
 * LLC under a royalty agreement with Alexander G. M. Smith.
 *
 * The contents of this source code file are MIT licensed.
 *
 * Copyright (c) 2016 by Alexander G. M. Smith.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef INCLUDE_GUI_CODE
  #define INCLUDE_GUI_CODE 0
#endif
  // Controls building with or without the graphical user interface.  The
  // commercial version has the GUI and the public version does not (you have
  // to use "hey" commands to achieve similar functionality).  The source code
  // included with the public version also has the GUI code removed).


#if INCLUDE_GUI_CODE
  #define PROGRAM_NAME_WORD "Fetchit"
    // Whenever we need to concatenate the program name into some text, this
    // string is used.  Thus it needs to be a macro.  And not have spaces or
    // weird characters since it gets appended to various file and other names.
  #define PROGRAM_NAME_FANCY "Fetchit!" B_UTF8_TRADEMARK
    // The program name as displayed to the user, with odd characters like the
    // trademark symbol.
#else // Building command line only version.
  #define PROGRAM_NAME_WORD "AGMSScriptOCron"
  #define PROGRAM_NAME_FANCY "AGMSScriptOCron"
#endif

#define SUITE_BASE "suite/vnd.agmsmith.scriptocron."
  // The various command suites all start with this phrase.  It's followed by a
  // name of a type of scriptable object, like "command" or "field" or
  // "application".  Needs to be the same for both commercial and public
  // versions of the program, so third party scripting works without changes.


/******************************************************************************
 * Global variables, and not-so-variable things too.  Grouped by functionality.
 */

extern const char *g_AncestorName; // Name of ancestor's name data in BMessage.
extern const char *g_DataName; // "data" stored in a BMessage.
extern const char *g_DataTypeInName; // Data type of input, used in a BMessage.
extern const char *g_DataTypeOutName; // Data type of output, used in BMessage.
extern const char *g_EmptyString; // Save some space by recycling nothing.
extern const char *g_ErrorName; // "error" code stored in a BMessage.
extern const char *g_KeyDateOffset; // Key for UserData about URL date offset.
extern const char *g_KeyDescription; // Command's user notes or description.
extern const char *g_KeyGroupBox; // Grouping Box Title key in Field UserData.
extern const char *g_KeyHelpText; // Help one liner for Fields and Commands.
extern const char *g_KeyOriginalModel; // Key for UserData about Model(s) used.
extern const char *g_KeyRenameGUIMode; // Renaming GUI UserData set by user.
extern const char *g_KeyRenameGUIInsert; // Renaming GUI UserData set by user.
extern const char *g_KeyRenameGUIRearrange; // Renaming GUI UserData set by user.
extern const char *g_KeyRenameGUISimple; // Renaming GUI UserData set by user.
extern const char *g_KeyRenameMonthConversion; // Renaming GUI UserData.
extern const char *g_KeyRenameSample; // Renaming GUI UserData set by user.
extern const char *g_MessageName; // "message" describes error in a BMessage.
extern const char *g_ModelName; // "model" scripting field for Model's name.
extern const char *g_NameIndex; // BMessage specifier for index numbers.
extern const char *g_NameName; // "name" BMessage field with new name.
extern const char *g_NameReboot; // "@Launch" for the trigger string.
extern const char *g_NameSOCLogFile; // "SOCLogFile" key in various places.
extern const char *g_NameSOCName; // "SOCName" key in various places.
extern const char *g_NameSOCTempDir; // "SOCTempDir" key in various places.
extern const char *g_NameSOCWorkDir; // "SOCWorkDir" key in various places.
extern const char *g_NameUnknown; // "Unknown" default name for missing things.
extern const char *g_NameUnnamed; // "Unnamed" default name for new things.
extern const char *g_ParentName; // Name of parent as BMessage returned value.
extern const char *g_PathName; // "pathname" element returned in a BMessage.
extern const char *g_ResultName; // "result" returned in a BMessage.
extern const char *g_SerialName; // "serial" scripting field for Model data.
extern const char *g_WaitingName; // "waiting" BMessage run status extra info.

extern bool g_FirstTimeAppRun;
  /* TRUE if this is the very first time this program has been run (detected by
  the lack of a settings file).  FALSE for later times.  Useful for doing very
  first time initialisation, like setting up auto-run on system boot. */

extern volatile bool g_SettingsNeedSaving;
  /* Set to TRUE by anything that changes something which may affect the
  settings file.  Once a minute the main application will check this flag and
  save the settings if needed, clearing the flag after it's done. */


/* These ones are only used in the GUI version of the program, but are common
so that they can be saved in the settings file. */

#if INCLUDE_GUI_CODE
extern volatile bool g_AdvancedMode;
  /* When TRUE the GUI is running in advanced mode, otherwise basic mode. */

extern volatile float g_FontScaleCommandNames;
  /* Scale factor for the system bold font when drawing the Command names. */

extern volatile float g_FontScaleLogText;
  /* Scale factor for the system plain font when drawing the log text. */

extern BRect g_InitialWindowFrame;
  /* A saved setting that specifies the initial window size and position. */

extern bool g_MetaMode;
  /* When TRUE the GUI shows some meta things that the ordinary user shouldn't
  fiddle with.  Not volatile since it's only set at startup. */

extern volatile bool g_ServerMode;
  /* When TRUE the program runs in server mode - error messages don't result in
  pop-up dialog boxes, but you can still see them in stderr.  Also the window
  is minimized, if it exists.  Don't need to lock the Application or go through
  scripting since this is simple enough to be read directly from multiple
  threads. */

extern volatile bool g_SimpleWindowTitle;
  /* Set to TRUE to show a simple window title, which is consistently the same,
  and thus is useful for automated window stack & tile operations.  Set to
  FALSE to use the fancier program name combined with the name of the current
  selection. */

#endif // INCLUDE_GUI_CODE


typedef enum TriggerTypeEnum {
  TRG_CRON = 0,
  TRG_NAMED_COMMAND,
  TRG_REBOOT,
  TRG_MANUAL,
  TRG_MAX} TriggerType;

typedef enum CommandStateEnum {
  CS_READY_TO_RUN = 0,
  CS_RUNNING,
  CS_WAIT_SUBSIDIARY, // Waiting for subsidiary child Commands to finish.
  CS_EDITING,
  CS_MAX} CommandState;

/* The various properties you can use in the scripting system for all the kinds
of objects we support.  Alphabetical order just because the list has grown so
long that it's hard to find them otherwise. */

typedef enum PropertyNumbersEnum
{
  PN_CHILDREN = 0,
  PN_COMMAND,
  PN_COMMAND_LINE,
  PN_DISABLE,
  PN_DUPLICATE,
  PN_EDIT,
  PN_FIELD,
  PN_LOG,
  PN_LOG_FILE_ENABLE,
  PN_MODEL,
  PN_NAME,
  PN_PATHS,
  PN_RUN,
  PN_SERIAL,
  PN_SUBSIDIARY,
  PN_TEMPLATE,
  PN_TRIGGER,
  PN_USER_DATA,
  PN_USER_TEMP_DIR,
  PN_VALUE,
  PN_MAX
} PropertyNumbers;

extern char * g_PropertyNames [PN_MAX];
  // Strings for the property names, useful for building scripting BMessages.
  // Or for parsing them.


/******************************************************************************
 * Scriptable objects like Commands or Fields are stored in a Map, with the key
 * being the case insenitive name of the object (actually the pointer obtained
 * from BHandler::Name) and the value being a pointer to the BHandler based
 * object.
 *
 * typedef map<const char *, BHandler *, LessThanCharStringStruct> ObjectMap;
 * Though to be more type safe, we'll define maps for specific object types
 * separately, as needed.  Maybe using the object itself rather than a pointer.
 */

struct LessThanCharStringStruct
{
  bool operator() ( /* Our less-than comparison function for sorting. */
    const char * const &ItemA,
    const char * const &ItemB) const
  {
    int Delta = strcasecmp (ItemA, ItemB);
    return Delta < 0;
  };
};


/******************************************************************************
 * This utility function copies a named field from the source BMessage to the
 * destination BMessage, optionally renaming it in the process.  If there are
 * multiple fields with that name (an array), they're all copied.  Note that
 * the two messages must be different, copying from a message to itself won't
 * work properly.
 */

extern status_t CopyBMessageField (
  BMessage *SourceMessagePntr, BMessage *DestMessagePntr,
  const char *FieldName, const char *NewName = NULL);


/******************************************************************************
 * Global utility function to display an error message and return.  The message
 * part describes the error, and if ErrorNumber is non-zero, gets the string
 * ", error code $X (standard description)." appended to it.  If the message
 * is NULL then it gets defaulted to "Something went wrong".  The title part
 * doesn't get displayed (no title bar in the dialog box, but you can see it in
 * the debugger as the window thread name), and defaults to "Error Message" if
 * you didn't specify one.  If running in command line mode, the error gets
 * printed to stderr rather than showing up in a dialog box.
 */

extern void DisplayErrorMessage (
  const char *MessageString = NULL,
  int ErrorNumber = 0,
  const char *TitleString = NULL);


/******************************************************************************
 * This utility function combines a date format string with the current time
 * and a date offset to generate text that includes the current time.  It's
 * usually used for file names that are date based.  Works by calling the
 * system "date" executable or maybe using strftime().
 *
 * DateFormat is a format string like "%Y%m%d%H%M.mp3".  See "date --help" or
 * the info page on the "date" command for more details.  Leave off the leading
 * plus sign, it will be added for you.  Odd characters will be escaped before
 * running it, for safety.
 *
 * The date offset is something like "1 day ago 2 hours ago", "next week" or
 * "tomorrow".  Use an empty string for no offset from the current time.  If it
 * is invalid, an empty string will be returned for the output.  See the info
 * page on the "date" command for details.
 */

extern void ExpandDateFormat (BString &Output,
  const char *DateFormat, const char *DateOffset = "");


/******************************************************************************
 * Print the usage info to the stream.  Optionally includes a list of all
 * commands and do word wrapping (zero to not wrap).
 */

extern void PrintUsage (ostream& OutputStream, int WrapWidth,
  bool ListCommands);


/******************************************************************************
 * A utility to run a CLI command from a string and save output to a string.
 */

extern int RunCLIString (BString &Output, const char *CommandLine);


/******************************************************************************
 * A utility function to backslash-escape dangerous characters in a string.
 */

extern void SafeEscape (BString &TheString);


#if INCLUDE_GUI_CODE
/******************************************************************************
 * Start up the GUI user interface.  Called by ScriptOCronApp::ReadyToRun.
 */

extern void SetSizesAndOpenWindow ();


/******************************************************************************
 * Add a one line text message to the status bar at the bottom of the window.
 * Does nothing if the window isn't open yet.  Multithread safe too.  Returns
 * true if it succesfully posted the message, false if message system isn't
 * running or a deadlock happened.
 */

extern bool AddStatusBarText (const char *StatusText);

#endif // INCLUDE_GUI_CODE



/******************************************************************************
 * $Log: ScriptOCron.h,v $
 * Revision 2.20  2018/02/15 20:59:43  agmsmith
 * Don't upgrade user data set by the user; add rename GUI choices to
 * a list of UserData that doesn't get upgraded.
 *
 * Revision 2.19  2018/01/24 20:26:18  agmsmith
 * Added a menu item to switch to a simple window title (just the program name)
 * for stack & tile convenience.  Corresponding entry added to the settings
 * file.
 *
 * Revision 2.18  2017/11/02 20:40:28  agmsmith
 * Add a user setting for changing the font size of the Command names and
 * another for the log text size.  Menu items cycle through the sizes.
 *
 * Revision 2.17  2017/07/31 20:21:40  agmsmith
 * Add the user specified temporary directory setting.
 *
 * Revision 2.16  2017/07/30 20:52:12  agmsmith
 * Added PN_PATHS scripting operation to get the paths to the work directory,
 * temp directory etc, for a particular Command.  Also rearrange existing code
 * to use the internal function that gets those names, so they're all
 * calculated in one spot without duplication.
 *
 * Add a GUI (two buttons) for Fetch Just New Files to reset the list of files
 * known to be already downloaded, to force redownloading or skip all of them
 * currently on the server.  Use the new PN_PATHS to find the file to
 * manipulate.
 *
 * Revision 2.15  2017/03/08 01:46:25  agmsmith
 * New function for adding status bar text, some string constants.
 *
 * Revision 2.14  2017/02/10 22:54:01  agmsmith
 * Made g_NameIndex an exported global variable.
 *
 * Revision 2.13  2017/01/29 01:40:59  agmsmith
 * Moved code for getting output from a piped command line to a separate
 * function.
 *
 * Revision 2.12  2016/11/02 19:08:05  agmsmith
 * Added HelpText for per-Field GUI status line text shown when mouse over
 * Field.
 *
 * Revision 2.11  2016/10/25 21:07:55  agmsmith
 * Added GroupBox UserData entry and key string.
 *
 * Revision 2.10  2016/09/17 22:51:38  agmsmith
 * Save a version number (date stamp) in "Original Model" user data in
 * Commands, so we can tell if a changed (could be older or newer) Model is
 * available.  Also update Commands with new Models upon loading.  Add a
 * UserData delete utility function, and allow for arrays of strings when
 * getting a UserData string.
 *
 * Revision 2.9  2016/09/16 21:20:49  agmsmith
 * Change "Original Model" to be generated when the Model is created rather
 * than added to a new Command when it is cloned from the Model.  Also add a
 * version number (date stamp) to the original model data, so we can tell if it
 * needs an update or already matches the current Model when loading a Command.
 *
 * Revision 2.8  2016/09/14 21:57:41  agmsmith
 * Moved code into functions SafeEscape and ExpandDateFormat so GUI version can
 * use them too.
 *
 * Revision 2.7  2016/09/09 19:58:45  agmsmith
 * Add AsDatedURL option to incorporate a date in a file name.
 *
 * Revision 2.6  2016/08/22 00:50:12  agmsmith
 * Add g_NameUnknown for "Unknown" default string, mostly for convenience.
 *
 * Revision 2.5  2016/07/27 20:28:18  agmsmith
 * Include key in returned UserData scripting BMessage so you can see which key
 * you got if accessing it via array indices (also lets you list the keys).
 * Clarify "Original Model" comments to have a space in the term.  Add the
 * g_KeyOriginalModel string to the exported variables.
 *
 * Revision 2.4  2016/04/20 20:44:05  agmsmith
 * Added more info to the Get Model scripting operation.  Now tracks a change
 * serial number for Models, lets you get it.  Also can now find out the data
 * types of the Model's input and output.
 *
 * Revision 2.3  2016/04/10 21:29:36  agmsmith
 * Add ancestor and parent to the Children scripting operation.
 *
 * Revision 2.2  2016/03/05 22:57:52  agmsmith
 * Add a scripting command to get the list of chilren of a Command.
 *
 * Revision 2.1  2016/02/24 15:51:41  agmsmith
 * Add the waiting field name to determine more detail about run status.
 *
 * Revision 1.15  2016/02/11 23:00:25  agmsmith
 * Added m_Subsidiary and scripting operations, alphabetical PN_ enum.
 *
 * Revision 1.14  2016/02/11 19:18:54  agmsmith
 * Work in progress, adding wait for children state, only enum done so far.
 *
 * Revision 1.13  2016/02/10 03:33:14  agmsmith
 * Moved disable functionality to a separate flag with its own scripting
 * commands.  Old @Disabled in the trigger string now removed when loading
 * settings.
 *
 * Revision 1.12  2015/12/19 04:41:06  agmsmith
 * Change @Reboot to be @Launch, and the button label to be Run at Startup.
 *
 * Revision 1.11  2015/12/16 16:18:29  agmsmith
 * Downloads renamed, now "FTP Download" is the default for new ones in Basic
 * mode.  Now saves and loads the initial window size, so your last window size
 * and position is remembered.  Goes back to default size if you make the
 * window too small or move it off screen.
 *
 * Revision 1.10  2015/12/13 20:13:56  agmsmith
 * Added Duplicate command and menu items, reworked new name generation too.
 *
 * Revision 1.9  2015/12/12 20:24:01  agmsmith
 * Added Meta mode and command line argument to trigger it.
 *
 * Revision 1.8  2015/12/09 22:46:05  agmsmith
 * Export the CopyBMessageField function.
 *
 * Revision 1.7  2015/12/04 22:25:10  agmsmith
 * Check for changes that affect settings, and save once a minute.
 *
 * Revision 1.6  2015/12/04 21:17:44  agmsmith
 * Added the ability to detect the first use of the program.
 *
 * Revision 1.5  2015/12/03 22:23:51  agmsmith
 * Log file GUI now working, enable/disable plus view log file button.
 *
 * Revision 1.4  2015/12/03 19:14:34  agmsmith
 * Logging to a file feature added, no GUI yet.
 *
 * Revision 1.3  2015/12/02 00:02:47  agmsmith
 * Wording.
 *
 * Revision 1.2  2015/12/01 00:38:48  agmsmith
 * Move the proprietary GUI code into a separate file, change program name for
 * GUI mode to Fetchit!TM.
 *
 * Revision 1.1  2015/11/30 13:27:09  agmsmith
 * Initial revision.  Common things shared by GUI and command line builds.
 */
