/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/ScriptOCron/RCS/URLConverter.cpp,v 1.7 2018/01/23 17:32:54 agmsmith Exp $
 *
 * This utility class converts URL strings into the separate elements of a URL,
 * and back.
 *
 * This URLConverter software is licensed under the MIT license.
 *
 * Copyright (c) 2015 by Alexander G. M. Smith.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/* Standard C Library. */

#include <stdio.h>
#include <ctype.h>

/* BeOS (Be Operating System) headers. */

#include <String.h> // For BString.

/* Our stuff. */

#include "URLConverter.h"



/******************************************************************************
 * Utility functions.
 */

/* Decode the string by converting %XX hex codes into characters.  Space can
also be decoded from "+". */

static void DecodeElement (BString &Source, BString &Dest)
{
  int32 Length = Source.Length ();
  char *DestPntr = Dest.LockBuffer (Length + 1);
  const char *SourcePntr = Source.String ();
  unsigned char Letter;
  while ((Letter = *SourcePntr++) != 0)
  {
    if (Letter == '+')
      *DestPntr++ = ' ';
    else if (Letter == '%' && ( /* Is it followed by two hex digits? */
    (SourcePntr[0] >= '0' && SourcePntr [0] <= '9') ||
    (SourcePntr[0] >= 'A' && SourcePntr [0] <= 'F') ||
    (SourcePntr[0] >= 'a' && SourcePntr [0] <= 'f')) && (
    (SourcePntr[1] >= '0' && SourcePntr [1] <= '9') ||
    (SourcePntr[1] >= 'A' && SourcePntr [1] <= 'F') ||
    (SourcePntr[1] >= 'a' && SourcePntr [1] <= 'f')))
    {
      int Hex = 0;

      Letter = *SourcePntr++;
      if (Letter >= '0' && Letter <= '9')
        Hex += Letter - '0';
      else
        Hex += toupper (Letter) - 'A' + 10;
      Hex *= 16;

      Letter = *SourcePntr++;
      if (Letter >= '0' && Letter <= '9')
        Hex += Letter - '0';
      else
        Hex += toupper (Letter) - 'A' + 10;

      *DestPntr++ = (char) Hex;
    }
    else // Doesn't need decoding.
      *DestPntr++ = Letter;
  }
  *DestPntr = 0;
  Dest.UnlockBuffer ();
}



/* Encode the string by converting illegible characters to hex encoded ones,
the format is "%XX" where the XX are the two hexadecimal digits for the
character.  Space can also be encoded as "+", but that doesn't work with wget
FTP mode, so use %20.  Encode : @ and % for URL parsing reasons.  Also encode
characters that would upset shell scripting of arguments in double quotes ($
and " and !  and # and `).  Usually encode a "/" too, you don't want a bare
slash in passwords and other non-directory parts. */

static void EncodeElement (BString &Source, BString &Dest,
  bool SlashesToo = true)
{
  int32 Length = Source.Length ();
  char *DestPntr = Dest.LockBuffer (Length * 3 + 1);
  const char *SourcePntr = Source.String ();
  unsigned char Letter;
  while ((Letter = *SourcePntr++) != 0)
  {
    if (Letter < 32 || Letter >= 127 || Letter == '@' || Letter == '%' ||
    Letter == ':' || Letter == '+' || Letter == ' ' || Letter == '$' ||
    Letter == '"' || Letter == '!' || Letter == '#' || Letter == '`' ||
    Letter == '&' || Letter == '\\' || /* SED confused by & and \ in s/a/b/ */
    (Letter == '/' && SlashesToo))
    {
      char Hex[4];
      sprintf (Hex, "%02X", (unsigned int) Letter);
      *DestPntr++ = '%';
      *DestPntr++ = Hex[0];
      *DestPntr++ = Hex[1];
    }
    else // Doesn't need encoding.
      *DestPntr++ = Letter;
  }
  *DestPntr = 0;
  Dest.UnlockBuffer ();
}



/******************************************************************************
 * The main URLConverter class.
 */

URLConverter::URLConverter ()
{
}


URLConverter::URLConverter (URLConverter &CopyFrom)
{
  for (int iElement = 0; iElement < UET_NUMBER_OF_ELEMENTS; iElement++)
    m_Elements[iElement] = CopyFrom.m_Elements[iElement];
}


void URLConverter::ClearAll ()
{
  for (int iElement = 0; iElement < UET_NUMBER_OF_ELEMENTS; iElement++)
    m_Elements[iElement] = "";
}


const char *URLConverter::GetValue (URLElementTypeEnum Element)
{
  if (Element < 0 || Element >= UET_NUMBER_OF_ELEMENTS)
    return NULL;

  return m_Elements[Element].String ();
}


void URLConverter::SetValue (URLElementTypeEnum Element, const char *NewString)
{
  if (Element < 0 || Element >= UET_NUMBER_OF_ELEMENTS)
    return;

  m_Elements[Element] = NewString;
}


void URLConverter::FromString (const char *Buffer)
{
  BString Element;
  ClearAll ();

  // Is there a protocol at the beginning?

  const char *ProtoEndPntr = strstr (Buffer, "://");
  const char *AfterProtoPntr;
  if (ProtoEndPntr != NULL)
  {
    Element.SetTo (Buffer, ProtoEndPntr - Buffer);
    DecodeElement (Element, m_Elements[UET_PROTOCOL]);
    AfterProtoPntr = ProtoEndPntr + 3 /* length of "://" */;
  }
  else
    AfterProtoPntr = Buffer;

  const char *PortEndPntr = strstr (AfterProtoPntr, "/");
  const char *PathStartPntr;
  if (PortEndPntr == NULL)
    PathStartPntr = Buffer + strlen (Buffer); // End of buffer, no path.
  else
    PathStartPntr = PortEndPntr + 1; // Skip over the first slash.

  const char *UserPassEndPntr = strchr (AfterProtoPntr, '@');
  if (UserPassEndPntr != NULL && UserPassEndPntr < PortEndPntr)
  {
    // Have a user and password section, ending at the "@".  Look for a
    // password ":" marker.

    const char *PassPntr = strchr (AfterProtoPntr, ':');
    if (PassPntr != NULL && PassPntr < UserPassEndPntr)
    {
      // Have a password too.

      Element.SetTo (PassPntr + 1, UserPassEndPntr - PassPntr - 1);
      DecodeElement (Element, m_Elements[UET_PASSWORD]);
    }
    else // No password, the whole thing is userid.
      PassPntr = UserPassEndPntr;

    Element.SetTo (AfterProtoPntr, PassPntr - AfterProtoPntr);
    DecodeElement (Element, m_Elements[UET_USERNAME]);

    UserPassEndPntr++; // Skip over the "@".
  }
  else // No user section.  Whole is address.
    UserPassEndPntr = AfterProtoPntr;

  // Address and port are between UserPassEndPntr and PortEndPntr.

  const char *AddressEndPntr = strchr (UserPassEndPntr, ':');
  if (AddressEndPntr != NULL && AddressEndPntr < PortEndPntr)
  {
    Element.SetTo (AddressEndPntr + 1, PortEndPntr - AddressEndPntr - 1);
    DecodeElement (Element, m_Elements[UET_PORT]);
  }
  else // No port, all address.
    AddressEndPntr = PortEndPntr;

  Element.SetTo (UserPassEndPntr, AddressEndPntr - UserPassEndPntr);
  DecodeElement (Element, m_Elements[UET_ADDRESS]);

  const char *PathEndPntr = strrchr (PathStartPntr, '/');
  if (PathEndPntr != NULL)
  {
    Element.SetTo (PathStartPntr, PathEndPntr - PathStartPntr);
    DecodeElement (Element, m_Elements[UET_PATH]);
    PathEndPntr++; // Skip over the last slash.
  }
  else
    PathEndPntr = PathStartPntr;

  Element.SetTo (PathEndPntr);
  DecodeElement (Element, m_Elements[UET_FILENAME]);

  // If not specified, guess the protocol from the server name.

  if (m_Elements[UET_PROTOCOL].Length () == 0)
  {
    if (m_Elements[UET_ADDRESS].IFindFirst ("www") == 0 /* right at start */)
      m_Elements[UET_PROTOCOL] = "http";
    else if (m_Elements[UET_ADDRESS].IFindFirst ("ftp") == 0)
      m_Elements[UET_PROTOCOL] = "ftp";
  }
}


void URLConverter::ToString (BString &OutputString)
{
  BString Encoded;

  // Convert to a URL by starting in the middle (the address) and appending or
  // prepending parts on the outside in a kind of nested fashion.
  // ftp://user:password@address:port/path/filename

  EncodeElement (m_Elements[UET_ADDRESS], Encoded);
  OutputString = Encoded;

  // Username and password.

  if (m_Elements[UET_USERNAME].Length () > 0)
  {
    OutputString.Prepend ("@");
    if (m_Elements[UET_PASSWORD].Length () > 0)
    {
      EncodeElement (m_Elements[UET_PASSWORD], Encoded);
      OutputString.Prepend (Encoded);
      OutputString.Prepend (":");
    }
    EncodeElement (m_Elements[UET_USERNAME], Encoded);
    OutputString.Prepend (Encoded);
  }

  // Protocol.

  if (m_Elements[UET_PROTOCOL].Length () > 0)
  {
    OutputString.Prepend ("://");
    EncodeElement (m_Elements[UET_PROTOCOL], Encoded);
    OutputString.Prepend (Encoded);
  }

  // Port

  if (m_Elements[UET_PORT].Length () > 0 && (
  (m_Elements[UET_PROTOCOL].ICompare ("HTTP") == 0 &&
    m_Elements[UET_PORT].ICompare ("80") != 0) ||
  (m_Elements[UET_PROTOCOL].ICompare ("HTTPS") == 0 &&
    m_Elements[UET_PORT].ICompare ("443") != 0) ||
  (m_Elements[UET_PROTOCOL].ICompare ("FTP") == 0 &&
    m_Elements[UET_PORT].ICompare ("21") != 0)))
  {
    OutputString.Append (":");
    EncodeElement (m_Elements[UET_PORT], Encoded);
    OutputString.Append (Encoded);
  }

  OutputString.Append ("/"); // The slash after the server info.

  if (m_Elements[UET_PATH].Length () > 0)
  {
    // Make sure the path has a trailing slash.

    EncodeElement (m_Elements[UET_PATH], Encoded, false /* Slashes allowed */);
    if (Encoded[Encoded.Length () - 1] != '/')
      Encoded.Append ("/");
    OutputString.Append (Encoded);
  }

  if (m_Elements[UET_FILENAME].Length () > 0)
  {
    EncodeElement (m_Elements[UET_FILENAME], Encoded, false /* Slashes ok */);
    OutputString.Append (Encoded);
  }
}



/******************************************************************************
 * Self test code.
 */

#if TEST_URL_CONVERTER
void TestURLConverter ()
{
  BString Buffer;
  static const char * TestStrings [] =
  {
    "",
    "/",
    "www",
    "agmsmith.ca",
    "www.agmsmith.ca",
    "ftp.agmsmith.ca",
    "http://agmsmith.ca",
    "http://agmsmith.ca/",
    "http://agmsmith.ca/Stuff.txt",
    "http://www.agmsmith.ca/Blort/Stuff.txt",
    "https://www.agmsmith.ca/Blort/Blah/Stuff.txt",
    "http://www.agmsmith.ca//Root/Blog/Stuff.txt",
    "https://www.agmsmith.ca/Blort/Blog/",
    "http://www.agmsmith.ca/Blort/Blog",
    "ftp://user:secret@www.agmsmith.ca:123/Blort/Stuff.txt",
    "ftp://user%40agmsmith.ca:secret@example.com:21/Stuff.txt",
    "file:///boot/home/Stuff.txt",
    "file:////boot/home/Stuff.txt",
    "///The end.///"
  };

  printf ("Testing $Header: /CommonBe/agmsmith/Programming/ScriptOCron/RCS/URLConverter.cpp,v 1.7 2018/01/23 17:32:54 agmsmith Exp $\n");
  URLConverter TestURL;

  TestURL.ToString (Buffer);
  printf ("Empty URLConverter: \"%s\"\n", Buffer.String ());

  BString Alphabet;
  for (int iChar = 1; iChar < 256; iChar++)
    Alphabet += (char) iChar;
  TestURL.SetValue (URLConverter::UET_ADDRESS, Alphabet.String ());
  TestURL.ToString (Buffer);
  printf ("Alphabet: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_ADDRESS, "www.agmsmith.ca");
  TestURL.ToString (Buffer);
  printf ("Just address: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PASSWORD, "secret");
  TestURL.ToString (Buffer);
  printf ("Added password, no userid: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_USERNAME, "user@example.com");
  TestURL.ToString (Buffer);
  printf ("Added userid with @: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_USERNAME, "userid");
  TestURL.ToString (Buffer);
  printf ("Added userid: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PASSWORD, "");
  TestURL.ToString (Buffer);
  printf ("Removed password: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_FILENAME, "file.blah");
  TestURL.ToString (Buffer);
  printf ("Added file name: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PATH, "path");
  TestURL.ToString (Buffer);
  printf ("Added path without slashes: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PATH, "path/");
  TestURL.ToString (Buffer);
  printf ("Added path with trailing slash: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PATH, "/path/");
  TestURL.ToString (Buffer);
  printf ("Added path with both slashes: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PROTOCOL, "fTp");
  TestURL.ToString (Buffer);
  printf ("Added protocol fTp: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PORT, "1234");
  TestURL.ToString (Buffer);
  printf ("Added port 1234: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PORT, "21");
  TestURL.ToString (Buffer);
  printf ("Added port 21 (the FTP one): \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PROTOCOL, "HTTP");
  TestURL.ToString (Buffer);
  printf ("Added protocol HTTP: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PORT, "80");
  TestURL.ToString (Buffer);
  printf ("Added port 80: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PROTOCOL, "https");
  TestURL.ToString (Buffer);
  printf ("Added protocol https: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_PORT, "443");
  TestURL.ToString (Buffer);
  printf ("Added port 443: \"%s\"\n", Buffer.String ());

  TestURL.SetValue (URLConverter::UET_FILENAME, "My File.txt");
  TestURL.ToString (Buffer);
  printf ("File with spaces: \"%s\"\n", Buffer.String ());

  for (int i = 0; i < (int) (sizeof (TestStrings) / sizeof (TestStrings[0]));
  i++)
  {
    TestURL.FromString (TestStrings[i]);
    TestURL.ToString (Buffer);
    printf ("\"%s\" converted to Prot:%s, User:%s, Pass:%s, Addr:%s, Port:%s, "
      "Path:%s, File:%s and then back to \"%s\".\n",
      TestStrings[i],
      TestURL.GetValue (URLConverter::UET_PROTOCOL),
      TestURL.GetValue (URLConverter::UET_USERNAME),
      TestURL.GetValue (URLConverter::UET_PASSWORD),
      TestURL.GetValue (URLConverter::UET_ADDRESS),
      TestURL.GetValue (URLConverter::UET_PORT),
      TestURL.GetValue (URLConverter::UET_PATH),
      TestURL.GetValue (URLConverter::UET_FILENAME),
      Buffer.String ());
  }
}
#endif


/******************************************************************************
 * $Log: URLConverter.cpp,v $
 * Revision 1.7  2018/01/23 17:32:54  agmsmith
 * Added copy constructor.
 *
 * Revision 1.6  2017/10/10 19:24:41  agmsmith
 * Also escape & and \ since they cause trouble when the file or path
 * name is in a SED expression, like s/blort/blah&blee/ where the
 * ampersand is replaced by the pattern matched.
 *
 * Revision 1.5  2016/09/13 21:24:59  agmsmith
 * Oops, wrong type of return value for GetValue when out of range.
 *
 * Revision 1.4  2016/09/09 18:15:58  agmsmith
 * Also escape the backslash character.
 *
 * Revision 1.3  2016/02/11 19:16:28  agmsmith
 * Needed to encode more special characters, for ones that bash shell uses and
 * ones that wget doesn't understand.
 *
 * Revision 1.2  2015/12/12 22:28:26  agmsmith
 * Encode slashes in passwords, user names, etc.
 *
 * Revision 1.1  2015/11/25 00:35:07  agmsmith
 * Initial revision
 */
