/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/ScriptOCron/RCS/ScriptOCron.cpp,v 2.91 2018/02/19 00:17:20 agmsmith Exp $
 *
 * This file implements the command line version of AGMSScriptOCron.
 *
 * This is a BeOS / Haiku OS program for editing and running commands without
 * needing to know about using the command line.  There's a graphical user
 * interface for editing the commands, letting you specify each argument
 * separately.  Running of commands can be triggered by "cron" style time based
 * events, the BeOS scripting system (see the Hey command), as well as a GUI
 * start button.
 *
 * Note that this uses the AGMS coding style.  That means no tabs, indents are
 * two spaces, m_ is the prefix for member variables, g_ is the prefix for
 * global names, C style comments, constants are in all capital letters and
 * most other things are mixed case, it's word wrapped to fit in 79 characters
 * per line to make proofreading on paper easier, and functions are listed in
 * reverse dependency order so that forward declarations (function prototypes
 * with no code) aren't needed.
 *
 * The command line only (no graphical user interface) AGMSScriptOCron software
 * is licensed under the MIT license.  The commercial Fetchit! software (the
 * graphical user interface layer on top of AGMSSCriptOCron) is sold under the
 * TuneTracker Systems Sales Terms and Conditions (summary: licensed to use a
 * single copy on a single computer, no reselling, no source code).  The source
 * code for the commercial version is proprietary, owned by TuneTracker Systems
 * LLC under a royalty agreement with Alexander G. M. Smith.
 *
 * The contents of this source code file are MIT licensed.
 *
 * Copyright (c) 2018 by Alexander G. M. Smith.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */


/* Standard C Library. */

#include <stdio.h>
#include <errno.h>
#include <signal.h> // For kill().
#include <unistd.h> // For fork, pipe, dup2, execve etc.

/* Standard C++ library. */

#include <iostream>
#include <sstream>

/* STL (Standard Template Library) headers. */

#include <map>
#include <vector>

/* BeOS (Be Operating System) headers. */

#include <Alert.h>
#include <Application.h>
#include <Archivable.h>
#include <ClassInfo.h> // For cast_as, probably obsolete.
#include <Directory.h>
#include <Entry.h>
#include <FindDirectory.h>
#include <Locker.h>
#include <Message.h>
#include <Path.h>
#include <PropertyInfo.h>
#include <Resources.h>
#include <String.h>

/* Our other stuff. */

#include "CronBitset.h"
#include "ScriptOCron.h"
#include "Trash.h"
#include "URLConverter.h"

/******************************************************************************
 * Global variables, and not-so-variable things too.  Grouped by functionality.
 */

static const char *g_AppSignature = "application/scriptocron";
  /* The application signature is the same for both the command line and
  commercial graphical user interface versions of the program, so that third
  party scripts will work with both.  It's also shorter than the usual
  x-vnd.blah so that it can be more easily used in "hey" commands. */

static const char *g_AddOnSubDirectory = "ScriptOCron";
  // Extra models and other data are stored in system & user AddOn directories
  // in a subdirectory with this name.  It's not the program name so that
  // AGMSScriptOCron and FetchIt! can share the extra data.

const char *g_AncestorName = "ancestor"; // Name of ancestor in a BMessage.
static const char *g_ChildListName = "ChildList"; // Used in archived Command.
static const char *g_CommandListName = "CommandList"; // Used in BArchive.
const char *g_DataName = "data"; // "data" field in a BMessage.
const char *g_DataTypeInName = "Datatype of Input"; // Key and BMessage field.
const char *g_DataTypeOutName = "Datatype of Output"; // Key and BMessage use.
static const char *g_DisabledName = "Disabled"; // For archiving.
const char *g_EmptyString = ""; // Save some space by recycling nothing.
const char *g_ErrorName = "error"; // Error code stored in a BMessage.
static const char *g_FieldListName = "FieldList"; // Used in archived Command.
static const char *g_LogDirName = "Logs"; // Log files go in this directory.
static const char *g_InternalName = "_name"; // Archived BHandler name field.
const char *g_KeyDateOffset = "DateOffset"; // UserData key AsDatedURL uses.
const char *g_KeyDescription = "Description"; // Command's user notes.
const char *g_KeyGroupBox = "GroupBox"; // Grouping Box Title key inside Field.
const char *g_KeyHelpText = "HelpText"; // Help key for Fields and Commands.
const char *g_KeyOriginalModel = "Original Model"; // UserData key.
const char *g_KeyRenameGUIMode = "RenameGUIMode"; // UserData set by user,
const char *g_KeyRenameGUIInsert = "RenameGUI.Insert"; // so don't upgrade it
const char *g_KeyRenameGUIRearrange = "RenameGUI.Rearrange"; // when doing
const char *g_KeyRenameGUISimple = "RenameGUI.Simple"; // general updating to
const char *g_KeyRenameMonthConversion = "RenameMonthConversion"; // a new
const char *g_KeyRenameSample = "RenameSampleName"; // model.
static const char *g_LogFileEnabledName = "LogFileEnabled"; // For archiving.
const char *g_MessageName = "message"; // Describes error in a BMessage.
static const char *g_ModelListName = "ModelList"; // Used for BArchived Models.
const char *g_ModelName = "model"; // Scripting field for Model's name.
static const char *g_NameBuiltIn = "builtin"; // Identifies built-in Models.
static const char *g_NameKey = "key"; // Name of UserData archived keys.
const char *g_NameIndex = "index"; // BMessage specifier for index #.
const char *g_NameName = "name"; // Names BMessage field with new name.
const char *g_NameReboot = "@Launch"; // For the trigger string.
const char *g_NameSOCLogFile = "SOCLogFile"; // Key for BMessages & Templates.
const char *g_NameSOCName = "SOCName"; // Key for BMessages & Templates.
const char *g_NameSOCTempDir = "SOCTempDir"; // Key for BMessages & Templates.
const char *g_NameSOCWorkDir = "SOCWorkDir"; // Key for BMessages & Templates.
const char *g_NameUnknown = "Unknown"; // Default name for missing things.
const char *g_NameUnnamed = "Unnamed"; // Default name for new things.
const char *g_ParentName = "parent"; // Name of parent Command in a BMessage.
const char *g_PathName = "pathname"; // Usually log path in a BMessage.
const char *g_ResultName = "result"; // Results returned in a BMessage.
const char *g_SerialName = "serial"; // Serial number returned in BMessage.
static const char *g_SettingsFileName = PROGRAM_NAME_WORD " Settings";
static const uint32 g_SettingsWhatCode = 'ASOC';
static const char *g_SubsidiaryName = "Subsidiary"; // For archived Command.
static const char *g_TemplateName = "Template"; // Used in archived Command.
static const char *g_TriggerName = "Trigger"; // Used in archived Command.
static const char *g_UserDataName = "UserData"; // Used in archiving it.
static const char *g_ValueName = "Value"; // Used in archived Fields.
const char *g_WaitingName = "waiting"; // BMessage run status extra info.

static const int MAX_LOG_LENGTH = 2000000;
#define MAX_LOG_LENGTH_AS_STRING "2000000"
  /* We stop adding output text to the in memory Command.m_LogAccumulated when
  it has gotten this big in bytes.  Given a maximum of 800 megabytes free
  memory for the application to use in 32 bit BeOS, don't make it too big
  (allow for up to 100 active downloads, some radio stations do have that many,
  plus there's a similar buffer in the GUI side). */

static const char * g_CommandStateNames [CS_MAX] =
{ // Readable versions of the states, mostly for error messages.
  "Ready to Run",
  "Running",
  "Waiting for Subsidiary Commands",
  "Editing"
};

static class ScriptOCronApp *g_App = NULL;
  /* A global reference to the main BApplication, non-null once the application
  has been initialised. */

bool g_FirstTimeAppRun = false;
  /* TRUE if this is the very first time this program has been run (detected by
  the lack of a settings file).  FALSE for later times.  Useful for doing very
  first time initialisation, like setting up auto-run on system boot. */

static BLocker *g_LockStdFilesPntr;
  /* Locks access to the standard output, input and error file handles.  Needed
  so you don't have two threads trying to set standard output to their own
  values.  Generally only lock when firing off a child process that needs
  different standard in/out/err.  Unlock when it's back to the normal standard
  in/out/err.  We don't lock every print statement, so theoretically some
  output may go to the wrong standard output.  But then the main program
  doesn't do much printing other than at startup and shutdown.  Might be worth
  explicitly using g_OriginalStdOut when printing. */

static int g_OriginalStdErr;
static int g_OriginalStdIn;
static int g_OriginalStdOut;
  /* Copies of the low level system file descriptors to the standard input,
  output and error files|terminal which were in effect when this program
  started.  They're also marked CLOEXEC (close on exec) so that they aren't
  inherited by child programs.  When needed, they can be duplicated and used to
  reset the standard input etc, back to the original files.  Needed so we can
  start child programs with different inherited stdin/err/out without using
  system(), popen() or fork(), which are slow in BeOS (copies the program's
  memory areas) and additionally seem to increase the odds of a runtime loader
  hang (some sort of race condition) in Haiku. */

static int g_QuitCountdown = -1;
  /* Set to the number of pulse timing events (about one every half second) to
  count down before the program quits.  Negative means stop counting.  Zero
  means quit at the next pulse event.  This is used to keep the program alive
  for a short while after someone requests that it quit, so that it has a
  chance to shut down running commands. */

volatile bool g_SettingsNeedSaving = false;
  /* Set to TRUE by anything that changes something which may affect the
  settings file.  Once a minute the main application will check this flag and
  save the settings if needed, clearing the flag after it's done. */


#if INCLUDE_GUI_CODE
volatile bool g_AdvancedMode = false;
  /* When TRUE the GUI is running in advanced mode, otherwise basic mode. */

volatile float g_FontScaleCommandNames = 1.25f;
  /* Scale factor for the system bold font when drawing the Command names. */

volatile float g_FontScaleLogText = 1.0f;
  /* Scale factor for the system be_fixed_font when drawing the log text. */

BRect g_InitialWindowFrame;
  /* A saved setting that specifies the initial window size and position. */

bool g_MetaMode = false;
  /* When TRUE the GUI shows some meta things that the ordinary user shouldn't
  fiddle with.  Not volatile since it's only set at startup. */

volatile bool g_ServerMode = false;
  /* When TRUE the program runs in server mode - error messages don't result in
  pop-up dialog boxes, but you can still see them in stderr.  Also the window
  is minimized, if it exists.  Don't need to lock the Application or go through
  scripting since this is simple enough to be read directly from multiple
  threads. */

volatile bool g_SimpleWindowTitle = false;
  /* Set to TRUE to show a simple window title, which is consistently the same,
  and thus is useful for automated window stack & tile operations.  Set to
  FALSE to use the fancier program name combined with the name of the current
  selection. */

#endif // INCLUDE_GUI_CODE


/* not const due to warnings */ char * g_PropertyNames [PN_MAX] =
{
  "Children", // PN_CHILDREN
  "Command", // PN_COMMAND
  "CommandLine", // PN_COMMAND_LINE
  "Disable", // PN_DISABLE
  "Duplicate", // PN_DUPLICATE
  "Edit", // PN_EDIT
  "Field", // PN_FIELD
  "Log", // PN_LOG
  "LogFileEnable", // PN_LOG_FILE_ENABLE
  "Model", // PN_MODEL
  "Name", // PN_NAME
  "Paths", // PN_PATHS
  "Run", // PN_RUN
  "Serial", // PN_SERIAL
  "Subsidiary", // PN_SUBSIDIARY,
  "Template", // PN_TEMPLATE
  "Trigger", // PN_TRIGGER
  "UserData", // PN_USER_DATA
  "UserTempDir", // PN_USER_TEMP_DIR
  "Value", // PN_VALUE
};


/* This array lists the scripting commands the top level Application can
handle, in a format that the scripting system can understand too. */

static struct property_info g_AppScriptingPropertyList [] =
{
  /* *name; commands[10]; specifiers[10]; *usage; extra_data; ... */

  {g_PropertyNames[PN_SERIAL],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns the change serial number for the Application overall.  It gets "
    "increased by some small amount (may wrap around to negative values "
    "eventually) when this object or any of its children have been changed.  "
    "Useful for knowing when to update a GUI displaying this object.",
    PN_SERIAL, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND],
    {B_COUNT_PROPERTIES, 0}, {B_DIRECT_SPECIFIER, 0},
    "Counts the number of Commands currently in the list of commands.",
    PN_COMMAND, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Create a new Command.  The Command's default name is \"Unnamed\".  You "
    "can set a name for the new Command by adding a string named \"name\" "
    "to the BMessage used for scripting.  If you want the new Command to be "
    "a copy of a Model Command, add a string named \"model\" with the name "
    "of the Model to use.  If no Model is specified, you get an empty Command "
    "(no Fields, no command line, etc).  The new Command starts in Edit mode, "
    "so it doesn't accidentally start running.  The Hey command would look "
    "like: hey " PROGRAM_NAME_WORD " create Command with "
    "name=NewCommandNameHere and model=ModelNameHere",
    PN_COMMAND, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND],
    {0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Specify a particular Command sub-object for further scripting.  "
    "It's better to refer to Commands by name than by index, since the array "
    "is kept in alphabetical order and adding or removing Commands will "
    "affect the indices of several other Commands.  The Hey command syntax "
    "would be: \"hey " PROGRAM_NAME_WORD " get something of Command "
    "NameOfCommandHere\"  For indexing, it would be: \"hey "
    PROGRAM_NAME_WORD " get something of Command [0]\" (note the space "
    "between Command and [0]).  If those don't work, try \"hey "
    PROGRAM_NAME_WORD " let Command [0] do get something\" which goes through "
    "a different code path to find the target object.",
    PN_COMMAND, {}, {}, {}},

  {g_PropertyNames[PN_MODEL],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets a list of the names of the Models, returned as an array of "
    "strings, in alphabetical order.  A Model is a saved Command that can be "
    "copied to make a new Command.  There are built-in ones for common tasks "
    "and ones which the user can create from existing Commands.  Also "
    "returns the data types of the input each model needs in an array of "
    "strings named \"Datatype of Input\", with an empty string for Models "
    "that don't have any inputs.  There's a corresponding array of "
    "\"Datatype of Output\" strings.  Those are used for Models that can be "
    "turned into Subsidiary Commands that can be chained together by "
    "matching data types of outputs from one subsidiary Command to the input "
    "of the next one.  This operation also returns the change serial number "
    "for the last time the model data was changed in a separate \"serial\" "
    "field.",
    PN_MODEL, {}, {}, {}},

  {g_PropertyNames[PN_MODEL],
    {B_COUNT_PROPERTIES, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the serial number for the last change of the Model data.  Lets you "
    "determine if your cached Models need to be updated.  If you want to know "
    "the number of Models, get the list of names and count them up.",
    PN_MODEL, {}, {}, {}},

  {g_PropertyNames[PN_MODEL],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Creates a Model from a Command.  An extra string called \"name\" "
    "specifies the name of the Command to be copied into the new Model (which "
    "will have the same name).  Fails if the Command doesn't exist or the "
    "name is already in use by another Model.",
    PN_MODEL, {}, {}, {}},

  {g_PropertyNames[PN_MODEL],
    {B_DELETE_PROPERTY, 0}, {B_NAME_SPECIFIER, 0},
    "Deletes the Model specified by the name specifier.  Try: \"hey "
    PROGRAM_NAME_WORD " delete Model NameOfModelHere\"",
    PN_MODEL, {}, {}, {}},

  {g_PropertyNames[PN_USER_TEMP_DIR],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns the user specified temporary files directory.  Commands will put "
    "their temporary files inside it, inside a " PROGRAM_NAME_WORD " folder "
    "which has subfolders for each Command's use.  If empty, the system "
    "default temporary directory will be used.  Use the \"Paths\" scripting "
    "property on a Command to get the actual path used.  Also returns the "
    "serial number at the time it was last changed, in key \"serial\".",
    PN_USER_TEMP_DIR, {}, {}, {}},

  {g_PropertyNames[PN_USER_TEMP_DIR],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Change the temporary files directory, where Commands temporarily "
    "put their downloaded files and other temporary things.  Usually this "
    "will be /tmp or something similar, varying depending on the operating "
    "system.  You can reset it back to the default one by specifying an empty "
    "string.  Trailing slashes aren't needed in your path, but it should "
    "start with a slash.  Usually you only need to change this if the default "
    "is on a disk without enough free space, or you don't like the way that "
    "the default /tmp directory gets automatically cleaned out when you "
    "reboot.",
    PN_USER_TEMP_DIR, {}, {}, {}},

  {g_PropertyNames[PN_USER_TEMP_DIR],
    {B_DELETE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Put the temporary files directory back to the system default one.",
    PN_USER_TEMP_DIR, {}, {}, {}},

  {0, {0}, {0}, 0, 0, {}, {}, {}} /* End of list of property commands. */
};


/* This array lists the scripting commands the Command object can handle.  Note
that the Name property is stored in the InternalName of the BHandler parent
class. */

static struct property_info g_CommandScriptingPropertyList [] =
{
  /* *name; commands[10]; specifiers[10]; *usage; extra_data; ... */
  {g_PropertyNames[PN_NAME],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the name of this Command.",
    PN_NAME, {}, {}, {}},

  {g_PropertyNames[PN_NAME],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Changes the name of this Command.  This will also affect the order of "
    "the Commands array (which is always in alphabetical order), so the index "
    "to this Command and others may change.  Fails if the new name is already "
    "in use or isn't suitable as a directory name (no slashes, not all "
    "periods, not empty).  You can change the name at any time, no need to "
    "switch into edit mode.  Will also rename the Work directory and Log file "
    "to match the new name, though if that fails you'll get an error but the "
    "Command rename still takes place.",
    PN_NAME, {}, {}, {}},

  {g_PropertyNames[PN_SERIAL],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns the change serial number for this Command, increased when this "
    "object or any of its children change in any way (including new log text, "
    "settings modifications, etc).",
    PN_SERIAL, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND],
    {B_DELETE_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Delete this Command.  Fails if the command is not in edit mode (running "
    "or idle).  Hey syntax is a bit odd, use \"hey " PROGRAM_NAME_WORD " let "
    "Command [3] do delete\".  A bit of a hack has been put in so you can use "
    "the more obvious \"delete Command [3]\" syntax, also \"hey "
    PROGRAM_NAME_WORD " delete Command of Command [3]\" does work.",
    PN_COMMAND, {}, {}, {}},

  {g_PropertyNames[PN_DUPLICATE],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Make a copy of this Command.  The new name of the command is \"Unnamed\" "
    "unless you specify it with a string called \"name\".  The new name must "
    "not exist, otherwise this operation will fail.  Can make a copy of a "
    "Command in any state; the new one will always start in the Edit state.",
    PN_DUPLICATE, {}, {}, {}},

  {g_PropertyNames[PN_EDIT],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Start editing this Command.  This creates a backup copy of the Command "
    "so that it can be restored to its original state if you cancel the "
    "editing.  Can only start editing if the Command isn't running or "
    "otherwise in use.  Changes the Command state to being edited, which will "
    "prevent other uses of the Command (like running it) while it is being "
    "edited.  Only works on non-subsidiary Commands; all subsidiary child "
    "Commands will also switch to edit mode when their ancestor Command "
    "changes.  Use \"hey " PROGRAM_NAME_WORD " let Command [3] do create "
    "edit\"",
    PN_EDIT, {}, {}, {}},

  {g_PropertyNames[PN_EDIT],
    {B_DELETE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Cancel the editing of this Command.  Restores it to its original state, "
    "unless you've somehow created another command with the old name.  In "
    "that rare case, you'll get an error and the Command and its backup copy "
    "are deleted rather than being restored.",
    PN_EDIT, {}, {}, {}},

  {g_PropertyNames[PN_EDIT],
    {B_EXECUTE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Finish editing the Command.  Gets rid of the backup copy of the original "
    "state of the Command, and changes the Command state to being ready to "
    "run.  The Hey command awkwardly uses \"do\" to mean execute, so it would "
    "have a double do like: hey " PROGRAM_NAME_WORD " let Command [0] do do "
    "Edit",
    PN_EDIT, {}, {}, {}},

  {g_PropertyNames[PN_EDIT],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns true if the Command is being edited, false otherwise.",
    PN_EDIT, {}, {}, {}},

  {g_PropertyNames[PN_RUN],
    {B_EXECUTE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Start running the Command.  Fails if it is being edited or is already "
    "running.  The Hey command awkwardly uses \"do\" to mean execute, so it "
    "would have a double do like: hey " PROGRAM_NAME_WORD " let Command [0] "
    "do do Run",
    PN_RUN, {}, {}, {}},

  {g_PropertyNames[PN_RUN],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns true if the Command is running or waiting for child Commands to "
    "finish running, false otherwise.  You can tell the difference by looking "
    "for the extra boolean \"waiting\", which will be true if the Command is "
    "waiting for child subsidiary Commands to finish.",
    PN_RUN, {}, {}, {}},

  {g_PropertyNames[PN_TEMPLATE],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Get the Template for building the command line used by this Command.",
    PN_TEMPLATE, {}, {}, {}},

  {g_PropertyNames[PN_TEMPLATE],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Set the Template for building the command line used by this Command.  "
    "Various magic words in the Template will be replaced with specific "
    "values.\n"
    "The most basic magic word is a Field name, which will get replaced with "
    "the Field's Value.\n"
    "Another is a Field name with \"AsDir\" appended, which uses the Value up "
    "to and including the last slash (/) character, nothing if no slash.\n"
    "Similarly, \"AsName\" is all the text after the last slash.\n"
    "\"AsUserData(name)\" gets replaced by the UserData under the named key "
    "attached to the Field, empty string if no UserData by that name or if it "
    "isn't a string.\n"
    "\"AsDatedString\" triggers date expansion if the Field Value starts "
    "with a plus sign, like AsDatedURL but without the conversion to and from "
    "a URL.\n"
    "\"AsDatedURL\" decodes the Field's value as a URL, and if the name "
    "portion starts with a plus sign, passes the name to the system \"date\" "
    "command as a format argument, with an optional UserData at key "
    "\"DateOffset\" string being used as a date offset, resulting in a CLI "
    "command that looks like 'date \"--date=1 day ago 2 minutes ago\" "
    "\"+Sermon%Y-%m%d.mp3\" | tr -d \"\\n\"'.  Date expansion is also done "
    "to the directory path if it starts with a plus sign.  Then the date "
    "expanded text is converted back to URL encoded format.\n"
    "\"AsNCFTP\" treats the Field's value as a URL and convertes it into "
    "arguments (separate parts for host name, directory, etc) for the "
    "ncftpput command.  Arguments are escaped and put in quotes for you.\n"
    "You can also append \"AsSafe\" to convert problematical characters "
    "(currently !$\"\\`) to their escaped equivalents.\n"
    "You can specify multiple postfixes in any order, but they will get "
    "processed in the AsUserData, AsDir, AsName, AsDatedString, AsDatedURL, "
    "AsNCFTP, AsSafe order.\n"
    "\"SOCName\" is replaced by the name of the ancestor Command.\n"
    "\"SOCMyName\" is replaced by the name of the current Command, even if "
    "it is a subsidiary one (SOCName is the one users should normally see).\n"
    "\"SOCWorkDir\" is replaced by the absolute path to the working "
    "directory for this command, ending without a slash, typically "
    "/boot/home/config/settings/" PROGRAM_NAME_WORD "/Work/CommandName, with "
    "the Command name sanitised (slashes changed to dashes and a few other "
    "things).  Subsidiary Commands use their ancestor's name so they all work "
    "from the same directory.  The work directory is used for things like a "
    "file containing the list of URLs to download or a file listing the "
    "server's contents as of the last run.  It is persistent between boots. "
    "You need to create the directory if needed; we just provide the name.\n"
    "\"SOCLogFile\" is replaced by the absolute path to the log file for this "
    "Command and its subsidiary Commands.\n"
    "\"SOCTempDir\" is for /tmp/" PROGRAM_NAME_WORD "/CommandName and is "
    "used for temporary files which get erased at the next system bootup.",
    PN_TEMPLATE, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND_LINE],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Get the current command line for this Command.  Will substitute "
    "Field names (and other related magic words) in the Template with their "
    "current Values to make the command line.",
    PN_COMMAND_LINE, {}, {}, {}},

  {g_PropertyNames[PN_TRIGGER],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Get the Trigger string used by this Command.",
    PN_TRIGGER, {}, {}, {}},

  {g_PropertyNames[PN_TRIGGER],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Set the Trigger string used by this Command.  A Trigger string can be "
    "several things, which control when this Command is started.  If it's an "
    "empty string, the Command can only be manually started.  If it's "
    "\"@Launch\" then the Command starts up when " PROGRAM_NAME_WORD " "
    "starts running (usually soon after the computer boots up, if it's "
    "configured that way).  If it's a Cron style string then the Command is "
    "started whenever the current time (local time zone) matches.  Finally, "
    "if it's the name of some other Command then this Command is started "
    "when that other Command finishes successfully (successful means with an "
    "exit code of zero).  I might as well explain Cron style time strings "
    "here.  Technically they're a modified version of the Unix crontab "
    "format (look it up online if you wish to read more).  It's a way of "
    "specifying dates and times, formatted as five fields separated by "
    "white space in this order (opposite of crontab order): DayOfWeek Month "
    "DayOfMonth Hour Minute.  Each field can contain a single value, a "
    "bunch of values separated by commas, a range of values using a hyphen, "
    "or be an asterisk to mean all possible values.  All fields must match "
    "to activate a time trigger.  So to specify the top of the hour at 2am "
    "on January 6th (years aren't mentioned, so it would happen every year "
    "at that time), it would be \"* 1 6 2 0\".  To specify 11:30pm on work "
    "days it would be \"Mon-Fri * * 23 30\".  For something happening every "
    "quarter hour from noon to just before 5pm (16:45 would be the last "
    "time) it would be \"* * * 12-16 0,15,30,45\".  The sneakiest may be "
    "for things like the first Friday of the month \"Fri * 1-7 0 0\" which "
    "activates at midnight if the day is a Friday and in the first 7 days "
    "of the month (one of them must be a Friday?).  Note that the standard "
    "Unix crontab interpretation does it differently when combining day of "
    "week and day of month.  The valid range for minutes is 0 to 59, hours "
    "0-23, days 1-31, months 1-12 (January to December), day of week 0-6 "
    "(Sunday, Monday, ..., Saturday).",
    PN_TRIGGER, {}, {}, {}},

  {g_PropertyNames[PN_CHILDREN],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the list of child Commands to this Command, and other relatives.  "
    "The \"result\" field is an array of strings naming the children, can be "
    "none.  Parent / child relationships are determined by trigger strings "
    "listing the name of the parent Command rather than a start time.  You "
    "can add an optional boolean of \"subsidiary\", set it to TRUE to "
    "restrict the list to just subsidiary child Commands.  Setting it to "
    "FALSE gives you a list of just the regular Command children.  Don't "
    "include the boolean if you want all children.  This operation also "
    "returns the names of some relatives.  The name of the ancestor Command "
    "is in a field named \"ancestor\", which only exists for Subsidiary "
    "Commands.  If the Command has a parent (an actually existing Command "
    "named by the Trigger string) it will be named in a \"parent\" field, "
    "otherwise the field will be missing.",
    PN_CHILDREN, {}, {}, {}},

  {g_PropertyNames[PN_DISABLE],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the disable status of this Command.  Result field is TRUE if it is "
    "disabled.",
    PN_DISABLE, {}, {}, {}},

  {g_PropertyNames[PN_DISABLE],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Set the disable status of this Command.  Add a boolean \"data\" field "
    "set to TRUE to disable the Command.  FALSE will enable it.  An enabled "
    "Command will run as usual.  A disabled Command won't even try to run at "
    "scheduled times, ignoring the Trigger string.  However, if it is a "
    "Subsidiary Command (is marked as Subsidiary and has the name of its "
    "parent Command specified in its Trigger) and that parent finishes "
    "running successfully, this Command will partly run.  A partial run will "
    "add a log entry saying that this Command was skipped because it was "
    "disabled, and any child Commands (ones with this Command listed as "
    "their trigger parent) will then be run.  This is needed so that you can "
    "disable individual processing steps (each one is a child Command of the "
    "previous step) without affecting later ones.",
    PN_DISABLE, {}, {}, {}},

  {g_PropertyNames[PN_SUBSIDIARY],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the Subsidiary status of this Command.  Result field is TRUE if it "
    "is marked as a Subsidiary.",
    PN_SUBSIDIARY, {}, {}, {}},

  {g_PropertyNames[PN_SUBSIDIARY],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Sets the Subsidiary status of this Command.  Set a boolean \"data\" "
    "BMessage field to TRUE to mark it as a Subsidiary, FALSE as a regular "
    "Command.  A Subsidiary Command usually has a parent Command listed in "
    "its Trigger string (ones without shouldn't happen, will be deleted "
    "when detected).  It will use its nearest non-subsidiary "
    "ancestor Command's log file for log messages.  The parents all the way "
    "up to that ancestor will wait for subsidiary children to finish running "
    "before they stop running.  Subsidiary Commands also follow their "
    "ancestor's edit state, and get deleted when the ancestor is deleted.  "
    "When the ancestor is renamed, Subsidiary Commands all get renamed as "
    "the ancestor's name plus a sequence number plus their original Model "
    "name (if it exists).  Models include their Subsidiary child Commands.",
    PN_SUBSIDIARY, {}, {}, {}},

  {g_PropertyNames[PN_LOG],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Get the accumulated log output (currently Standard Output and Standard "
    "Error combined) from the execution of this Command.  The accumulated "
    "data (which is stored in memory) is also cleared at this point.  The "
    "serial number is also changed whenever new log data arrives.  If you are "
    "running a particularly verbose command, it's best to periodically get "
    "the log output data to avoid missing anything since each Command's "
    "in-memory buffer only holds " MAX_LOG_LENGTH_AS_STRING " bytes.",
    PN_LOG, {}, {}, {}},

  {g_PropertyNames[PN_LOG_FILE_ENABLE],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns a boolean that is TRUE if the output from this Command is being "
    "logged to a file.  FALSE if it's not being filed away.",
    PN_LOG_FILE_ENABLE, {}, {}, {}},

  {g_PropertyNames[PN_LOG_FILE_ENABLE],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Turns on or off file logging of this Command's output.  Takes effect "
    "only when the Command starts running, not during.  The boolean \"data\" "
    "field needs to be set to TRUE to enable file logging, false to turn it "
    "off.  Regular in-memory logging proceeds irregardless.  Logged text "
    "will be appended to the file; it's up to you to delete it if it gets "
    "too big.  The log file will have the same name as the Command (minus "
    "slashes) and be stored in the Logs directory in the program's "
    "subdirectory in the user settings directory.  Typically it will be "
    "named something like /boot/home/config/settings/" PROGRAM_NAME_WORD
    "/Logs/CommandName.  If you don't want it there, replace the file with "
    "a symbolic link to a different file in a different place.  Note that "
    "Haiku circa 2018 occasionally crashes when appending to a file, so "
    "the default for new Commands is no logging.",
    PN_LOG_FILE_ENABLE, {}, {}, {}},

  {g_PropertyNames[PN_PATHS],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns strings for the various stock paths for this Command.  The one "
    "at key \"SOCTempDir\" is the temporary directory path for this "
    "particular Command (may get erased during reboots, usually inside /tmp, "
    "a sanitised version of the Command name will be somewhere in the path).  "
    "\"SOCWorkDir\" is the directory for the Command's more persistent data, "
    "usually somewhere under the " PROGRAM_NAME_WORD " settings directory.  "
    "There's also \"SOCLogFile\" for the path to the log file, though the "
    "file might not exist.  Returned strings don't end in a slash.",
    PN_PATHS, {}, {}, {}},

  {g_PropertyNames[PN_FIELD],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Create a new Field inside the Command.  The Command needs to be in Edit "
    "mode.  The Field's default name is \"Unnamed\".  You can set a name for "
    "the new Field by adding a string named \"name\" to the BMessage used for "
    "scripting.  The Hey command would look like: hey " PROGRAM_NAME_WORD
    " let Command [0] do create Field with name=NewFieldNameHere",
    PN_FIELD, {}, {}, {}},

  {g_PropertyNames[PN_FIELD],
    {B_COUNT_PROPERTIES, 0}, {B_DIRECT_SPECIFIER, 0},
    "Counts the number of Fields in the Command.",
    PN_FIELD, {}, {}, {}},

  {g_PropertyNames[PN_FIELD],
    {0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Specify a particular Field sub-object of a Command for further "
    "scripting.  It's better to refer to Fields by name than by index, since "
    "the array is kept in alphabetical order and adding or removing Fields "
    "will affect the indices of several other Fields.",
     PN_FIELD, {}, {}, {}},

  {0, {0}, {0}, 0, 0, {}, {}, {}} /* End of list of property commands. */
};


/* This array lists the scripting commands the Field object can handle. */

static struct property_info g_FieldScriptingPropertyList [] =
{
  /* *name; commands[10]; specifiers[10]; *usage; extra_data; ... */
  {g_PropertyNames[PN_NAME],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the name of this Field.",
    PN_NAME, {}, {}, {}},

  {g_PropertyNames[PN_NAME],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Changes the name of this Field.  This will also affect the order of "
    "the Fields array (which is always in alphabetical order), so the index "
    "to this Field and others may change.  Fails if the new name is already "
    "in use or the parent Command isn't in edit mode.",
    PN_NAME, {}, {}, {}},

  {g_PropertyNames[PN_SERIAL],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns the change serial number for this Field, increased when this "
    "object changes in any way (usually due to user editing).",
    PN_SERIAL, {}, {}, {}},

  {g_PropertyNames[PN_FIELD],
    {B_DELETE_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Delete this Field.  Fails if the parent command is not in edit mode.",
    PN_FIELD, {}, {}, {}},

  {g_PropertyNames[PN_VALUE],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the Value of the Field.",
    PN_VALUE, {}, {}, {}},

  {g_PropertyNames[PN_VALUE],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Sets the Value of the Field.  This is the user provided string that "
    "will be substituted in the command line everywhere the Field name is "
    "present.",
    PN_VALUE, {}, {}, {}},

  {0, {0}, {0}, 0, 0, {}, {}, {}} /* End of list of property commands. */
};


/* This array lists the scripting commands for manipulating UserData.  It's
implemented by both Commands and Fields.  Unfortunately, we can't put it in the
BApplication too due to class inheritance limitations. */

static struct property_info g_UserDataScriptingPropertyList [] =
{
  /* *name; commands[10]; specifiers[10]; *usage; extra_data; ... */

  {g_PropertyNames[PN_USER_DATA],
    {B_SET_PROPERTY, 0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "UserData is used for storing key and value pairs of user specified data "
    "into scriptable objects, currently Commands and "
    "Fields.  It's often used by a GUI on top of this program to store "
    "things such as the data type related UI to pop up to edit a particular "
    "Field.  The key (passed in the name specifier of the scripting "
    "operation) is a non-empty UTF-8 string, with case insensitive matching.  "
    "The value can be any of the BMessage field data types (including a "
    "BMessage) and can also be an array of them.  This scripting operation "
    "changes the value stored in the UserData entry for the specified name "
    "or at the specified index to the contents of your \"data\" argument(s).  "
    "If you don't specify any data, then the contents become NULL and when "
    "you get it next time, the \"result\" field will be missing.  If the "
    "entry doesn't exist and you used the name specifier, it will be created "
    "(thus the lack of a B_CREATE_PROPERTY operation).  For example: hey "
    PROGRAM_NAME_WORD " set UserData \"DisplayType\" of field \"A-URL\" of "
    "Command \"Download New Files\" to \"URL\"",
    PN_USER_DATA, {}, {}, {}},

  {g_PropertyNames[PN_USER_DATA],
    {B_DELETE_PROPERTY, 0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Deletes the UserData entry for the specified name or at the specified "
    "index.",
    PN_USER_DATA, {}, {}, {}},

  {g_PropertyNames[PN_USER_DATA],
    {B_COUNT_PROPERTIES, 0}, {B_DIRECT_SPECIFIER, 0},
    "Counts the number of UserData entries currently stored.  "
    "Coincidentally and not surprisingly, valid index specifiers are from "
    "zero up to this number minus one.",
    PN_USER_DATA, {}, {}, {}},

  {g_PropertyNames[PN_USER_DATA],
    {B_GET_PROPERTY, 0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Gets the value stored in the UserData entry for the specified name or "
    "at the specified index number (zero based, be wary of the shuffling "
    "which happens when items are added or removed - they're kept in "
    "alphabetical order).  Returned in the usual \"result\" output, which "
    "will be missing if the stored value is NULL.\n\n"

    "Here are some standard UserData keys/values:\n\n"

    "Datatype of Input/string.  Names the kind of data this Command uses as an "
    "input.\n\n"

    "Datatype of Output/string.  Names the kind of data this Command outputs.  "
    "Particular data types are:\n"
    "\"BaseURL\" for a text file named SOCWorkDir/BaseURL containing the URL "
    "to the remote directory (ends with a slash) of files to be downloaded, "
    "without any end of line characters.\n"
    "\"DownloadURLList\" for a text file named SOCWorkDir/URLsToDownload "
    "containing a list of URLS (one per line) to be downloaded.\n"
    "\"ProcessDir\" for all the files in directory SOCTempDir/Processing.  "
    "They'll get batch processed in some way, perhaps renaming or adjusting "
    "audio levels.\n\n"

    "Description/string.  User notes, can be multiple paragraphs.\n\n"

    "DisplayType/[CheckBox | JustNewOps | LocalDir | LocalFile | LocalPath | "
    "MultipleChoice | MultipleChoiceWithText | "
    "String | URL | URLDirOnly | Renamer]\n"
    "Specifies how to display a Field in the GUI version of the program.  If "
    "not specified, you get the usual String display.  A DisplayType of "
    "\"CheckBox\" shows a checkbox which sets the Field's value to \"0\" or "
    "\"1\".  \"JustNewOps\" is a specially made user interface for the "
    "\"Fetch Just New Files\" Command Model which has buttons to reset the "
    "list of downloaded files in various ways (its Field is just a dummy).  "
    "\"MultipleChoice\" uses a multiple choice pop-up menu, with "
    "UserData[\"Choices\"] specifying the choices separated by vertical bar "
    "characters, and UserData[\"Labels\"] contains wordier text describing "
    "each choice (defaults to the corresponding Choices value if an empty "
    "string or if not specified).  \"MultipleChoiceWithText\" is similar but "
    "has an additional text box where the user can type in new choices.  The "
    "\"LocalDir\" one uses a file requester to select a directory, the "
    "resulting path is absolute if it starts with a slash, or relative to the "
    "current Station directory.  Often the path ends with a slash for "
    "directories, though the user can edit it.  The \"LocalFile\" similarly "
    "uses a file requester to select a file, relative to the Station "
    "directory.  \"LocalPath\" lets the user choose either a file or a "
    "directory.  The \"URL\" DisplayType breaks apart the URL (only FTP and "
    "HTTP(S) ones) into all the parts and lets the user edit each part "
    "individually before reassembling the result into a new URL (with "
    "encoding of odd characters too).  \"URLDirOnly\" is similar except that "
    "it doesn't ask for a file name and only works for FTP.  \"Renamer\" "
    "prompts for information needed to rename a file and builds a SED (stream "
    "editor) command string to rename an original file name into a new "
    "name.\n\n"

    "Documentation/string.  Documentation for the Command (could be for "
    "Fields too, but it isn't currently displayed).  Typically will list the "
    "inputs and outputs for subsidiary Commands.  Can be multiple "
    "paragraphs.\n\n"

    "GroupBox/string.  Starts a multiple Field grouping box, with the string "
    "as the title of the box.  The box contains this Field and the following "
    "ones (in alphabetical order) up to the last Field in the Command or "
    "stops when a Field with another GroupBox definition is encountered.  "
    "A string of \"NoBox\" turns off the grouping box for that Field and "
    "following ones.  If the string starts with \"+\" then the box is visible "
    "in both basic and advanced modes (and the + isn't shown to the user), "
    "otherwise the box is only visible in basic mode.\n\n"

    "GUITitles/string.  Changes the title of various GUI elements used "
    "in displaying a Field like the LocalDir one from the default of \"Save "
    "files to\", \"Change\", \"Show\", \"Choose save directory for\" and "
    "\"Use\" to be whatever the strings are.  The different elements "
    "separated by \"|\" characters and the order of the elements is Field "
    "DisplayType specific.  We haven't gotten around to implementing it for "
    "all field types.\n\n"

    "HelpText/string.  One line (max about 120 characters) of help text "
    "about a Field or a Command, which will be displayed in the status bar "
    "of the GUI version of the program when the mouse is over that Field or "
    "Command.\n\n"

    "Original Model/string+string.  If present then the user has not edited "
    "the template (they may have changed fields etc).  The original Model's "
    "name is given by the string.  A second string stores the version of the "
    "Model; the datestamp when the Model was saved.  Later on when doing "
    "software updates, we can replace the Template and other properties of "
    "the Command with the more recent Template from the Model of the same "
    "name.  When the user changes the Template of a Command, this entry is "
    "deleted.\n\n"

    "VisibleModes/string.  A single character that specifies when the "
    "related Command (subsidiary or not) or Field is visible to the user in "
    "the FetchIt user interface.  ' ' (space) hides it, 'A' makes it show "
    "only in Advanced mode, 'B' makes it appear in Basic mode, 'C' shows up "
    "in both modes.  If not specified, it shows up in all modes.",
    PN_USER_DATA, {}, {}, {}},

  {0, {0}, {0}, 0, 0, {}, {}, {}} /* End of list of property commands. */
};



/* Utility function for searching a g_...ScriptingPropertyList array to see if
a given command is in it. */

static bool IsItOurScriptingCommand (struct property_info *InfoPntr,
  const char *PropertyName, uint32 SpecificationKind)
{
  int iKind;

  for (; InfoPntr->name != NULL; InfoPntr++)
  {
    if (strcasecmp (PropertyName, InfoPntr->name) == 0)
    {
      for (iKind = 0; iKind < (int) sizeof (InfoPntr->specifiers) /
      (int) sizeof (InfoPntr->specifiers[0]); iKind++)
      {
        if (InfoPntr->specifiers[iKind] == 0)
          break; // Zero terminated array of specifier kinds.
        if (InfoPntr->specifiers[iKind] == SpecificationKind)
          return true; // Found it.
      }
    }
  }
  return false;
}



/******************************************************************************
 * A base class for common stuff that our BeOS/Haiku system scriptable object
 * needs.  Also implements the change tracking serial number system and the
 * UserData (a key/value dictionary for user use) scripting functionality.  The
 * base BHandler's name is used as the object's name.
 *
 * Subclasses respond to scripting messages thus the BHandler parent class.
 * It's also archiveable into a BMessage, thus the _EXPORT.  For multitasking
 * simplicity and because there isn't any performance need for it, the
 * application is used as the BLooper.  That does mean the app's BLooper has
 * zillions of BHandlers in its list, hope it doesn't get too slow.
 */

class _EXPORT BaseChangeScriptable : public BHandler
{
public: // Constructor and destructor.
  BaseChangeScriptable (const char *NewName);
  BaseChangeScriptable (BMessage *ArchiveMessagePntr);
  virtual ~BaseChangeScriptable ();

public: // BeOS virtual functions and other BeOS callbacks.
  virtual status_t Archive (BMessage *ArchiveMessagePntr,
    bool Deep = true) const;
  virtual status_t GetSupportedSuites (BMessage *MessagePntr);
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual BHandler *ResolveSpecifier (BMessage *MessagePntr, int32 Index,
    BMessage *SpecifierMsgPntr, int32 SpecificationKind,
    const char *PropertyNamePntr);

public: // Our functions.
  void AddToLooper (BLooper *LooperPntr);

  virtual void ChangeHappened (bool SettingsAffected = true);
    /* Update our serial number so that users of this object can quickly tell
    that it has changed by comparing serial numbers.  Also inform our parent
    object (whatever it is) that a change happened (which is why this is a
    virtual function).  If SettingsAffected is TRUE, it will also request a
    save of the settings file. */

  int GetSerial ();
  void RemoveFromLooper ();

  bool UserDataAddString (const char *KeyPntr, const char *ValueStringPntr);
    /* Creates a UserData entry under the given Key, with the given string
    value.  Returns TRUE if successful, FALSE if something is wrong.  If the
    key already exists the value will be added as a second or later string in
    the "result" BMessage field. */

  bool UserDataDelete (const char *KeyPntr);
    /* Deletes the UserData entry under the given Key.  Returns TRUE if
    successful, FALSE if something is wrong, including key not found. */

  const char * UserDataGetString (const char *KeyPntr,
    const char *DefaultValue = NULL, int32 Index = 0);
    /* Looks for the UserData under the given key.  If it exists and is a
    string then it's returned.  Otherwise the DefaultValue will be returned.
    If you don't specify DefaultValue or it's NULL, an empty string will be
    used for that purpose.  Index is for UserData arrays of strings; usually
    there's only one string. */

  status_t UserDataSet (const char *KeyPntr, BMessage *SourceMessagePntr);
    /* Copies the "data" or "result" fields (priority given to "data") from the
    SourceMessagePntr to the UserData under the given key name.  If something
    goes wrong (out of memory or no source fields found), may leave behind an
    empty UserData entry under that key. */

  void UserDataUpgradeFromNewerModel (BMessage *ModelPntr);
    // Grab newer UserData from the model and replace existing entries with the
    // new ones from the Model.  Leaves UserData not mentioned in the Model
    // unchanged (presumably that is end user provided stuff like notes).

protected: // Member variables, types, etc.
 typedef map<const char *, BMessage, LessThanCharStringStruct>
   UserDataMap; // Note key string is manually allocated and freed.

  int m_ChangeSerial;
    /* Incremented every time something is changed in this object, or in a
    child object. */

  UserDataMap m_UserData;
    /* A collection of key and value pairs for storing user specified data.
    Values are BMessages containing "result" field(s), which are usually a
    single string but can be any BMessage data type.  Often used by a GUI on
    top of this program to store things associated with Commands (whether to
    use the basic or advanced GUI for a Command) and Fields (such as Field data
    type). */
};



/******************************************************************************
 * A Field is a holder for a chunk of information that the user specifies, such
 * as a file name or URL, which gets inserted in the command line at a certain
 * spot.  It can also specify the type of information, so we can use specific
 * ways of prompting the user for the Field's value (plain string input vs
 * popping up a file requestor if we know it is a file name).
 *
 * It responds to scripting messages for doing things to the Field.  It's also
 * archiveable into a BMessage, thus the _EXPORT.
 */

class _EXPORT Field : public BaseChangeScriptable
{
public: // Constructor and destructor.
  Field (const char *NewName);
  Field (BMessage *ArchiveMessagePntr);
  virtual ~Field ();

public: // BeOS virtual functions and other BeOS callbacks.
  virtual status_t Archive (BMessage *ArchiveMessagePntr,
    bool Deep = true) const;
  virtual status_t GetSupportedSuites (BMessage *MessagePntr);
  static BArchivable *Instantiate (BMessage *ArchiveMessagePntr);
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual BHandler *ResolveSpecifier (BMessage *MessagePntr, int32 Index,
    BMessage *SpecifierMsgPntr, int32 SpecificationKind,
    const char *PropertyNamePntr);

public: // Member functions.
  virtual void ChangeHappened (bool SettingsAffected = true);

  void ExpandMagicWords (BString &CommandLine);
    /* Replaces all keywords related to this field in the given string. */

  void SetOwnerCommand (class Command *CommandPntr);

private: // Member variables.
  class Command *m_OwnerCommand;
    /* Points to the Command that owns this field, or NULL.  Useful for
    propagating changes. */

  BString m_Value;
    /* The main value of this Field.  This string will replace every occurrence
    the Field's name in the command line. */
};



/******************************************************************************
 * A Command contains everything needed to run a command line program (values
 * for the Fields, the command line template, log output).  It also responds to
 * scripting commands for doing things with the command and is archiveable into
 * a BMessage, thus the _EXPORT.
 */

class _EXPORT Command : public BaseChangeScriptable
{
public: // Constructor and destructor.
  Command (const char *NewName);
  Command (BMessage *ArchiveMessagePntr);
  virtual ~Command ();

public: // BeOS virtual functions and other BeOS callbacks.
  virtual status_t Archive (BMessage *ArchiveMessagePntr,
    bool Deep = true) const;
  virtual status_t GetSupportedSuites (BMessage *MessagePntr);
  static BArchivable *Instantiate (BMessage *ArchiveMessagePntr);
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual BHandler *ResolveSpecifier (BMessage *MessagePntr, int32 Index,
    BMessage *SpecifierMsgPntr, int32 SpecificationKind,
    const char *PropertyNamePntr);

private: // More class specific types, for use only by this class.
 typedef map<const char *, class Field *, LessThanCharStringStruct>
   FieldMap;

public: // Member functions, some semi-private internal ones (used by Fields).
  void AncestryReset (); // Makes child list empty, NULL parent & ancestor.
  bool AncestryUpdateAncestry (int MaximumDepth); // Find ancestor, can fail.
  void AncestryUpdateParent (); // Find parent and add to parent's child list.
  void AppendLogText (const char *AddText, int32 AddLength = -1);
  virtual void ChangeHappened (bool SettingsAffected = true);
  bool DoesNamedFieldExist (const char *FieldNamePntr);
  void InternalAddChildren (const char *AncestorName, int Depth);
  void InternalAddField (Field *FieldPntr);
  void InternalRemoveChildren ();
  void InternalRemoveField (Field *FieldPntr);
  inline bool IsEditing () const
    { return m_CommandState == CS_EDITING; };
  inline bool IsRunningOrWaiting () const
    { return m_CommandState == CS_RUNNING ||
      m_CommandState == CS_WAIT_SUBSIDIARY; };
  inline bool IsSubsidiary () const
    { return m_Subsidiary; };
  void ParentCommandFinished ();
  void PollAndUpdate (const struct tm *CurrentTime);

  void SetStateRecursive (CommandState NewState);
   // Normally for internal use only.  Sets the state and clears pending runs
   // due to a trigger by another Command.  Recursively sets the state of all
   // subsidiary Commands too.  Does a ChangeHappened() if state changed.

  void UpgradeFromNewerModel (BMessage *ModelPntr);
    // Grab newer data (Template, UserData) from the model (a newer version of
    // the one which was cloned to make this Command) to update this Command so
    // that it works with newer software and settings.  Caller will update
    // Command's Model version number.  Recursively updates the Fields from the
    // relevant portions of the same Model.

  void UseParentNameAsTrigger ();

private: // Completely private member functions.
  void ExpandTemplateIntoCommandLine ();
  void GetSanitisedName (BString &OutputName);
  void GetStockNamesAndPaths (BString *OutputName, BString *OutputLogFile,
    BString *OutputTempDir, BString *OutputWorkDir);
  void ParseTriggerString ();
  status_t RunCommand (char *ErrorTextPntr, int ErrorTextSize);
  status_t StopCommand (char *ErrorTextPntr, int ErrorTextSize);

private: // Member variables.
  Command *m_CommandAncestor;
    /* The first ancestor (parent or parent of parent etc) Command of this
    Command which is also a non-subsidiary Command.  NULL if no ancestor (a
    regular Command or an orphaned subsidiary Command which shouldn't exist).
    Will write log entries to the ancestor's log file, follow its edit state,
    etc, see documentation of Subsidiary for details.  Just for convenient
    reference, it's not definitive. */

  vector <Command *> m_CommandChildren;
    /* A list of all children of this Command, usually including both
    subsidiary and regular Commands.  When the Command is in the global
    application's list (map actually) of Commands, then the children are owned
    by the app.  When the Command is independent then it owns the listed
    children (all non-subsidiary ones are removed) and they get deallocated
    when the Command is deallocated.  That way the Command and related children
    can be handled as a block, for saving in Models and the undo saved state.
    Similarly, children are stored separately in the settings file for
    Commands, but are nested recursively for Models.  The canonical list of all
    Commands is stored in the application's map, and parents are defined by
    Trigger strings.  Can be invalid at times, regenerated (see
    RecalculateAncestry) after loading, creation, renaming and trigger changing
    operations. */

  Command *m_CommandParent;
    /* The parent Command of this Command, NULL if none.  Just for convenient
    reference, it's not definitive (the Trigger string is). */

  int m_ChildProcessExitCode;
    /* The exit code from the child process running the command line.  Well,
    the child is usually the /bin/sh command interpreter, but it passes on up
    the byte sized exit codes if asked nicely. */

  pid_t m_ChildProcessID;
    /* Process ID of the child (the /bin/sh program) executing the command
    line.  Set to -1 when the child isn't running.  Also used as the process
    group number for the /bin/sh and its children, so they can conveniently all
    be killed off when required. */

  BString m_CommandLine;
    /* The last used command line for this Command.  Generated by replacing
    Field names with the corresponding Field Values.  See also
    m_CommandLineSerial. */

  int m_CommandLineSerial;
    /* Serial number of this Command at the last time the m_CommandLine was
    generated from the m_Template and Field Values.  If it is different from
    the Command's current serial number, then the m_CommandLine needs to be
    regenerated the next time it is used. */

  CommandState m_CommandState;
    /* States to prevent various things from interfering, like trying to edit a
    running command.  Usually changed through SetStateRecursive (). */

  BMessage m_CopyOfCommandBeforeEditing;
    /* When editing a Command, a copy of its state (including fields) is saved
    here in BArchive format, in case the user wants to cancel editing and
    revert back to the original state.  Subsidiary Commands are included in
    this BMessage too. */

  bool m_Disabled;
    /* Set to TRUE to disable the Command.  A disabled Command won't even try
    to run at scheduled times.  However, if it is the child of a Command which
    finishes running then it is partly run.  A partial run will add a log entry
    saying that this Command was skipped because it was disabled, and any child
    Commands will then be run. */

  FieldMap m_FieldMap;
    /* A collection of Fields, which are used to build the command line when
    this Command is executed.  The Fields are also scriptable objects and get
    saved inside the Command when it is Archived.  The name of the Field is the
    key. */

  struct tm m_LastRunStartTime;
    /* Time when the program was last run, broken down into seconds, minutes,
    etc.  Local time zone assumed.  Initialised to all zeroes, so you can
    detect the not-run-yet state, which happens after loading and maybe
    editing. */

  BString m_LogAccumulated;
    /* A string containing the accumulated standard output and standard error
    text from running the Command.  Gets cleared when someone gets the current
    log text.  Limited to at most MAX_LOG_LENGTH bytes stored.  Not used in
    subsidiary Commands. */

  bool m_LogFileEnabled;
    /* TRUE if logging to a file is enabled for this Command.  Will open (for
    append) the file when the Command starts running, and close it when
    finished running, and write the Commands output to it while running.  Not
    used in subsidiary Commands. */

  BFile m_LogFile;
    /* The BFile is usually opened on the log file.  Closed when the Command
    isn't running.  InitCheck() will tell you if the file is open.  Not used in
    subsidiary Commands, they use their ancestor's log. */

  int m_NoChildrenPollCount;
    /* Counts the number of poll passes when we're waiting in state
    CS_WAIT_SUBSIDIARY for running children to finish and there are no
    children.  If this happens a few times, then we can stop running.  Needed
    because children take a poll or two to start running. */

  int m_PipeStdOutFromChild[2];
    /* Pairs of file handles for the pipes for communicating with the child
    process running the actual Command.  [0] is the FD (file descriptor) number
    for the read end of the pipe, [1] is the write end.  Both are set to -1
    when the pipe is not open.  They're set to O_CLOEXEC (close on executing)
    mode in the main program thread so that copies (due to the way fork()
    works) of them in later Command child threads are automatically closed.
    They're also set to O_NONBLOCK mode so that reads from them will return
    immediately if there is no data available. */

  bool m_Subsidiary;
    /* TRUE if this is a Subsidiary Command, FALSE for regular Commands. */

  BString m_Template;
    /* The template used for building command lines.  Magic words, like Field
    names, get replaced with related strings, like Field Values.  Also see the
    "Original Model" key in UserData, which keeps track of whether the user has
    edited the Template. */

  BString m_Trigger;
    /* The time and other things trigger for starting this Command, as
    specified by the user.  "@Launch" starts it when the program starts.  A
    cron style string specifies that it starts at the given times (parsed value
    is stored in m_TriggerCronBits).  An empty string means it can only be
    manually started.  Otherwise it is the name of another Command (which may
    not exist yet) and specifies that it starts when the other command finishes
    successfully.  See m_TriggerType for which one it is. */

  CronBitset m_TriggerCronBits;
    /* If the Trigger is a cron style string, the decoded bitsets representing
    the time are stored here.  For other trigger types, this isn't used. */

  bool m_TriggerNamedCommandHasFinished;
    /* Set to TRUE when the Command named by m_Trigger (now known as the parent
    Command) has finished successfully.  Cleared when this Command changes
    state, so that it is only used when the parent Command finishes running
    while this one is in the ready to run state. */

  TriggerType m_TriggerType;
    /* When m_Trigger is changed, the string is parsed to determine what kind
    it is (manual, cron, reboot, name of a Command) and the result stored in
    this enum to save on having to figure it out every time it's used. */
};



/******************************************************************************
 * ScriptOCronApp is the BApplication class for this program.  Besides a bit of
 * bookkeeping (startup, shutdown), this main thread handles scripting messages
 * that manipulate the Commands and Fields.  Conveniently, this avoids
 * multiprocessing pitfalls.  All the operations are fairly fast, so there
 * shouldn't be a bottleneck with using a single message processing thread.  If
 * built with the GUI, the window thread also uses scripting messages sent to
 * this thread to do all its actions on Commands and Fields.
 */

class ScriptOCronApp : public BApplication
{
public: // Constructor and destructor.
  ScriptOCronApp ();
  ~ScriptOCronApp ();

public: // BeOS virtual functions.
  virtual void AboutRequested ();
  virtual void DispatchMessage (BMessage *MessagePntr, BHandler *Target);
  virtual status_t GetSupportedSuites (BMessage *MessagePntr);
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual void Pulse ();
  virtual void ReadyToRun ();
  virtual BHandler *ResolveSpecifier (BMessage *MessagePntr, int32 Index,
    BMessage *SpecifierMsgPntr, int32 SpecificationKind,
    const char *PropertyNamePntr);


public: // Member functions, some semi-private internal ones used by Commands.
  void ChangeHappened (bool SettingsAffected = true);
  void DeleteAllCommands ();
  void DeleteAllModels ();
  inline bool DoesNamedCommandExist (const char *CommandNamePntr) const
    { return  (InternalFindCommand (CommandNamePntr) != NULL); };
  const char *GetSettingsDirectory ();
  inline const char * GetUserSpecifiedTemporaryDirectory ()
    { return m_UserSpecifiedTemporaryDirectory.String (); };
  inline void InternalAddCommand (Command *CommandPntr)
    { InternalAddCommandGuts (CommandPntr, CommandPntr->Name (), 0);
      RecalculateAncestry (); };
  void InternalAddCommandGuts (Command *CommandPntr,
    const char *AncestorName, int Depth);
  Command * InternalFindCommand (const char *CommandNamePntr) const;
  inline void InternalRemoveCommand (Command *CommandPntr)
    { InternalRemoveCommandGuts (CommandPntr);
      RecalculateAncestry (); };
  void InternalRemoveCommandGuts (Command *CommandPntr);
  void LoadModels (BMessage *ArchivePntr, bool BuiltIn);
  status_t LoadSaveSettings (bool DoLoad);
  void RecalculateAncestry ();
  void UpdateWithNewerModels ();

private: // Internal member functions.
  void DefaultSettings ();
  status_t LoadSaveSettingsFile (bool DoLoad, bool JustModels,
    BDirectory &Directory, const char *FileName);

private: // Class specific types.
 typedef map<const char *, class Command *, LessThanCharStringStruct>
   CommandMap;

 typedef map<const char *, BMessage *, LessThanCharStringStruct> ModelMap;

private: // Member variables.
  int m_ChangeSerial;
    /* Incremented every time something is changed in this object, or in a
    child object or in the settings.  Settings will be saved on program exit if
    anything has changed. */

  CommandMap m_CommandMap;
    /* A collection of commands, which scripting messages can be sent to.  They
    also show up in the list of commands in the main window.  There may be a
    few outside this Map, such as ones being edited (the user works on a copy
    of the command when editing).  The name of the command is the key. */

  ModelMap m_ModelMap;
    /* A collection of Models (saved Commands), each stored in archived
    BMessage format with the name of the Model (referencing a string within the
    BMessage) being the key in the map.  Used when creating a new Command to
    start off with things set for common tasks, rather than a fully empty
    Command.  See also m_SerialWhenModelsLastChanged. */

  BMessage *m_QuitRequestedStashedMessage;
    /* To delay the quit until Commands have stopped running (need a few more
    Pulse() times), when we get the B_QUIT_REQUESTED message it gets detached
    from the message queue and stored here.  After we're finished, we reply to
    it with the appropriate boolean flag in "result" and a message type of
    B_REPLY. */

  int m_SerialWhenModelsLastChanged;
    /* The m_ChangeSerial at the time when the models in m_ModelMap were last
    modified. */

  int m_SerialWhenSettingsSaved;
    /* The m_ChangeSerial at the time when the settings were last saved or
    loaded, so we can detect if they get modified. */

  BPath m_SettingsDirectoryPath;
    /* The constructor initialises this to the settings directory path.  It
    never changes after that. */

  bigtime_t m_TimeSettingsWereLastSaved;
    /* Used for automatically saving the settings once in a while. */

  BString m_UserSpecifiedTemporaryDirectory;
    /* The master temporary directory.  If it's an empty string, the system
    default temporary directory gets used.  Otherwise this path is used.
    Shouldn't end in a slash.  Doesn't need to exist since the program doesn't
    directly use it, however the bash templates usually create it when they
    need it (using mkdir --parents "SOCTempDir/Processing" or something like
    that). */

  int m_UserSpecifiedTemporaryDirectorySerial;
    /* m_ChangeSerial when m_UserSpecifiedTemporaryDirectory last changed. */
};



/******************************************************************************
 * UserData is stored in BMessages, usually in the "result" field.  But it
 * arrives in a field named "data" and sometimes the key string is also stored
 * in the BMessage.  This utility function copies a named field from the source
 * BMessage to the destination BMessage, optionally renaming it in the process.
 * If there are multiple fields with that name (an array), they're all copied.
 * Note that the two messages must be different, copying from a message to
 * itself won't work properly.  Returns B_NAME_NOT_FOUND and does nothing if
 * the source field doesn't exist.
 */

status_t CopyBMessageField (BMessage *SourceMessagePntr,
  BMessage *DestMessagePntr, const char *FieldName, const char *NewName)
{
  if (SourceMessagePntr == NULL || DestMessagePntr == NULL ||
  SourceMessagePntr == DestMessagePntr || FieldName == NULL ||
  FieldName[0] == 0 || (NewName != NULL && NewName[0] == 0))
    return B_BAD_VALUE; // Bad input arguments, do nothing.
  if (NewName == NULL)
    NewName = FieldName;

  int32 CountFound = 0;
  const void *DataPntr = NULL;
  status_t ErrorCode;
  bool FixedSize = false;
  type_code TypeFound = 0;

  ErrorCode = SourceMessagePntr->GetInfo (FieldName, &TypeFound, &FixedSize);
  if (ErrorCode != B_OK)
    return ErrorCode;
  ErrorCode = SourceMessagePntr->GetInfo (FieldName, &TypeFound, &CountFound);
  if (ErrorCode != B_OK)
    return ErrorCode;

  for (int32 iCount = 0; iCount < CountFound; iCount++)
  {
    ssize_t DataSize = 0;
    ErrorCode = SourceMessagePntr->FindData (FieldName, TypeFound, iCount,
      &DataPntr, &DataSize);
    if (ErrorCode != B_OK)
      return ErrorCode;

    ErrorCode = DestMessagePntr->AddData (NewName, TypeFound, DataPntr,
      DataSize, FixedSize, (iCount == 0) ? CountFound : 1 /* prealloc */);
    if (ErrorCode != B_OK)
      return ErrorCode;
  }
  return B_OK;
}



/******************************************************************************
 * Global utility function to return the current time as a readable string.
 * Not multithread safe - uses a global string array to return the result.
 */

static const char * CurrentTimeAsString ()
{
  static char TimeString[30];
  time_t TimeNow;
  struct tm BrokenUpTime;

  time (&TimeNow);
  localtime_r (&TimeNow, &BrokenUpTime);
  asctime_r (&BrokenUpTime, TimeString);

  // Strip off the end of line character.
  char *EOL = strchr (TimeString, '\n');
  if (EOL != NULL)
    *EOL = 0;

  return TimeString;
}



/******************************************************************************
 * Global utility function to display an error message and return.  The message
 * part describes the error, and if ErrorNumber is non-zero, gets the string
 * ", error code $X (standard description)." appended to it.  If the message
 * is NULL then it gets defaulted to "Something went wrong".  The title part
 * doesn't get displayed (no title bar in the dialog box, but you can see it in
 * the debugger as the window thread name), and defaults to "Error Message" if
 * you didn't specify one.  If running in command line mode, the error gets
 * printed to stderr rather than showing up in a dialog box.
 */

void DisplayErrorMessage (
  const char *MessageString = NULL,
  int ErrorNumber = 0,
  const char *TitleString = NULL)
{
  char ErrorBuffer [PATH_MAX + 1500];

  if (TitleString == NULL)
    TitleString = PROGRAM_NAME_WORD " Error Message";

  if (MessageString == NULL)
  {
    if (ErrorNumber == 0)
      MessageString = "No error, no message, why bother?";
    else
      MessageString = "Something went wrong";
  }

  if (ErrorNumber != 0)
  {
    sprintf (ErrorBuffer, "%s, error code $%X/%d (%s) has occurred.",
      MessageString, ErrorNumber, ErrorNumber, strerror (ErrorNumber));
    MessageString = ErrorBuffer;
  }

#if INCLUDE_GUI_CODE
  if (!g_ServerMode)
  {
    BAlert *AlertPntr;
    AlertPntr = new BAlert (TitleString, MessageString,
      "Acknowledge", NULL, NULL, B_WIDTH_AS_USUAL, B_STOP_ALERT);
    AlertPntr->SetShortcut (0, B_ESCAPE);
    if (AlertPntr != NULL)
      AlertPntr->Go (); // Synchronous mode, will wait until done.
  }
  else // Print message on standard error.
#endif //INCLUDE_GUI_CODE
  cerr << TitleString << ": " << MessageString << endl;
}



/******************************************************************************
 * This utility function combines a date format string with the current time
 * and a date offset to generate text that includes the current time.  It's
 * usually used for file names that are date based.  Works by calling the
 * system "date" executable or maybe using strftime().
 *
 * DateFormat is a format string like "%Y%m%d%H%M.mp3".  See "date --help" or
 * the info page on the "date" command for more details.  Leave off the leading
 * plus sign, it will be added for you.  Odd characters will be escaped before
 * running it, for safety.
 *
 * The date offset is something like "1 day ago 2 hours ago", "next week" or
 * "tomorrow".  Use an empty string for no offset from the current time.
 *
 * If the date command fails, an empty string will be returned for the output.
 * See the info page on the "date" command for details.
 */

void ExpandDateFormat (
  BString &Output,
  const char *DateFormat,
  const char *DateOffset)
{
  BString CLIDateCommand ("date ");

  // Insert the date offset argument if present.  Will modify the current date
  // with things like "1 year ago".

  if (*DateOffset != 0 )
  {
    BString DateOffsetEscaped (DateOffset);
    SafeEscape (DateOffsetEscaped);

    CLIDateCommand.Append ("\"--date=");
    CLIDateCommand.Append (DateOffsetEscaped);
    CLIDateCommand.Append ("\" ");
  }

  // Create the format part of the date command.  A leading plus sign is needed
  // to identify that it is a format for "date" to interpret.

  CLIDateCommand.Append ("\"+");
  BString DateFormatEscaped (DateFormat);
  SafeEscape (DateFormatEscaped);
  CLIDateCommand.Append (DateFormatEscaped);

  // Strip trailing newline from the output.

  CLIDateCommand.Append ("\" | tr -d \"\\n\"");

  // Run the date command and save the output, if any.  No output if
  // something's wrong, like a bad date offset.

  RunCLIString (Output, CLIDateCommand.String ());
}



/******************************************************************************
 * A utility function to get the next globally defined (so that multiple
 * simultaneous users get unique values) serial number.
 */

static int GetNextSerialNumber ()
{
#if __HAIKU__
  static long g_NextSerial = 0;
#else // BeOS
  static vint32 g_NextSerial = 0;
#endif

  return (int) atomic_add (&g_NextSerial, 1);
}



/******************************************************************************
 * A utility to run a CLI command from a string and save output to a string.
 * Returns the exit code from running the CLI command (zero customarily means
 * success), or negative if something went wrong.  Sets the output to an empty
 * string if things go wrong fast enough (a failed CLI command writes error
 * messages to stderr with no output to stdout), or to partial output if it ran
 * a bit before failing.
 */

int RunCLIString (BString &Output, const char *CommandLine)
{
#define MEASURE_CLI_RUN_TIME 0 /* For debugging set to non-zero. */
#if MEASURE_CLI_RUN_TIME
  bigtime_t StartTime = system_time ();
#endif

  Output = g_EmptyString;

  int PipeStdOutFromChild[2]; // Read pipe file descriptor [0], write at [1].

  if (pipe (PipeStdOutFromChild) != 0) // Returns -1 if failed, 0 if okay.
    return errno; // Pipe creation failed.

  g_LockStdFilesPntr->Lock (); // Hope it worked, garbled output otherwise.

  // Replace the standard output (both ours right now and the child when it
  // inherits it) with the pipe.  dup2 overwrites (closes) the old standard
  // out, and clears the close-on-exec flag too in the duplicate handle it
  // created for the new standard out.

  dup2 (PipeStdOutFromChild[1], STDOUT_FILENO);
  close (PipeStdOutFromChild[1]); // We're not using the write end of the pipe.

  // Child doesn't need to see the read end of the pipe, it just writes to the
  // pipe, so make it close-on-exec when the child starts.

  int PipeFlags = fcntl (PipeStdOutFromChild [0], F_GETFD);
  PipeFlags |= FD_CLOEXEC;
  fcntl (PipeStdOutFromChild [0], F_SETFD, PipeFlags);

  // Load the program we want to run, inheriting the current file handles.
  // Note that it doesn't do a fork (doesn't inherit the memory space of this
  // program and then throw it away), so it's faster, but trickier to set up.

  const char *ArgumentArray [4];
  ArgumentArray [0] = "/bin/sh";
  ArgumentArray [1] = "-c"; // Option for bash to run a string as a command.
  ArgumentArray [2] = CommandLine; // The string to run.
  ArgumentArray [3] = NULL;

  thread_id NewProgramID = load_image (
    3 /* argument count */, ArgumentArray, const_cast<const char **>(environ));

  status_t ErrorCode = NewProgramID; // Get a negative ID if load failed.
  if (ErrorCode >= 0)
  {
    // Switch the new program to a new process group (zero means actual group
    // number is copied from the thread ID, so it will be unique), so signals
    // sent the the main program (like control-C SIGTERM) don't also affect the
    // child.  Also makes the new child the head of a process group, so if
    // someone kills it, all the commands fired off by the child are killed
    // too.

    setpgid (NewProgramID, NewProgramID);

    ErrorCode = resume_thread (NewProgramID);
  }

  // Reset our standard output back to the normal one.  Child remains using the
  // pipe write end, and we share standard input and error with the child.

  dup2 (g_OriginalStdOut, STDOUT_FILENO);
  g_LockStdFilesPntr->Unlock ();

  // Read the read end of the pipe until end of file (happens when the child
  // has exited).

  if (ErrorCode >= 0)
  {
    while (true)
    {
      char Buffer [4096]; // BeOS pipes hold at most 4K, Haiku 64K.
      ssize_t AmountRead;
      AmountRead = read (PipeStdOutFromChild[0], Buffer, sizeof (Buffer));
      if (AmountRead <= 0)
        break; // End of file reached (zero), or error (-1).
      Output.Append (Buffer, AmountRead);
    }
  }
  close (PipeStdOutFromChild[0]);

  // Wait for the program to finish running and give us an exit code.

  status_t ProcessExitCode = 0;
  wait_for_thread (NewProgramID, &ProcessExitCode);

#if MEASURE_CLI_RUN_TIME
  bigtime_t Duration = system_time () - StartTime;
  printf ("RunCLIString took %.3fs for \"%s\" (%d)\n",
    Duration / 1000000.0f, CommandLine, (int) ProcessExitCode);
#endif

  if (ErrorCode < 0)
  {
    printf ("RunCLIString: Failed to start /bin/sh, code %08X (%s).  Was "
      "trying to evaluate \"%s\".\n",
      (int) ErrorCode, strerror (ErrorCode), CommandLine);
    return ErrorCode;
  }

  return ProcessExitCode;
}



/******************************************************************************
 * A utility function to backslash-escape dangerous characters in a string.
 */

void SafeEscape (BString &TheString)
{
  BString SafeString;
  char *SafePntr = SafeString.LockBuffer (2 * TheString.Length () + 1);

  const char *SourcePntr = TheString.String (); // Can be NULL for new strings.
  while (SourcePntr != NULL && *SourcePntr != 0)
  {
    char Letter = *SourcePntr++;
    if (Letter == '$' || Letter == '"' || Letter == '\\' || Letter == '`')
      *SafePntr++ = '\\'; // One of the ones needing a backslash escape.
    *SafePntr++ = Letter;
  }
  *SafePntr++ = 0;
  SafeString.UnlockBuffer ();
  TheString = SafeString;
}



/******************************************************************************
 * Word wrap a long line of text into shorter LineLength column lines and print
 * the result on the given output stream.  Appends a new line.  Zero LineLength
 * means don't wrap.
 */

static void WrapTextToStream (ostream& OutputStream, const char *TextPntr,
  const int LineLength = 79)
{
  if (LineLength <= 0)
  {
    OutputStream << TextPntr << endl;
    return;
  }

  char *StringPntr;
  char  TempString [LineLength+1]; // How modern, variable size locals!

  TempString[LineLength] = 0; /* Only needs to be done once. */

  while (*TextPntr != 0)
  {
    while (*TextPntr == ' ' || *TextPntr == '\t')
      TextPntr++; /* Skip leading spaces and tabs, but not newlines. */
    if (*TextPntr == 0)
      break; /* It was all spaces, don't print any more. */

    strncpy (TempString, TextPntr, LineLength);

    /* Advance StringPntr to the end of the temp string, partly to see how long
    it is (rather than doing strlen).  Stop at new line characters too, so they
    appear as a shorter line and don't wrap funny. */

    StringPntr = TempString;
    while (*StringPntr != 0 && *StringPntr != '\n')
      StringPntr++;

    if (StringPntr - TempString < LineLength)
    {
      /* This line fits completely. */

      if (*StringPntr == '\n')
        *StringPntr++ = 0; // Don't print line feed character + skip it too.

      OutputStream << TempString << endl;
      TextPntr += StringPntr - TempString;
      continue;
    }

    /* Advance StringPntr to the last space in the temp string. */

    while (StringPntr > TempString)
    {
      if (isspace (*StringPntr))
        break; /* Found the trailing space. */
      else /* Go backwards, looking for the trailing space. */
        StringPntr--;
    }

    /* Remove more trailing spaces at the end of the line, in case there were
    several spaces in a row. */

    while (StringPntr > TempString && isspace (StringPntr[-1]))
      StringPntr--;

    /* Print the line of text and advance the text pointer too. */

    if (StringPntr == TempString)
    {
      /* This line has no spaces, don't wrap it, just split off a chunk. */
      OutputStream << TempString << endl;
      TextPntr += strlen (TempString);
      continue;
    }

    *StringPntr = 0; /* Cut off after the first trailing space. */
    OutputStream << TempString << endl;
    TextPntr += StringPntr - TempString;
  }
}



/******************************************************************************
 * Print the usage info to the stream.  Optionally includes a list of all
 * commands and do word wrapping (zero to not wrap).
 */

void PrintUsage (ostream& OutputStream, int WrapWidth, bool ListCommands)
{
  struct property_info *PropInfoPntr;

#if INCLUDE_GUI_CODE
  WrapTextToStream (OutputStream,
    PROGRAM_NAME_FANCY " Downloader and command launcher.\n\n"
    "Copyright © 2018 TuneTracker Systems LLC.  Sold "
    "under the TuneTracker Systems Sales Terms and Conditions.", WrapWidth);
#else
  WrapTextToStream (OutputStream,
    PROGRAM_NAME_FANCY " - For editing and running commands.\n\n"
    PROGRAM_NAME_FANCY " copyright © 2018 by Alexander G. M. Smith.  "
    "Licensed under the MIT license.  Send e-mail to agmsmith@ncf.ca "
    "with questions or comments.", WrapWidth);
#endif // INCLUDE_GUI_CODE
  WrapTextToStream (OutputStream, "\n$Revision: 2.91 $ "
    __DATE__ ", " __TIME__ ".", WrapWidth);

#if INCLUDE_GUI_CODE
  if (!ListCommands)
    return; // Shorter about box for commercial GUI version.
#endif

  WrapTextToStream (OutputStream,
    "\nThis program maintains a library of commands which can be executed "
    "when requested or when a specified time of day arrives.  The commands "
    "can also be edited with a graphical user interface (available "
    "commercially) designed to keep the complexity down for ordinary "
    "users.\n\n"

    "A specialist sets up a template for each kind of command, specifying the "
    "full command line arguments and marking out key fields within those "
    "arguments for user editing, along with providing descriptions and "
    "default values for those key fields.  Then the ordinary user only needs "
    "to fill in the key fields, assisted by the descriptions.\n\n"

    "Commands are triggered either by a cron style time specification, "
    "by the user hitting the Run button, or by BeOS/Haiku scripting "
    "operations (which can be sent via the \"Hey\" utility program).\n\n"

    "Speaking of \"Hey\", scripting operations can be used for "
    "creating new Commands, running them and even modifying "
    "existing ones.  Send Hey's GetSuites command to the various different "
    "objects in " PROGRAM_NAME_WORD " (the Application itself, Command "
    "objects, Field objects) to see what scripting operations are available.  "
    "You can also see the list of scripting operations if you run "
    PROGRAM_NAME_WORD " from the command line.", WrapWidth);

  if (!ListCommands)
    return;

  /* Go through all our scripting commands and add a description of each one to
  the usage text. */

  static struct SuiteStruct {
    property_info *m_PropInfoPntr;
    const char *m_SuiteName;
  } ScriptingSuites [] =
  {
    {g_AppScriptingPropertyList, "application"},
    {g_CommandScriptingPropertyList, "command"},
    {g_FieldScriptingPropertyList, "field"},
    {g_UserDataScriptingPropertyList, "userdata"}
  };

  int iSuite;
  for (iSuite = 0;
  iSuite < (int) sizeof (ScriptingSuites) / (int) sizeof (ScriptingSuites[0]);
  iSuite++)
  {
    OutputStream << "\nScripting operations for " SUITE_BASE <<
      ScriptingSuites[iSuite].m_SuiteName << ":" << endl << endl;
    for (PropInfoPntr = ScriptingSuites[iSuite].m_PropInfoPntr;
    PropInfoPntr->name != 0; PropInfoPntr++)
    {
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY:
          OutputStream << "Get " << PropInfoPntr->name << endl;
          break;

        case B_SET_PROPERTY:
          OutputStream << "Set " << PropInfoPntr->name << " NewValue" << endl;
          break;

        case B_COUNT_PROPERTIES:
          OutputStream << "Count " << PropInfoPntr->name << endl;
          break;

        case B_CREATE_PROPERTY:
          OutputStream << "Create " << PropInfoPntr->name << endl;
          break;

        case B_DELETE_PROPERTY:
          OutputStream << "Delete " << PropInfoPntr->name << endl;
          break;

        case B_EXECUTE_PROPERTY:
          OutputStream << "Execute " << PropInfoPntr->name << endl;
          break;

        default:
          OutputStream << "N/A " << PropInfoPntr->name << endl;
          break;
      }
      WrapTextToStream (OutputStream, PropInfoPntr->usage, WrapWidth);
      OutputStream << endl;
    }
  }
}



/******************************************************************************
 * Wrapper function for printing usage via stream commands, like
 * cout << PrintUsage;
 */

ostream& PrintUsage (ostream& OutputStream)
{
  PrintUsage (OutputStream, 79 /* WrapWidth */, true /* ListCommands */);
  OutputStream.flush ();
  return OutputStream;
}



/******************************************************************************
 * Implementation of the BaseChangeScriptable class, constructor, destructor
 * and the rest of the member functions in mostly alphabetical order.
 */

BaseChangeScriptable::BaseChangeScriptable (const char *NewName) :
  BHandler (NewName),
  m_ChangeSerial (GetNextSerialNumber ())
{
}


BaseChangeScriptable::BaseChangeScriptable (BMessage *ArchiveMessagePntr) :
  BHandler (ArchiveMessagePntr),
  m_ChangeSerial (GetNextSerialNumber ())
{
  int32 CountFound = 0;
  status_t ErrorCode;
  type_code TypeFound = 0;
  UserDataMap::iterator UserDataIter;

  // Load up the UserData key/value pairs.

  ErrorCode = ArchiveMessagePntr->
    GetInfo (g_UserDataName, &TypeFound, &CountFound);
  if (ErrorCode == B_OK && TypeFound == B_MESSAGE_TYPE && CountFound > 0)
  {
    int32 iUserData;
    for (iUserData = 0; iUserData < CountFound; iUserData++)
    {
      BMessage UserDataMessage;
      ErrorCode = ArchiveMessagePntr->
        FindMessage (g_UserDataName, iUserData, &UserDataMessage);
      if (ErrorCode != B_OK)
        continue; // Skip bad ones.
      const char *KeyPntr = NULL;
      if (UserDataMessage.FindString (g_NameKey, &KeyPntr) != B_OK)
        continue; // Always need a key.
      if (KeyPntr == NULL || *KeyPntr == 0)
        continue; // Don't allow empty string keys.

      UserDataIter = m_UserData.find (KeyPntr);
      if (UserDataIter != m_UserData.end ())
        continue; // Duplicate key, ignore redundant keys.

      KeyPntr = strdup (KeyPntr); // Allocate our own memory for string.
      if (KeyPntr == NULL)
        continue;

      BMessage *ValueMessagePntr = &(m_UserData[KeyPntr]); // Allocates BMsg.
      ValueMessagePntr->what = B_SIMPLE_DATA;

      // Copy the "result" field(s) from the saved data.  Could be anything,
      // not necessarily just a single string.

      ErrorCode = CopyBMessageField (&UserDataMessage, ValueMessagePntr,
        g_ResultName);
      if (ErrorCode != B_OK) // Can happen if no "result" field.
        ValueMessagePntr->MakeEmpty (); // Effectively NULL user data.
    }
  }
}


BaseChangeScriptable::~BaseChangeScriptable ()
{
  RemoveFromLooper ();

  // Need to explicitly deallocate the UserData key strings.  The BMessage
  // portion of the key/value pair is automatically deallocated by the map.

  UserDataMap::iterator Iter;
  for (Iter = m_UserData.begin (); Iter != m_UserData.end (); Iter++)
  {
    free (const_cast<char *> (Iter->first));
  }
  m_UserData.clear ();
}


/* Adds this scriptable object (or rather its BHandler ancestor) to the given
BLooper (usually the main BApplication).  Needed for receiving scripting
messages, even if ResolveSpecifier explicitly returns this object as the
BHandler for a particular message. */

void BaseChangeScriptable::AddToLooper (BLooper *LooperPntr)
{
  if (Looper () == NULL && LooperPntr != NULL && LooperPntr->Lock ())
  {
    LooperPntr->AddHandler (this);
    LooperPntr->Unlock ();
  }
}


status_t BaseChangeScriptable::Archive (
  BMessage *ArchiveMessagePntr, bool Deep) const
{
  status_t ErrorCode = BHandler::Archive (ArchiveMessagePntr, Deep);
  if (ErrorCode != B_OK)
    return ErrorCode;

  UserDataMap::const_iterator UserDataIter; // Const since in const function.
  for (UserDataIter = m_UserData.begin ();
  UserDataIter != m_UserData.end (); UserDataIter++)
  {
    // Copy the UserData BMessage (which has a "result" field containing the
    // user's data) and add the key to it.  That combo then goes in the
    // archive.  Note that there can be several "result" fields (for an array
    // of user values) or none; they'll all get copied.

    BMessage TempMessage (B_SIMPLE_DATA);

    ErrorCode = TempMessage.AddString (g_NameKey, UserDataIter->first);
    if (ErrorCode != B_OK)
      return ErrorCode;

    ErrorCode = CopyBMessageField (const_cast<BMessage *> (
      &UserDataIter->second), &TempMessage, g_ResultName);
    if (ErrorCode != B_OK && ErrorCode != B_NAME_NOT_FOUND)
      return ErrorCode;

    ErrorCode = ArchiveMessagePntr->AddMessage (g_UserDataName, &TempMessage);
    if (ErrorCode != B_OK)
      return ErrorCode;
  }

  return B_OK;
}


void BaseChangeScriptable::ChangeHappened (bool SettingsAffected)
{
  m_ChangeSerial = GetNextSerialNumber ();
}


int BaseChangeScriptable::GetSerial ()
{
  return m_ChangeSerial;
}


status_t BaseChangeScriptable::GetSupportedSuites (BMessage *MessagePntr)
{
  BPropertyInfo TempPropInfo (g_UserDataScriptingPropertyList);
  MessagePntr->AddString ("suites", SUITE_BASE "userdata");
  MessagePntr->AddFlat ("messages", &TempPropInfo);
  return BHandler::GetSupportedSuites (MessagePntr);
}


void BaseChangeScriptable::MessageReceived (BMessage *MessagePntr)
{
  status_t              ErrorCode = B_ERROR;
  const char           *PropertyNamePntr;
  BMessage              ReplyMessage (B_REPLY);
  int32                 SpecifierIndex;
  int32                 SpecifierKind;
  BMessage              SpecifierMessage;
  char                  TempString [PATH_MAX + 1024];

  switch (MessagePntr->what)
  {
    case B_GET_PROPERTY:
    case B_SET_PROPERTY:
    case B_COUNT_PROPERTIES:
    case B_CREATE_PROPERTY:
    case B_DELETE_PROPERTY:
    case B_EXECUTE_PROPERTY:
      // Look at specifier[0] to find out what the scripting does.
      if (B_OK == MessagePntr->SetCurrentSpecifier (0))
        ErrorCode = MessagePntr->GetCurrentSpecifier (&SpecifierIndex,
        &SpecifierMessage, &SpecifierKind, &PropertyNamePntr);
      else // No specifiers, assume it is a direct operation on this UserData.
      {
        PropertyNamePntr = g_PropertyNames[PN_USER_DATA];
        SpecifierKind = B_DIRECT_SPECIFIER;
        ErrorCode = B_OK;
      }

      if (ErrorCode == B_OK &&
      IsItOurScriptingCommand (g_UserDataScriptingPropertyList,
      PropertyNamePntr, SpecifierKind))
        break; // It's one of ours, do it.
      // Else fall through, and punt the message to the parent class.

    default: // Not one of our messages, let the parent handle it.
      BHandler::MessageReceived (MessagePntr);
      return;
  }

  snprintf (TempString, sizeof (TempString),
    "Scripting operation code %c%c%c%c (get, set, count, etc) or access type "
    "%d (direct=1, index=2, reverse index=3, range=4, reverse range=5, "
    "name=6, ID=7) is unsupported for property \"%s\" of object \"%s\"",
    (char) (MessagePntr->what >> 24),
    (char) (MessagePntr->what >> 16),
    (char) (MessagePntr->what >>  8),
    (char) (MessagePntr->what >>  0),
    (int) SpecifierKind, PropertyNamePntr, Name ());
  ErrorCode = B_BAD_SCRIPT_SYNTAX;

  // Handle the UserData property.

  if (strcasecmp (PropertyNamePntr, g_PropertyNames[PN_USER_DATA]) == 0)
  {
    // Decode the specifier into a key for the UserData.

    const char *KeyPntr = NULL;

    if (SpecifierKind == B_NAME_SPECIFIER)
    {
      ErrorCode = SpecifierMessage.FindString (g_NameName, &KeyPntr);
      if (ErrorCode != B_OK || KeyPntr == NULL || KeyPntr[0] == 0)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString),
          "\"%s\" string field is missing or empty in the scripting specifier "
          "for the \"%s\" property of object \"%s\"",
          g_NameName, PropertyNamePntr, Name ());
        goto ErrorExit;
      }
    }
    else if (SpecifierKind == B_INDEX_SPECIFIER)
    {
      // Look for the numbered (indexed) UserData key, the slow way.

      int32 DesiredIndex = 0;
      ErrorCode = SpecifierMessage.FindInt32 (g_NameIndex, &DesiredIndex);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field is "
          "missing in the scripting specifier for the \"%s\" property of "
          "object \"%s\"",
          g_NameIndex, PropertyNamePntr, Name ());
        goto ErrorExit;
      }

      if (DesiredIndex < 0 || DesiredIndex >= (int) m_UserData.size ())
      {
        ErrorCode = B_BAD_INDEX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field for "
          "property \"%s\" is %d, which is not between 0 and %d, for "
          "object \"%s\"",
          g_NameIndex, PropertyNamePntr, (int) DesiredIndex,
          (int) m_UserData.size() - 1, Name ());
        goto ErrorExit;
      }

      UserDataMap::iterator UserDataIter = m_UserData.begin ();
      int32 i = 0;
      while (i < DesiredIndex && UserDataIter != m_UserData.end ())
      {
        UserDataIter++;
        i++;
      }
      if (UserDataIter == m_UserData.end ())
        goto ErrorExit; // Shouldn't happen.

      KeyPntr = UserDataIter->first;
    }

    if (MessagePntr->what == B_GET_PROPERTY)
    {
      if (KeyPntr == NULL)
        goto ErrorExit;

      UserDataMap::iterator UserDataIter = m_UserData.find (KeyPntr);
      if (UserDataIter == m_UserData.end ())
      {
        ErrorCode = B_NAME_NOT_FOUND;
        snprintf (TempString, sizeof (TempString), "\"%s\" key does not "
          "exist for UserData of object \"%s\"", KeyPntr, Name ());
        goto ErrorExit;
      }

      // Also include the key in the reply message so users can find the key
      // strings using the array index access method.

      ReplyMessage.AddString (g_NameKey, KeyPntr);

      // Finally copy over the UserData value, all the "result" fields in the
      // UserData BMessage.  Could be strings or other things.

      ErrorCode = CopyBMessageField (&UserDataIter->second,
        &ReplyMessage, g_ResultName);
      if (ErrorCode != B_OK && ErrorCode != B_NAME_NOT_FOUND)
      {
        snprintf (TempString, sizeof (TempString), "\"%s\" key copying data "
          "to output BMessage failed for object \"%s\"", KeyPntr, Name ());
        goto ErrorExit;
      }
      goto SuccessExit;
    }

    if (MessagePntr->what == B_SET_PROPERTY)
    {
      // Copy the "data" field(s) into the saved BMessage "result" field(s).
      // If the key doesn't exist, create a new key/BMessage.  May leave an
      // empty UserData BMessage if it fails (out of memory or no source data).

      UserDataSet (KeyPntr, MessagePntr);

      ChangeHappened ();
      goto SuccessExit;
    }

    // Count the number of UserData entries.

    if (MessagePntr->what == B_COUNT_PROPERTIES &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, m_UserData.size ());
      goto SuccessExit;
    }

    if (MessagePntr->what == B_DELETE_PROPERTY)
    {
      if (!UserDataDelete (KeyPntr))
      {
        ErrorCode = B_NAME_NOT_FOUND;
        snprintf (TempString, sizeof (TempString), "\"%s\" key does not "
          "exist for UserData of object \"%s\"", KeyPntr, Name ());
        goto ErrorExit;
      }

      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on UserData property.
  }

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return;

SuccessExit: /* Success. */
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
}


/* Removes this object from the BLooper it is assigned to.  Will no longer
receive scripting messages. */

void BaseChangeScriptable::RemoveFromLooper ()
{
  BLooper *MyLooperPntr = Looper ();
  if (MyLooperPntr != NULL && MyLooperPntr->Lock ())
  {
    MyLooperPntr->RemoveHandler (this);
    MyLooperPntr->Unlock ();
  }
}


BHandler *BaseChangeScriptable::ResolveSpecifier (BMessage *MessagePntr,
  int32 Index, BMessage *SpecifierMsgPntr, int32 SpecificationKind,
  const char *PropertyNamePntr)
{
#if 0
printf ("Resolving Specifier in BaseChangeScriptable object, Message is: ");
MessagePntr->PrintToStream();
printf ("Specifier index %d, Specifier kind %d, Property \"%s\"\n",
  (int) Index, (int) SpecificationKind, PropertyNamePntr);
#endif

  if (Index == 0 && IsItOurScriptingCommand (g_UserDataScriptingPropertyList,
  PropertyNamePntr, SpecificationKind))
  {
    return this; /* It's one of our properties, self will handle message. */
  }

  /* Handle an unrecognized scripting Field, let the parent figure it out. */

  return BHandler::ResolveSpecifier (
    MessagePntr, Index, SpecifierMsgPntr, SpecificationKind, PropertyNamePntr);
}


bool BaseChangeScriptable::UserDataAddString (
  const char *KeyPntr, const char *ValueStringPntr)
{
  if (KeyPntr == NULL || *KeyPntr == 0 || ValueStringPntr == NULL)
    return false; // Don't allow empty string keys.

  UserDataMap::iterator UserDataIter;
  BMessage *ValueMessagePntr = NULL;

  UserDataIter = m_UserData.find (KeyPntr);
  if (UserDataIter == m_UserData.end ())
  {
    // Need to make a new UserData entry, doesn't exist.

    KeyPntr = strdup (KeyPntr); // Allocate our own memory for string.
    if (KeyPntr == NULL)
      return false;

    ValueMessagePntr = &(m_UserData[KeyPntr]); // Allocates BMsg.
    ValueMessagePntr->what = B_SIMPLE_DATA;
  }
  else
    ValueMessagePntr = &(UserDataIter->second);

  ValueMessagePntr->AddString (g_ResultName, ValueStringPntr);
  return true;
}


bool BaseChangeScriptable::UserDataDelete (const char *KeyPntr)
{
  if (KeyPntr == NULL)
    return false;

  UserDataMap::iterator UserDataIter = m_UserData.find (KeyPntr);
  if (UserDataIter == m_UserData.end ())
    return false; // Key not found.

  char *OldKeyPntr = const_cast<char *> (UserDataIter->first);
  m_UserData.erase (UserDataIter);
  free (OldKeyPntr);

  return true;
}


const char * BaseChangeScriptable::UserDataGetString (
  const char *KeyPntr, const char *DefaultValue, int32 Index)
{
  if (DefaultValue == NULL)
    DefaultValue = g_EmptyString;

  if (KeyPntr == NULL)
    return DefaultValue;

  UserDataMap::iterator UserDataIter;

  UserDataIter = m_UserData.find (KeyPntr);
  if (UserDataIter == m_UserData.end ())
    return DefaultValue; // Key not found.

  BMessage *ValueMessagePntr = &(UserDataIter->second);
  const char *ReturnValue = NULL;
  if (B_OK != ValueMessagePntr->FindString (g_ResultName, Index,
  &ReturnValue) || ReturnValue == NULL)
    return DefaultValue;
  return ReturnValue;
}


// Copy the "data" field(s) or "result" ones (if no "data") into the saved
// BMessage "result" field(s), wiping out any previous contents.  If the key
// doesn't exist, create a new key/BMessage.

status_t BaseChangeScriptable::UserDataSet (
  const char *KeyPntr, BMessage *SourceMessagePntr)
{
  status_t ErrorCode;

  if (KeyPntr == NULL || *KeyPntr == 0 /* disallow empty string keys */)
      return B_BAD_VALUE;

  BMessage *ValueMessagePntr = NULL;
  UserDataMap::iterator UserDataIter = m_UserData.find (KeyPntr);
  if (UserDataIter == m_UserData.end ()) // New key needed.
  {
    // Need to explicitly allocate memory for the key string.

    char *NewKeyPntr = strdup (KeyPntr);
    if (NewKeyPntr == NULL)
      return B_NO_MEMORY;

    ValueMessagePntr = &(m_UserData[NewKeyPntr]);
    ValueMessagePntr->what = B_SIMPLE_DATA;
  }
  else // Key exists, reuse related BMessage structure.
    ValueMessagePntr = &(UserDataIter->second);

  ValueMessagePntr->MakeEmpty ();

  // Is there a "data" field in the passed in user data BMessage?

  int32 CountFound = 0;
  type_code TypeFound = 0;
  const char *NameOfFieldToCopy = g_DataName;
  ErrorCode = SourceMessagePntr->GetInfo (g_DataName, &TypeFound, &CountFound);
  if (ErrorCode != B_OK || CountFound <= 0)
    NameOfFieldToCopy = g_ResultName; // No "data", use "result" instead.

  // Copy the provided data into the UserData BMessage's "result" field.

  ErrorCode = CopyBMessageField (SourceMessagePntr, ValueMessagePntr,
    NameOfFieldToCopy, g_ResultName);
  if (ErrorCode != B_OK) // Field not found, or other errors, okayish.
    ValueMessagePntr->MakeEmpty ();

  return ErrorCode;
}


void BaseChangeScriptable::UserDataUpgradeFromNewerModel (BMessage *ModelPntr)
{
  status_t ErrorCode;
  int32 iUserData;

  for (iUserData = 0; true; iUserData++)
  {
    BMessage ModelDataMessage;
    ErrorCode = ModelPntr->FindMessage (
      g_UserDataName, iUserData, &ModelDataMessage);
    if (ErrorCode != B_OK)
      break; // Presumably ran out of UserData fields in the Model's BMessage.

    const char *KeyPntr = NULL;
    if (ModelDataMessage.FindString (g_NameKey, &KeyPntr) != B_OK)
      continue; // Bad entry, no key, skip it.

    // Don't trample over data created by the actual user when upgrading.

    static const char *KeysNotToBeUpgraded [] =
    {
      g_KeyDescription,
      g_KeyRenameGUIMode,
      g_KeyRenameGUIInsert,
      g_KeyRenameGUIRearrange,
      g_KeyRenameGUISimple,
      g_KeyRenameMonthConversion,
      g_KeyRenameSample,
      NULL
    };

    const char **SkipItPntr;
    for (SkipItPntr = KeysNotToBeUpgraded; *SkipItPntr != NULL; SkipItPntr++)
    {
      if (strcasecmp (*SkipItPntr, KeyPntr) == 0)
        break;
    }
    if (*SkipItPntr != NULL)
      continue; // Don't trample over this key and value.

    // Mostly for debugging, see if the string contents have changed.

    const char *OldString = UserDataGetString (KeyPntr);
    const char *NewString = g_EmptyString;
    if (ModelDataMessage.FindString (g_ResultName, &NewString) == B_OK)
    {
      if (strcmp (NewString, OldString) != 0)
      {
        printf ("Upgrading \"%s\" user data \"%s\"\n"
          "  from \"%s\"\n"
          "  to   \"%s\"\n",
          Name (), KeyPntr, OldString, NewString);
      }
    }
    else
      printf ("Upgrading \"%s\" non-string user data \"%s\".\n",
        Name (), KeyPntr);

    UserDataSet (KeyPntr, &ModelDataMessage);
  }
}



/******************************************************************************
 * Implementation of the Command class, constructor, destructor and the rest of
 * the member functions in mostly alphabetical order.
 */

Command::Command (const char *NewName) :
  BaseChangeScriptable (NewName),
  m_CommandAncestor (NULL),
  m_CommandParent (NULL),
  m_ChildProcessExitCode (0),
  m_ChildProcessID (-1),
  m_CommandLineSerial (-1),
  m_CommandState (CS_EDITING),
  m_Disabled (false),
  m_LogFileEnabled (false),
  m_Subsidiary (false),
  m_TriggerNamedCommandHasFinished (false),
  m_TriggerType (TRG_MANUAL)
{
  memset (&m_LastRunStartTime, 0, sizeof (m_LastRunStartTime));

  int i;
  for (i = 0; i < 2; i++)
  {
    m_PipeStdOutFromChild [i] = -1;
  }
}


Command::Command (BMessage *ArchiveMessagePntr) :
  BaseChangeScriptable (ArchiveMessagePntr),
  m_CommandAncestor (NULL),
  m_CommandParent (NULL),
  m_ChildProcessExitCode (0),
  m_ChildProcessID (-1),
  m_CommandLineSerial (-1),
  m_CommandState (CS_EDITING),
  m_Disabled (false),
  m_LogFileEnabled (false),
  m_Subsidiary (false),
  m_TriggerNamedCommandHasFinished (false),
  m_TriggerType (TRG_MANUAL)
{
  int32 CountFound = 0;
  status_t ErrorCode;
  type_code TypeFound = 0;

  memset (&m_LastRunStartTime, 0, sizeof (m_LastRunStartTime));

  int i;
  for (i = 0; i < 2; i++)
  {
    m_PipeStdOutFromChild [i] = -1;
  }

  ErrorCode = ArchiveMessagePntr->FindBool (g_DisabledName, &m_Disabled);
  if (ErrorCode != B_OK)
    m_Disabled = false;

  ErrorCode = ArchiveMessagePntr->FindBool (g_LogFileEnabledName,
  &m_LogFileEnabled);
  if (ErrorCode != B_OK)
    m_LogFileEnabled = false;

  ErrorCode = ArchiveMessagePntr->FindBool (g_SubsidiaryName, &m_Subsidiary);
  if (ErrorCode != B_OK)
    m_Subsidiary = false;

  ErrorCode = ArchiveMessagePntr->FindString (g_TemplateName, &m_Template);
  if (ErrorCode != B_OK)
    m_Template = g_EmptyString;

  // Read and process the trigger string.

  ErrorCode = ArchiveMessagePntr->FindString (g_TriggerName, &m_Trigger);
  if (ErrorCode != B_OK)
    m_Trigger = g_EmptyString;

  // Remove "@Disabled " prefix from old style trigger strings.  It's now a
  // separate flag.

  const char *OldStyleDisabled = "@Disabled";
  if (m_Trigger.ICompare (OldStyleDisabled, strlen (OldStyleDisabled)) == 0)
  {
    m_Disabled = true;
    m_Trigger.Remove (0, strlen (OldStyleDisabled));
    while (isspace (m_Trigger.ByteAt (0)))
      m_Trigger.Remove (0, 1); // Remove leading white space.
  }

  ParseTriggerString ();

  // Load all the Fields inside this Command.

  ErrorCode = ArchiveMessagePntr->
    GetInfo (g_FieldListName, &TypeFound, &CountFound);
  if (ErrorCode == B_OK && TypeFound == B_MESSAGE_TYPE && CountFound > 0)
  {
    int32 iField;
    for (iField = 0; iField < CountFound; iField++)
    {
      BMessage ArchiveMsg;
      ErrorCode = ArchiveMessagePntr->
        FindMessage (g_FieldListName, iField, &ArchiveMsg);
      if (ErrorCode != B_OK)
        continue; // Skip bad ones.
      BArchivable *DearchivedPntr = instantiate_object (&ArchiveMsg);
      if (DearchivedPntr == NULL)
        continue; // Unable to find C++ class of object to rebuild it, etc.

      Field *FieldPntr = cast_as (DearchivedPntr, Field);
      if (FieldPntr != NULL)
        InternalAddField (FieldPntr);
      else // Some other kind of object, someone's been hacking the settings?
        delete DearchivedPntr;
    }
  }

  // Load child subsidiary Commands.  They exist in Models and saved undo state
  // Commands.  They're not used in saved settings (children stored separately
  // each in their own Command entry).  Conceptually, the children are owned by
  // the Command while the Command isn't hooked into the global map of
  // Commands.  When it does get connected, the children will be added to the
  // global list (with renaming to give them unique names).

  ErrorCode = ArchiveMessagePntr->
    GetInfo (g_ChildListName, &TypeFound, &CountFound);
  if (ErrorCode == B_OK && TypeFound == B_MESSAGE_TYPE && CountFound > 0)
  {
    int32 iChild;
    for (iChild = 0; iChild < CountFound; iChild++)
    {
      BMessage ArchiveMsg;
      ErrorCode = ArchiveMessagePntr->
        FindMessage (g_ChildListName, iChild, &ArchiveMsg);
      if (ErrorCode != B_OK)
        continue; // Skip bad ones.
      BArchivable *DearchivedPntr = instantiate_object (&ArchiveMsg);
      if (DearchivedPntr == NULL)
        continue; // Unable to find C++ class of object to rebuild it, etc.

      Command *ChildPntr = cast_as (DearchivedPntr, Command);
      if (ChildPntr != NULL)
      {
        m_CommandChildren.push_back (ChildPntr);
        ChildPntr->m_CommandParent = this;
      }
      else // Some other kind of object, someone's been hacking the settings?
        delete DearchivedPntr;
    }
  }
}


Command::~Command ()
{
  // Delete all Fields in this Command.

  FieldMap::iterator FieldIter;
  for (FieldIter = m_FieldMap.begin ();
  FieldIter != m_FieldMap.end (); FieldIter++)
  {
    Field *FieldPntr = FieldIter->second;
    delete FieldPntr;
  }
  m_FieldMap.clear ();

  // Delete all subsidiary children.

  for (int i = 0; i < (int) m_CommandChildren.size (); i++)
  {
    Command *ChildPntr = m_CommandChildren [i];
    if (ChildPntr->IsSubsidiary ())
      delete ChildPntr;
    else
    {
      char TempString [1024];
      snprintf (TempString, sizeof (TempString),
        "Internal fault, while deleting Command \"%s\", found a "
        "non-subsidiary child \"%s\"", Name (), ChildPntr->Name ());
      DisplayErrorMessage (TempString, B_BAD_INDEX, "~Command");
    }
  }
  m_CommandChildren.clear ();
}


/* Reset various things used during ancestry determination.  Makes child list
empty, sets parent to NULL and clears the recursive flag. */

void Command::AncestryReset ()
{
  m_CommandAncestor = NULL;
  m_CommandParent = NULL;
  m_CommandChildren.clear ();
}


/* Update the pointer to the ancestor of this subsidiary Command.  Look for the
closest parent Command which is non-subsidiary.  If none then this is an
orphaned subsidiary Command which we will delink in preparation for deletion,
and return FALSE to get the caller to do the deletion.  Fix infinite loops due
to circular links between a bottom of tree subsidiary Command and its ancestor
(the ancestor is also a child of the Command).  Delete such links to keep the
subsidiary structure in tree form, for simplicity in the rest of the code. */

bool Command::AncestryUpdateAncestry (int MaximumDepth)
{
  int i;
  Command *AnAncestorPntr = m_CommandParent;
  for (i = 0; i < MaximumDepth; i++)
  {
    if (AnAncestorPntr == NULL)
      break; // Ran out of parents.

    if (!AnAncestorPntr->m_Subsidiary)
    { // Found a regular Command, will be our ancestor.
      m_CommandAncestor = AnAncestorPntr;

      // Verify that the ancestor isn't also a child of this command.

      for (i = 0; i < (int) m_CommandChildren.size (); i++)
      {
        if (m_CommandChildren[i] != m_CommandAncestor)
          continue;

        // Break the parent-of-ancestor circular link between the ancestor and
        // this Command.  It's still our ancestor, but not our child.

        char TempString [1024];
        snprintf (TempString, sizeof (TempString),
          "Removing Command \"%s\" as a parent of Command \"%s\" since it's "
          "also a descendant.  We don't handle subsidiary Command circular "
          "references, though non-subsidiary cicularity is fine",
          this->Name (), m_CommandAncestor->Name ());
        DisplayErrorMessage(TempString, B_BAD_INDEX, "AncestryUpdateAncestry");

        m_CommandChildren.erase (m_CommandChildren.begin () + i);
        m_CommandAncestor->m_CommandParent = NULL;
        m_CommandAncestor->m_Trigger = g_EmptyString;
        m_CommandAncestor->m_TriggerType = TRG_MANUAL;
        m_CommandAncestor->ChangeHappened ();
        break;
      }

      return true; // Yes, it's still our ancestor, but not our child any more.
    }
    AnAncestorPntr = AnAncestorPntr->m_CommandParent;
  }

  // No valid ancestor, will be deleted soon, disconnect related Commands to
  // save on having to do a full rebuild of the parent and child pointers.

  // Remove this Command from parent's child list.

  if (m_CommandParent != NULL)
  {
    for (i = 0; i < (int) m_CommandParent->m_CommandChildren.size (); i++)
    {
      if (m_CommandParent->m_CommandChildren[i] != this)
        continue;
      m_CommandParent->m_CommandChildren.erase (
        m_CommandParent->m_CommandChildren.begin () + i);
      i--;
    }
  }
  m_CommandParent = NULL;

  // Remove this Command as a parent of its children.

  for (i = 0; i < (int) m_CommandChildren.size (); i++)
    m_CommandChildren[i]->m_CommandParent = NULL;
  m_CommandChildren.clear ();

  return false;
}


/* Find parent from the trigger string and add self to parent's child list. */

void Command::AncestryUpdateParent ()
{
  if (m_TriggerType != TRG_NAMED_COMMAND)
    m_CommandParent = NULL;
  else
    m_CommandParent = g_App->InternalFindCommand (m_Trigger.String ());

  if (m_CommandParent != NULL)
    m_CommandParent->m_CommandChildren.push_back (this);
}


/* Archives the Command object into a BMessage, for storage (usually in the
settings file).  If Deep is specified, subsidiary Commands are included. */

status_t Command::Archive (BMessage *ArchiveMessagePntr, bool Deep) const
{
  status_t ErrorCode;

  ErrorCode = BaseChangeScriptable::Archive (ArchiveMessagePntr, Deep);
  if (ErrorCode != B_OK)
    return ErrorCode;

  // BHandler's Archive function adds the class name of "Command" too rather
  // than "BHandler", so this isn't really needed, at least in BeOS R5.03 and
  // Haiku OS.
/*
  ErrorCode = ArchiveMessagePntr->AddString ("class", "Command");
  if (ErrorCode != B_OK)
    return ErrorCode;
*/

  ErrorCode = ArchiveMessagePntr->AddBool (g_DisabledName, m_Disabled);
  if (ErrorCode != B_OK)
    return ErrorCode;

  ErrorCode = ArchiveMessagePntr->AddBool (g_LogFileEnabledName,
    m_LogFileEnabled);
  if (ErrorCode != B_OK)
    return ErrorCode;

  ErrorCode = ArchiveMessagePntr->AddBool (g_SubsidiaryName, m_Subsidiary);
  if (ErrorCode != B_OK)
    return ErrorCode;

  ErrorCode = ArchiveMessagePntr->AddString (g_TemplateName, m_Template);
  if (ErrorCode != B_OK)
    return ErrorCode;

  ErrorCode = ArchiveMessagePntr->AddString (g_TriggerName, m_Trigger);
  if (ErrorCode != B_OK)
    return ErrorCode;

#if 0 // Dump parent, children and ancestors names for debugging.
  if (m_CommandParent != NULL)
    ArchiveMessagePntr->AddString ("Parent", m_CommandParent->Name ());
  if (m_CommandAncestor != NULL)
    ArchiveMessagePntr->AddString ("Ancestor", m_CommandAncestor->Name ());
  for (int i = 0; i < (int) m_CommandChildren.size (); i++)
    ArchiveMessagePntr->AddString ("Child", m_CommandChildren[i]->Name ());
#endif

  // Fields always get archived, no situation where you wouldn't want them.

  FieldMap::const_iterator FieldIter; // Since this function is const too.
  for (FieldIter = m_FieldMap.begin ();
  FieldIter != m_FieldMap.end (); FieldIter++)
  {
    BMessage ArchiveMsg (B_ARCHIVED_OBJECT);

    Field *FieldPntr = FieldIter->second;
    ErrorCode = FieldPntr->Archive (&ArchiveMsg, Deep);
    if (ErrorCode != B_OK)
      return ErrorCode;
    ErrorCode =
      ArchiveMessagePntr->AddMessage (g_FieldListName, &ArchiveMsg);
    if (ErrorCode != B_OK)
      return ErrorCode;
  }

  // Recursively dump the subsidiary child subsidiary Commands if requested.

  if (Deep)
  {
    for (int i = 0; i < (int) m_CommandChildren.size (); i++)
    {
      Command *ChildPntr = m_CommandChildren [i];
      if (ChildPntr->m_Subsidiary)
      {
        BMessage ChildArchive (B_ARCHIVED_OBJECT);
        ErrorCode = ChildPntr->Archive (&ChildArchive, Deep);
        if (ErrorCode != B_OK)
          return ErrorCode;

        ErrorCode =
          ArchiveMessagePntr->AddMessage (g_ChildListName, &ChildArchive);
        if (ErrorCode != B_OK)
          return ErrorCode;
      }
    }
  }

  return ErrorCode;
}


/* Appends text to the log string, and to the log file too if it is open.
Negative length means figure out the length.  If this is a subsidiary Command,
the ancestor's log file and settings are used.  Does a ChangeHappened() to the
Command which is doing the logging. */

void Command::AppendLogText (const char *AddText, int32 AddLength)
{
  if (AddText == NULL || AddText[0] == 0 || AddLength == 0)
    return;

  Command *LogCommandPntr = this;
  BFile *LogFilePntr = &m_LogFile;
  BString *LogAccumulatedPntr = &m_LogAccumulated;
  if (IsSubsidiary () && m_CommandAncestor != NULL)
  {
    LogCommandPntr = m_CommandAncestor;
    LogFilePntr = &m_CommandAncestor->m_LogFile;
    LogAccumulatedPntr = &m_CommandAncestor->m_LogAccumulated;
  }

  if (AddLength < 0)
    AddLength = strlen (AddText);

  // Write to the log file.

  if (LogFilePntr->InitCheck () == B_OK)
  {
    ssize_t AmountWritten = LogFilePntr->Write (AddText, AddLength);
    if (AmountWritten != AddLength)
      LogFilePntr->Unset (); // Stop writing on error.
  }

  // Extra text past the max length of our in-memory buffer gets dropped.

  int32 OldLength = LogAccumulatedPntr->Length ();
  if (OldLength + AddLength >= MAX_LOG_LENGTH)
    AddLength = MAX_LOG_LENGTH - OldLength;
  if (AddLength > 0)
    LogAccumulatedPntr->Append (AddText, AddLength);

  LogCommandPntr->ChangeHappened (false /* Settings not changed, just log */);
}


/* This Command has been changed, increment the serial number and let the
parent objects know about the change. */

void Command::ChangeHappened (bool SettingsAffected)
{
  BaseChangeScriptable::ChangeHappened (SettingsAffected);

  if (m_CommandAncestor != NULL)
    m_CommandAncestor->ChangeHappened (SettingsAffected);
  else // Regular Command or we are the Ancestor, pass it up to the app.
    g_App->ChangeHappened (SettingsAffected);
}


/* Like the function name says... */

bool Command::DoesNamedFieldExist (const char *FieldNamePntr)
{
  FieldMap::iterator FieldIter;
  FieldIter = m_FieldMap.find (FieldNamePntr);
  return FieldIter != m_FieldMap.end ();
}


void Command::ExpandTemplateIntoCommandLine ()
{
  // Is the previously stored command line out of date?

  int CurrentSerialNumber = GetSerial ();
  if (m_CommandLineSerial != CurrentSerialNumber)
  {
    // Yes it is, generate command line by substituting keywords in the
    // Template text.  Each field does its own substitution.

    m_CommandLine = m_Template;

    FieldMap::iterator FieldIter;
    FieldIter = m_FieldMap.begin ();
    while (FieldIter != m_FieldMap.end ())
    {
      FieldIter->second->ExpandMagicWords (m_CommandLine);
      FieldIter++;
    }

    // SOCMyName - Script-O-Cron Command's true name.

    m_CommandLine.ReplaceAll ("SOCMyName", Name ());

    BString SOCName;
    BString SOCLogFile;
    BString SOCTemp;
    BString SOCWork;
    GetStockNamesAndPaths (&SOCName, &SOCLogFile, &SOCTemp, &SOCWork);

    // SOCName - Script-O-Cron Command's ancestor's name, useful for hiding
    // subsidiary Command mangled names from the end user.
    m_CommandLine.ReplaceAll (g_NameSOCName, SOCName.String ());

    // SOCWorkDir - semi-persistent working directory for the Command.  The
    // ancestor's name is used for subsidiary Commands so they all work from
    // the same directory.
    m_CommandLine.ReplaceAll (g_NameSOCWorkDir, SOCWork.String ());

    // SOCLogFile - path to the log file for the Command and subsitiaries.
    m_CommandLine.ReplaceAll (g_NameSOCLogFile, SOCLogFile.String ());

    // SOCTempDir - temporary stuff directory for the Command.
    m_CommandLine.ReplaceAll (g_NameSOCTempDir, SOCTemp.String ());

    // Done, mark it so!
    m_CommandLineSerial = CurrentSerialNumber;
  }
}


void Command::GetSanitisedName (BString &OutputName)
{
  OutputName = Name ();

  // Look for names which are all periods, like "", ".", "..", "..." (for
  // AGMSRamFileSystem compatibility).  Prefix them with "#".

  int i;
  for (i = 0; i < OutputName.Length (); i++)
  {
    if (OutputName.ByteAt (i) != '.')
      break;
  }
  if (i >= OutputName.Length ())
    OutputName.Prepend ("#");

  OutputName.ReplaceSet ("/\"$!'", '-'); // No slashes and so on in file names.
}


/* Returns the path and other names for some stock things related to the
Command.  If NULL is specified, that particular name won't be calculated. */

void Command::GetStockNamesAndPaths (BString *OutputName,
  BString *OutputLogFile, BString *OutputTempDir, BString *OutputWorkDir)
{
  Command *CommandToUseForNames = this;
  if (m_CommandAncestor != NULL)
    CommandToUseForNames = m_CommandAncestor;

  BString SanitisedName;
  CommandToUseForNames->GetSanitisedName (SanitisedName);

  // SOCName - Script-O-Cron Command's ancestor's name, useful for hiding
  // subsidiary Command mangled names from the end user.

  if (OutputName != NULL)
    *OutputName = CommandToUseForNames->Name ();

  // SOCLogFile - path to the log file for the Command and subsidiaries.

  if (OutputLogFile != NULL)
  {
    BPath LogPath (g_App->GetSettingsDirectory (), g_LogDirName);
    LogPath.Append (SanitisedName.String ());
    *OutputLogFile = LogPath.Path ();
  }

  // SOCTempDir - temporary stuff directory for the Command.

  if (OutputTempDir != NULL)
  {
    status_t ErrorCode = 0;
    BPath TempPath;
    const char *UserTempDir = g_App->GetUserSpecifiedTemporaryDirectory ();
    if (UserTempDir [0] == 0)
    {
#if __HAIKU__
      ErrorCode = find_directory (B_SYSTEM_TEMP_DIRECTORY,
        &TempPath, false /* Create if needed */);
#else
      ErrorCode = find_directory (B_COMMON_TEMP_DIRECTORY,
        &TempPath, false /* Create if needed */);
#endif
    }
    else // User wants a non-default temporary directory.
    {
      ErrorCode = TempPath.SetTo (UserTempDir);
    }

    TempPath.Append (PROGRAM_NAME_WORD);
    TempPath.Append (SanitisedName.String ());
    *OutputTempDir = TempPath.Path ();
  }

  // SOCWorkDir - semi-persistent working directory for the Command.  The
  // ancestor's name is used for subsidiary Commands so they all work from
  // the same directory.

  if (OutputWorkDir != NULL)
  {
    BPath WorkPath (g_App->GetSettingsDirectory (), "Work");
    WorkPath.Append (SanitisedName.String ());
    *OutputWorkDir = WorkPath.Path ();
  }
}


/* Returns a list of the scripting commands a Command understands. */

status_t Command::GetSupportedSuites (BMessage *MessagePntr)
{
  BPropertyInfo TempPropInfo (g_CommandScriptingPropertyList);
  MessagePntr->AddString ("suites", SUITE_BASE "command");
  MessagePntr->AddFlat ("messages", &TempPropInfo);
  return BaseChangeScriptable::GetSupportedSuites (MessagePntr);
}


/* Class global function for recreating objects from a BMessage archive. */

BArchivable *Command::Instantiate (BMessage *ArchiveMessagePntr)
{
  if (validate_instantiation (ArchiveMessagePntr, "Command"))
    return new Command (ArchiveMessagePntr);
  return NULL;
}


/* Add all children (as specified by child pointers, not Trigger names) to the
global list of Commands.  Can assume it is a tree, without circular references.
Conceptually part of the recursive InternalAddCommandGuts. */

void Command::InternalAddChildren (const char *AncestorName, int Depth)
{
  int i;
  for (i = 0; i < (int) m_CommandChildren.size (); i++)
  {
    Command *ChildPntr = m_CommandChildren [i];
    g_App->InternalAddCommandGuts (ChildPntr, AncestorName, Depth);
  }
}


/* Add the given Field to the Map of all Fields.  Used by the Field object when
it is renaming itself.  Also used when adding a newly dearchived Field.
Replaces any existing Field with the same name.  Also updates the BLooper and
parent pointer. */

void Command::InternalAddField (Field *FieldPntr)
{
  FieldMap::iterator FieldIter;
  FieldIter = m_FieldMap.find (FieldPntr->Name ());
  if (FieldIter != m_FieldMap.end ())
  {
    Field *OldFieldPntr = FieldIter->second;

    if (OldFieldPntr == FieldPntr)
      return; // It's already added to our Map of Fields.  Do nothing.

    m_FieldMap.erase (FieldIter);
    delete OldFieldPntr; // Will deLooper it as part of destructor.
  }

  m_FieldMap[FieldPntr->Name ()] = FieldPntr;
  FieldPntr->AddToLooper (g_App);
  FieldPntr->SetOwnerCommand (this);
}


/* Remove all subsidiary Command children from the global map of Commands,
leaving them linked together by the parent and children pointers.  Child
regular Commands are not removed, since they're not conceptually controlled by
the ancestor Command that started the removal.  Conceptually part of the
recursive InternalRemoveCommandGuts. */

void Command::InternalRemoveChildren ()
{
  int i;
  for (i = 0; i < (int) m_CommandChildren.size (); i++)
  {
    Command *ChildPntr = m_CommandChildren [i];
    if (!ChildPntr->IsSubsidiary ())
    {
      // Break the link to this child.  It's a regular Command that's not being
      // removed.

      ChildPntr->m_CommandParent = NULL;
      m_CommandChildren.erase (m_CommandChildren.begin () + i);
      i--;
      continue;
    }

    // Have a subsidiary child here.

    g_App->InternalRemoveCommandGuts (ChildPntr);
  }
}


/* Remove the given Field from the Map of all Fields (but don't delete it).
Used by the Field object when it is renaming itself.  Also updates the BLooper
and parent pointer. */

void Command::InternalRemoveField (Field *FieldPntr)
{
  FieldMap::iterator FieldIter;
  FieldIter = m_FieldMap.find (FieldPntr->Name ());
  if (FieldIter != m_FieldMap.end ())
  {
    Field *OldFieldPntr = FieldIter->second;
    if (OldFieldPntr == FieldPntr)
    {
      m_FieldMap.erase (FieldIter);
      OldFieldPntr->RemoveFromLooper ();
      OldFieldPntr->SetOwnerCommand (NULL);
    }
  }
}


/* Process mostly scripting messages. */

void Command::MessageReceived (BMessage *MessagePntr)
{
  const char           *ArgumentString = NULL;
  status_t              ErrorCode = B_ERROR;
  const char           *PropertyName;
  BMessage              ReplyMessage (B_REPLY);
  int32                 SpecifierIndex;
  int32                 SpecifierKind;
  BMessage              SpecifierMessage;
  char                  TempString [PATH_MAX + 1024];

  switch (MessagePntr->what)
  {
    case B_GET_PROPERTY:
    case B_SET_PROPERTY:
    case B_COUNT_PROPERTIES:
    case B_CREATE_PROPERTY:
    case B_DELETE_PROPERTY:
    case B_EXECUTE_PROPERTY:
      // Look at specifier[0] to find out what the scripting does.
      if (B_OK == MessagePntr->SetCurrentSpecifier (0))
        ErrorCode = MessagePntr->GetCurrentSpecifier (&SpecifierIndex,
        &SpecifierMessage, &SpecifierKind, &PropertyName);
      else // No specifiers, assume it is a direct operation on this Command.
      {
        PropertyName = g_PropertyNames[PN_COMMAND];
        SpecifierKind = B_DIRECT_SPECIFIER;
        ErrorCode = B_OK;
      }

      if (ErrorCode == B_OK &&
      IsItOurScriptingCommand (g_CommandScriptingPropertyList,
      PropertyName, SpecifierKind))
        break; // It's one of ours, do it.
      // Else fall through, and punt the message to the parent class.

    default: // Not one of our messages, let the parent handle it.
      BaseChangeScriptable::MessageReceived (MessagePntr);
      return;
  }

  snprintf (TempString, sizeof (TempString),
    "Scripting operation code %c%c%c%c (get, set, count, etc) or access type "
    "%d (direct=1, index=2, reverse index=3, range=4, reverse range=5, "
    "name=6, ID=7) is unsupported for property \"%s\" of the Command "
    "object named \"%s\".",
    (char) (MessagePntr->what >> 24),
    (char) (MessagePntr->what >> 16),
    (char) (MessagePntr->what >>  8),
    (char) (MessagePntr->what >>  0),
    (int) SpecifierKind, PropertyName, Name ());
  ErrorCode = B_BAD_SCRIPT_SYNTAX;

  // Handle the Name property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_NAME]) == 0)
  {
    // Get the current name.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, Name ());
      goto SuccessExit;
    }

    // Set the new Name.  This also means changing the global Map of Commands
    // since they are in alphabetical order.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode == B_OK)
      {
        if (IsSubsidiary ())
        {
          snprintf (TempString, sizeof (TempString),
            "Can't change name of Command \"%s\" because it is a subsidiary.  "
            "Instead, make it non-subsidiary or change the ancestor \"%s\"",
            Name (),
            (m_CommandAncestor == NULL) ? "?" : m_CommandAncestor->Name ());
          ErrorCode = B_BUSY;
          goto ErrorExit;
        }

        if (g_App->DoesNamedCommandExist (ArgumentString))
        {
          snprintf (TempString, sizeof (TempString),
            "Command named \"%s\" already exists, can't rename \"%s\" to it",
            ArgumentString, Name ());
          ErrorCode = B_NAME_IN_USE;
          goto ErrorExit;
        }

        if (ArgumentString [0] == 0)
        {
          snprintf (TempString, sizeof (TempString),
            "Command named \"%s\" can't be renamed to \"%s\" because the new "
            "name is empty and that would be confusing",
            Name (), ArgumentString);
          ErrorCode = B_BAD_VALUE;
          goto ErrorExit;
        }

        // Does it contain slashes?  Back or forwards both cause problems when
        // we try to use the name as a directory name (such as for the log
        // file, or work directory, etc).

        if (strchr (ArgumentString, '/') != NULL ||
        strchr (ArgumentString, '\\') != NULL)
        {
          snprintf (TempString, sizeof (TempString),
            "Command named \"%s\" can't be renamed to \"%s\" because the new "
            "name contains slashes which will break things when we try to "
            "use the name as a directory name or file name (such as for the "
            "log file for this Command)",
            Name (), ArgumentString);
          ErrorCode = B_BAD_VALUE;
          goto ErrorExit;
        }

        // Make sure the new name isn't all periods.

        bool AllPeriods = true;
        const char *SourcePntr = ArgumentString;
        while (*SourcePntr != 0)
        {
          if (*SourcePntr != '.')
          {
            AllPeriods = false;
            break;
          }
          SourcePntr++;
        }
        if (AllPeriods)
        {
          snprintf (TempString, sizeof (TempString),
            "Command named \"%s\" can't be renamed to \"%s\" because the new "
            "name is entirely periods, which are reserved for magic directory "
            "names, like the parent directory, alternate parent directory, "
            "etc. (see AGMSLink explanation)",
            Name (), ArgumentString);
          ErrorCode = B_BAD_VALUE;
          goto ErrorExit;
        }

        // New Command name is good, prepare to rename related files.

        BString OldLogPathName;
        BString OldWorkPathName;
        GetStockNamesAndPaths (NULL, &OldLogPathName, NULL, &OldWorkPathName);

        // Change the actual name.

        g_App->InternalRemoveCommand (this);
        SetName (ArgumentString);
        g_App->InternalAddCommand (this);
        ChangeHappened ();

        // Rename the log file and work directory.  Not an error if they don't
        // exist.  Do both even if one fails.

        BString NewLogPathName;
        BString NewWorkPathName;
        GetStockNamesAndPaths (NULL, &NewLogPathName, NULL, &NewWorkPathName);

        ErrorCode = B_OK;
        if (rename (OldLogPathName.String (), NewLogPathName.String ()))
        {
          if (errno != ENOENT)
          {
            ErrorCode = errno;
            snprintf (TempString, sizeof (TempString),
              "Unable to rename log file for Command \"%s\" from "
              "\"%s\" to \"%s\"",
              Name (), OldLogPathName.String (), NewLogPathName.String ());
          }
        }

        if (rename (OldWorkPathName.String(), NewWorkPathName.String()))
        {
          if (errno != ENOENT)
          {
            ErrorCode = errno;
            snprintf (TempString, sizeof (TempString),
              "Unable to rename work directory for Command \"%s\" from "
              "\"%s\" to \"%s\"",
              Name (), OldWorkPathName.String(), NewWorkPathName.String());
          }
        }

        if (ErrorCode != B_OK)
          goto ErrorExit;
        goto SuccessExit;
      }
      else
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Name property "
          "from.  Was for Command \"%s\"",
          g_DataName, Name ());
        goto ErrorExit;
      }
    }

    goto ErrorExit; // Unknown operation on Name property.
  }

  // Handle the Serial property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_SERIAL]) == 0)
  {
    // Get the current serial number.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, GetSerial ());
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Serial property.
  }

  // Handle the Command property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_COMMAND]) == 0)
  {
    if (MessagePntr->what == B_DELETE_PROPERTY /* any kind of specifier OK */)
    {
      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't delete Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      // Delete the corresponding log file and work directory, by moving it to
      // trash.  Confusingly they both have the same name, but the trasher
      // appends sequence numbers if needed, so this works.  Okay to fail.

      BString LogPathName;
      BString WorkPathName;
      GetStockNamesAndPaths (NULL, &LogPathName, NULL, &WorkPathName);
      ErrorCode = trash (LogPathName.String ());
      ErrorCode = trash (WorkPathName.String ());

      g_App->InternalRemoveCommand (this); // DeLoopers this too.
      delete this; // Careful, can't reference member variables after here.
      g_App->ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Command property.
  }

  // Handle the Duplicate property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_DUPLICATE]) == 0)
  {
    if (MessagePntr->what == B_CREATE_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK)
        ArgumentString = g_NameUnnamed;

      if (g_App->DoesNamedCommandExist (ArgumentString))
      {
        ErrorCode = B_NAME_IN_USE;
        snprintf (TempString, sizeof (TempString),
          "Can't create a duplicate Command named \"%s\" because it "
          "already exists", ArgumentString);
        goto ErrorExit;
      }

      // Make a copy of the command in BMessage format, including subsidiary
      // Commands.

      BMessage MsgCopyOfCommand (B_ARCHIVED_OBJECT);
      ErrorCode = Archive (&MsgCopyOfCommand, true /* deep */);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Pre-duplication failed to archive Command \"%s\" into a BMessage",
          Name ());
        goto ErrorExit;
      }

      BArchivable *DearchivedPntr = instantiate_object (&MsgCopyOfCommand);
      Command *DuplicateCommandPntr = NULL;
      if (DearchivedPntr == NULL ||
      (DuplicateCommandPntr = cast_as (DearchivedPntr, Command)) == NULL)
      {
        delete DuplicateCommandPntr; // NULL or wrong kind of object.
        ErrorCode = B_ERROR;
        snprintf (TempString, sizeof (TempString),
          "Failed to instantiate a duplicate of Command \"%s\"", Name ());
        goto ErrorExit;
      }

      DuplicateCommandPntr->SetName (ArgumentString);
      g_App->InternalAddCommand (DuplicateCommandPntr);
      DuplicateCommandPntr->SetStateRecursive (CS_EDITING);

      DuplicateCommandPntr->ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Duplicate property.
  }

  // Handle the Template property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_TEMPLATE]) == 0)
  {
    // Get the current Template string.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, m_Template);
      goto SuccessExit;
    }

    // Set the Template string, but only in Edit mode.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Template property "
          "from.  Was for Command \"%s\"",
          g_DataName, Name ());
        goto ErrorExit;
      }

      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't set Template of Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      m_Template = ArgumentString;

      // Delete the "Original Model" UserData entry now that the user has
      // edited the template.

      UserDataDelete (g_KeyOriginalModel);

      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Template property.
  }

  // Handle the Trigger property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_TRIGGER]) == 0)
  {
    // Get the current Trigger string.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, m_Trigger);
      goto SuccessExit;
    }

    // Set the Trigger string, but only in Edit mode.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Trigger property "
          "from.  Was for Command \"%s\"",
          g_DataName, Name ());
        goto ErrorExit;
      }

      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't set Trigger of Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      m_Trigger = ArgumentString;
      ParseTriggerString ();

      ChangeHappened (); // Let old ancestor know we changed.
      BString MyName (Name ()); // Save my name, in case of deletion.
      g_App->RecalculateAncestry (); // May delete this Command!
      if (g_App->DoesNamedCommandExist (MyName.String ()))
        ChangeHappened (false); // Let new ancestor know we changed.

      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Trigger property.
  }

  // Handle the Children property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_CHILDREN]) == 0)
  {
    // Get the list of children, and some relatives.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      bool IncludeRegular = true;
      bool IncludeSubsidiary = true;
      bool TempBool = false;
      if (MessagePntr->FindBool ("subsidiary", &TempBool) == B_OK)
      {
        if (TempBool)
          IncludeRegular = false;
        else
          IncludeSubsidiary = false;
      }

      for (int i = 0; i < (int) m_CommandChildren.size (); i++)
      {
        Command *ChildPntr = m_CommandChildren [i];
        if (ChildPntr->IsSubsidiary ())
        {
          if (IncludeSubsidiary)
            ReplyMessage.AddString (g_ResultName, ChildPntr->Name ());
        }
        else // Regular Command.
        {
          if (IncludeRegular)
            ReplyMessage.AddString (g_ResultName, ChildPntr->Name ());
        }
      }

      // Add the ancestor's name. Only exists in subsidiary Commands.

      if (m_CommandAncestor != NULL)
        ReplyMessage.AddString (g_AncestorName, m_CommandAncestor->Name ());

      // Add the parent's name, if there is a parent.

      if (m_CommandParent != NULL)
        ReplyMessage.AddString (g_ParentName, m_CommandParent->Name ());

      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Children property.
  }

  // Handle the CommandLine property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_COMMAND_LINE]) == 0)
  {
    // Get the current command line string.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ExpandTemplateIntoCommandLine ();
      ReplyMessage.AddString (g_ResultName, m_CommandLine);
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on CommandLine property.
  }

  // Handle the Edit property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_EDIT]) == 0)
  {
    // Get editing state.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddBool (g_ResultName, m_CommandState == CS_EDITING);
      goto SuccessExit;
    }

    if (IsSubsidiary ())
    {
      snprintf (TempString, sizeof (TempString),
        "Can't change edit state of Command named \"%s\" because it is a "
        "subsidiary.  Edit the ancestor \"%s\" instead or make it "
        "non-subsidiary", Name (),
        (m_CommandAncestor == NULL) ? "?" : m_CommandAncestor->Name ());
      ErrorCode = B_BUSY;
      goto ErrorExit;
    }

    // Start editing.

    if (MessagePntr->what == B_CREATE_PROPERTY) // All kinds of specifiers OK.
    {
      if (m_CommandState != CS_READY_TO_RUN)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't edit Command named \"%s\" because it is busy %s",
          Name (), g_CommandStateNames [m_CommandState]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      // Make a copy of the current state of the command, including contained
      // fields.

      m_CopyOfCommandBeforeEditing.MakeEmpty ();
      m_CopyOfCommandBeforeEditing.what = B_ARCHIVED_OBJECT;
      ErrorCode = Archive (&m_CopyOfCommandBeforeEditing, true /* deep */);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Failed to make a backup copy of Command named \"%s\" before "
          "starting editing.  Editing operation aborted", Name ());
        goto ErrorExit;
      }

      SetStateRecursive (CS_EDITING); // Also does a ChangeHappened (false).
      goto SuccessExit;
    }

    // Cancel editing.

    if (MessagePntr->what == B_DELETE_PROPERTY) // All kinds of specifiers OK.
    {
      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't cancel editing of Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      // Restore the orginal version of the command by dearchiving the copy,
      // deleting the old Command and inserting the newly dearchived copy.  If
      // no previous copy exists (such as from totally new Commands which start
      // in Edit state), the Command disappears silently.

      BArchivable *DearchivedPntr =
        instantiate_object (&m_CopyOfCommandBeforeEditing);
      Command *OriginalCommandPntr = NULL;
      if (DearchivedPntr == NULL ||
      (OriginalCommandPntr = cast_as (DearchivedPntr, Command)) == NULL)
        delete DearchivedPntr; // NULL or wrong kind of object.

      g_App->InternalRemoveCommand (this);
      if (OriginalCommandPntr != NULL)
      {
        g_App->InternalAddCommand (OriginalCommandPntr);
        OriginalCommandPntr->SetStateRecursive (CS_READY_TO_RUN);
        OriginalCommandPntr->ChangeHappened ();
      }
      else // At least let parent know we're going away.
        g_App->ChangeHappened ();
      delete this; // No more references to member variables after this!
      goto SuccessExit;
    }

    // Finish editing (Execute Edit).

    if (MessagePntr->what == B_EXECUTE_PROPERTY) // All kinds of specifiers OK.
    {
      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't finish editing of Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      // Delete the backup copy of the command as it was before editing.

      m_CopyOfCommandBeforeEditing.MakeEmpty ();
      m_CopyOfCommandBeforeEditing.what = 0;
      SetStateRecursive (CS_READY_TO_RUN); // Also does ChangeHappened (false).
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Edit property.
  }

  // Handle the Run property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_RUN]) == 0)
  {
    // Start running (Execute Run).

    if (MessagePntr->what == B_EXECUTE_PROPERTY) // All kinds of specifiers OK.
    {
      if (g_QuitCountdown >= 0)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't run Command named \"%s\" because " PROGRAM_NAME_WORD
          " is in the process of shutting down", Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      if (m_CommandState != CS_READY_TO_RUN)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't run Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_READY_TO_RUN]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      ErrorCode = RunCommand (TempString, sizeof (TempString));
      if (ErrorCode != B_OK)
        goto ErrorExit;

      snprintf (TempString, sizeof (TempString),
        "\n%s \"%s\" started on %s.\n",
        IsSubsidiary () ? "Subcommand" :
        "====================================================================="
        "===========\n\nCommand",
        Name (), CurrentTimeAsString ());
      AppendLogText (TempString);
      AppendLogText (m_CommandLine.String ()); // Can be really long.
      AppendLogText ("\n");

      // Can't use SetStateRecursive (CS_RUNNING), since only this Command is
      // running, the child subsidiary Commands will start on their own after
      // this Command finishes running.
      m_CommandState = CS_RUNNING;

      ChangeHappened (false /* Settings file doesn't need saving */);
      goto SuccessExit;
    }

    // Stop running, abort, abort!

    if (MessagePntr->what == B_DELETE_PROPERTY) // All kinds of specifiers OK.
    {
      if (!IsRunningOrWaiting ())
      {
        snprintf (TempString, sizeof (TempString),
          "Can't stop Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_RUNNING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      ErrorCode = StopCommand (TempString, sizeof (TempString));
      if (ErrorCode != B_OK)
        goto ErrorExit;

      goto SuccessExit;
    }

    // Get the current run state.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddBool (g_ResultName,
        m_CommandState == CS_RUNNING || m_CommandState == CS_WAIT_SUBSIDIARY);
      ReplyMessage.AddBool (g_WaitingName,
        m_CommandState == CS_WAIT_SUBSIDIARY);
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Run property.
  }

  // Handle the Log property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_LOG]) == 0)
  {
    // Get the accumulated log output.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, m_LogAccumulated);
      m_LogAccumulated = g_EmptyString;
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Log property.
  }

  // Handle the LogFile property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_LOG_FILE_ENABLE]) == 0)
  {
    // Get the current logging enabled status.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddBool (g_ResultName, m_LogFileEnabled);
      goto SuccessExit;
    }

    // Set the logging enabled status.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      bool TempBool = false;
      ErrorCode = MessagePntr->FindBool (g_DataName, &TempBool);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" boolean value to set the \"%s\" "
          "property.  Was for Command \"%s\"",
          g_DataName, PropertyName, Name ());
        goto ErrorExit;
      }

      m_LogFileEnabled = TempBool;
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on LogFile property.
  }

  // Handle the Paths property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_PATHS]) == 0)
  {
    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      BString SOCName;
      BString SOCLogFile;
      BString SOCTemp;
      BString SOCWork;
      GetStockNamesAndPaths (&SOCName, &SOCLogFile, &SOCTemp, &SOCWork);

      ReplyMessage.AddString (g_NameSOCName, SOCName);
      ReplyMessage.AddString (g_NameSOCLogFile, SOCLogFile);
      ReplyMessage.AddString (g_NameSOCTempDir, SOCTemp);
      ReplyMessage.AddString (g_NameSOCWorkDir, SOCWork);
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Paths property.
  }

  // Handle the Disable property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_DISABLE]) == 0)
  {
    // Get the current disable status.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddBool (g_ResultName, m_Disabled);
      goto SuccessExit;
    }

    // Set the disabled state to a new value.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      bool TempBool = false;
      ErrorCode = MessagePntr->FindBool (g_DataName, &TempBool);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" boolean value to set the \"%s\" "
          "property.  Was for Command \"%s\"",
          g_DataName, PropertyName, Name ());
        goto ErrorExit;
      }

      m_Disabled = TempBool;
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Disable property.
  }

  // Handle the Subsidiary property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_SUBSIDIARY]) == 0)
  {
    // Get the current Subsidiary status.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddBool (g_ResultName, m_Subsidiary);
      goto SuccessExit;
    }

    // Set the Subsidiary status.  Will check during loading and other ancestor
    // tree changing operations for validity (needs a parent Command).

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      bool TempBool = false;
      ErrorCode = MessagePntr->FindBool (g_DataName, &TempBool);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" boolean value to set the \"%s\" "
          "property.  Was for Command \"%s\"",
          g_DataName, PropertyName, Name ());
        goto ErrorExit;
      }

      m_Subsidiary = TempBool;

      ChangeHappened (); // Let old ancestor know we are changing.
      BString MyName (Name ()); // Save my name, in case of deletion.
      g_App->RecalculateAncestry (); // May delete this Command!
      if (g_App->DoesNamedCommandExist (MyName.String ()))
        ChangeHappened (false);  // Let new ancestor know about the change.

      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Subsidiary property.
  }

  // Handle the Field property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_FIELD]) == 0)
  {
    // Count the number of Fields.

    if (MessagePntr->what == B_COUNT_PROPERTIES &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, m_FieldMap.size ());
      goto SuccessExit;
    }

    // Create a new Field, optionally with a name.

    if (MessagePntr->what == B_CREATE_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK)
        ArgumentString = g_NameUnnamed;

      if (!IsEditing ())
      {
        snprintf (TempString, sizeof (TempString),
          "Can't create a Field named \"%s\" because the related "
          "Command named \"%s\" isn't in edit mode",
          ArgumentString, Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      if (DoesNamedFieldExist (ArgumentString))
      {
        snprintf (TempString, sizeof (TempString),
          "Can't create a Field named \"%s\" because it already exists "
          "inside Command \"%s\"",
          ArgumentString, Name ());
        ErrorCode = B_NAME_IN_USE;
        goto ErrorExit;
      }

      Field *NewFieldPntr = new Field (ArgumentString);
      InternalAddField (NewFieldPntr);
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on a Field property.
  }

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return;

SuccessExit: /* Success. */
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
}


/* The parent Command has finished running.  If we were waiting for it to
finish, then get ready to start running this Command on the next poll. */

void Command::ParentCommandFinished ()
{
  if (m_CommandState != CS_READY_TO_RUN)
    return;

  m_TriggerNamedCommandHasFinished = true;
}


/* Examine the m_Trigger string to see what kind it is, and update related
variables with decoded values. */

void Command::ParseTriggerString ()
{
  if (m_Trigger.Length () == 0)
    m_TriggerType = TRG_MANUAL;
  else if (strcasecmp (m_Trigger.String (), g_NameReboot) == 0)
    m_TriggerType = TRG_REBOOT;
  else if (m_TriggerCronBits.FromString (m_Trigger.String ()) == NULL)
    m_TriggerType = TRG_CRON;
  else
    m_TriggerType = TRG_NAMED_COMMAND;
}


/* Do periodic polling operations for the command.  If it's running, see if
there is any output text and log it.  Also see if it has finished running and
get the final status code.  If it's ready to run, see if it wants to start
running.  If it's waiting for children, just sit around with the log file open
until they've stopped for a couple of polling iterations. */

void Command::PollAndUpdate (const struct tm *CurrentTime)
{
  if (m_CommandState == CS_RUNNING)
  {
    // Try reading the pipe.  If successful, return and don't bother checking
    // to see if the child program is dead (want to empty out the pipe first).

    status_t ErrorCode = B_OK;

    if (m_PipeStdOutFromChild [0] != -1)
    {
      char ReadBuffer [1024 * 64]; // BeOS gives us 4K, Haiku 64K at most.
      ssize_t AmountRead = -1;

      AmountRead = read (m_PipeStdOutFromChild [0], ReadBuffer,
        sizeof (ReadBuffer));
      if (AmountRead < 0)
      {
        ErrorCode = errno;
        if (ErrorCode == EAGAIN)
          return; // No new data available yet, a normal situation.
        // Read error of some sort, fall through and see if the program is
        // dead.
      }
      else if (AmountRead == 0)
      {
        // End of file, means the other end of the pipe has been closed.  Fall
        // through and see if the child program is dead.
      }
      else // AmountRead > 0
      {
        AppendLogText (ReadBuffer, AmountRead);
        return;
      }
    }

    // No more data is readable, close the pipe.

    if (m_PipeStdOutFromChild [0] != -1)
    {
      close (m_PipeStdOutFromChild [0]);
      m_PipeStdOutFromChild [0] = -1;
    }

    // See if the child program has exited.  Retrieve the return code if so.
    // This also "reaps" a zombie child (final step in the process life cycle).
    // The error message will be printed out after all child Commands have
    // finished, so that it will be the last entry in the log.

    m_ChildProcessExitCode = -1;
    pid_t ProcessIDAffected =
      waitpid (m_ChildProcessID, &m_ChildProcessExitCode, WNOHANG);
    if (ProcessIDAffected == 0)
      return; // Child hasn't stopped.
    m_ChildProcessID = -1;

    // Notify child Commands (subsidiary and regular ones too) that this one
    // has finished successfully.  Leave the log file open for subsidiary
    // Commands to write to, while we wait for them to finish running.

    if (m_ChildProcessExitCode == 0)
    {
      for (int i = 0; i < (int) m_CommandChildren.size (); i++)
      {
        Command *ChildPntr = m_CommandChildren [i];
        ChildPntr->ParentCommandFinished ();
      }
    }

    m_NoChildrenPollCount = 0;
    m_CommandState = CS_WAIT_SUBSIDIARY; // Can't use SetStateRecursive ().
    ChangeHappened (false /* Settings file doesn't need saving */);
    return;
  }

  if (m_CommandState == CS_WAIT_SUBSIDIARY)
  {
    // Wait for the child Commands (just subsidiary ones) to have finished
    // running for a little while, since there may be a poll tick when they
    // aren't running prior to starting up.

    bool AChildIsRunning = false;
    for (int i = 0; i < (int) m_CommandChildren.size (); i++)
    {
      Command *ChildPntr = m_CommandChildren [i];
      if (!ChildPntr->IsSubsidiary ())
        continue;

      if (ChildPntr->IsRunningOrWaiting ())
      {
        AChildIsRunning = true;
        break;
      }
    }

    if (AChildIsRunning)
      m_NoChildrenPollCount = 0;
    else
      m_NoChildrenPollCount++;

    if (m_NoChildrenPollCount >= 3)
    {
      // No children running for a little while, can assume they're all
      // finished, so stop waiting and go back to idle.  Inherit error code
      // from children (should be valid now that they've stopped).

      for (int i = 0; i < (int) m_CommandChildren.size (); i++)
      {
        Command *ChildPntr = m_CommandChildren [i];
        if (!ChildPntr->IsSubsidiary ())
          continue;
        if (m_ChildProcessExitCode == 0)
          m_ChildProcessExitCode = ChildPntr->m_ChildProcessExitCode;
      }

      // Print out the final error code now, so that it's at the end of the
      // log.

      char TempString [PATH_MAX + 1024];
      snprintf (TempString, sizeof (TempString),
        "%s \"%s\" finished on %s.\nIt provided an exit code of %d (by "
        "convention zero means OK).\n",
        IsSubsidiary () ? "Subcommand" : "Command", Name (),
        CurrentTimeAsString (), m_ChildProcessExitCode);
      AppendLogText (TempString);

      m_LogFile.Unset (); // Stop logging, closes the file.
      m_CommandState = CS_READY_TO_RUN;
      ChangeHappened (false /* Settings file doesn't need saving */);
    }
    return;
  }

  if (m_CommandState == CS_READY_TO_RUN)
  {
    // See if this Command should start running now.

    bool bStartRunning = false;
    switch (m_TriggerType)
    {
      case TRG_CRON:
        if (m_LastRunStartTime.tm_min != CurrentTime->tm_min ||
        m_LastRunStartTime.tm_hour != CurrentTime->tm_hour ||
        m_LastRunStartTime.tm_mday != CurrentTime->tm_mday ||
        m_LastRunStartTime.tm_mon != CurrentTime->tm_mon ||
        m_LastRunStartTime.tm_year != CurrentTime->tm_year ||
        m_LastRunStartTime.tm_isdst != CurrentTime->tm_isdst)
        {
          // Last start time is in a different minute than current time (only
          // start something once per minute).

          if (m_TriggerCronBits.TimeMatches (CurrentTime))
            bStartRunning = true;
        }
        break;

      case TRG_NAMED_COMMAND:
        if (m_TriggerNamedCommandHasFinished)
        {
          bStartRunning = true;
          m_TriggerNamedCommandHasFinished = false;
        }
        break;

      case TRG_REBOOT:
        if (m_LastRunStartTime.tm_year == 0 &&
        m_LastRunStartTime.tm_min == 0 &&
        m_LastRunStartTime.tm_hour == 0 &&
        m_LastRunStartTime.tm_mday == 0 &&
        m_LastRunStartTime.tm_mon == 0)
          bStartRunning = true;
        break;

      default:
        break;
    }

    if (bStartRunning)
    {
      if (m_Disabled && IsSubsidiary ())
      {
        // Special case for disabled subsidiary Commands, which run their
        // children anyway.

        char TempString [PATH_MAX + 1024];
        snprintf (TempString, sizeof (TempString),
          "Subcommand \"%s\" is disabled, actions skipped on %s.\n",
          Name (), CurrentTimeAsString ());
        AppendLogText (TempString);

        m_NoChildrenPollCount = 0;
        m_CommandState = CS_WAIT_SUBSIDIARY; // Can't use SetStateRecursive ().
        ChangeHappened (false /* Settings file doesn't need saving */);

        // Tell the child commands that they may now run.

        for (int i = 0; i < (int) m_CommandChildren.size (); i++)
        {
          Command *ChildPntr = m_CommandChildren [i];
          ChildPntr->ParentCommandFinished ();
        }
      }
      else if (!m_Disabled)
      { // Run the Command, reusing the manual run method for convenience.
        BMessage RunScriptingMessage (B_EXECUTE_PROPERTY);
        RunScriptingMessage.AddSpecifier (g_PropertyNames[PN_RUN]);
        RunScriptingMessage.AddSpecifier (g_PropertyNames[PN_COMMAND], Name ());
        g_App->PostMessage (&RunScriptingMessage);
      }
    }
  }
}


/* A Command is looking for some object to execute a scripting command. */

BHandler * Command::ResolveSpecifier (
  BMessage *MessagePntr,
  int32 Index,
  BMessage *SpecifierMsgPntr,
  int32 SpecificationKind,
  const char *PropertyNamePntr)
{
  status_t ErrorCode = B_OK;
  char TempString [PATH_MAX + 1024];

#if 0
printf ("Resolving Specifier in Command object, Message is: ");
MessagePntr->PrintToStream();
printf ("Specifier index %d, Specifier kind %d, Property \"%s\"\n",
  (int) Index, (int) SpecificationKind, PropertyNamePntr);
#endif

  /* See if it is one of our sub-object selection properties. */

  if (strcasecmp (PropertyNamePntr, g_PropertyNames[PN_FIELD]) == 0)
  {
    // Looking for a named Field?

    if (SpecificationKind == B_NAME_SPECIFIER)
    {
      const char *NamePntr = NULL;
      ErrorCode = SpecifierMsgPntr->FindString (g_NameName, &NamePntr);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString),
          "\"%s\" string field is missing in the scripting specifier for "
          "the \"%s\" property", g_NameName, PropertyNamePntr);
        goto ErrorExit;
      }

      FieldMap::iterator FieldIter;
      FieldIter = m_FieldMap.find (NamePntr);
      if (FieldIter != m_FieldMap.end ())
      {
        // Found the named Field, pass on the rest of the scripting command
        // to the Field object.

        MessagePntr->PopSpecifier ();
        return FieldIter->second;
      }
      ErrorCode = B_NAME_NOT_FOUND;
      snprintf (TempString, sizeof (TempString),
        "Property \"%s\" sub-object specified by name \"%s\" not found inside "
        "Command named \"%s\"",
        PropertyNamePntr, NamePntr, Name ());
      goto ErrorExit;
    }
    else if (SpecificationKind == B_INDEX_SPECIFIER)
    {
      // Look for the numbered (indexed) Field, the slow way.

      int32 DesiredIndex = 0;
      ErrorCode = SpecifierMsgPntr->FindInt32 (g_NameIndex, &DesiredIndex);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field is "
          "missing in the scripting specifier for the \"%s\" property of "
          "Command \"%s\"",
          g_NameIndex, PropertyNamePntr, Name ());
        goto ErrorExit;
      }

      if (DesiredIndex < 0 || DesiredIndex >= (int) m_FieldMap.size ())
      {
        ErrorCode = B_BAD_INDEX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field for "
          "property \"%s\" is %d, which is not between 0 and %d, for "
          "Command \"%s\"",
          g_NameIndex, PropertyNamePntr, (int) DesiredIndex,
          (int) m_FieldMap.size() - 1, Name ());
        goto ErrorExit;
      }

      FieldMap::iterator FieldIter;
      FieldIter = m_FieldMap.begin ();
      int32 i = 0;
      while (i < DesiredIndex && FieldIter != m_FieldMap.end ())
      {
        FieldIter++;
        i++;
      }
      if (FieldIter != m_FieldMap.end ())
      {
        MessagePntr->PopSpecifier ();
        return FieldIter->second;
      }
      ErrorCode = B_ERROR;
      strcpy (TempString, "Command::ResolveSpecifier shouldn't happen");
      goto ErrorExit;
    }
  }

  if (Index == 0 && IsItOurScriptingCommand (g_CommandScriptingPropertyList,
  PropertyNamePntr, SpecificationKind))
  {
    return this; /* It's one of our properties, self will handle message. */
  }

  /* Handle an unrecognized scripting command, let the parent figure it out. */

  return BaseChangeScriptable::ResolveSpecifier (
    MessagePntr, Index, SpecifierMsgPntr, SpecificationKind, PropertyNamePntr);

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  BMessage ReplyMessage (/* B_REPLY or maybe */ B_MESSAGE_NOT_UNDERSTOOD);
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return NULL; // Stop trying to resolve the message.
}


status_t Command::RunCommand (char *ErrorTextPntr, int ErrorTextSize)
{
  // Time stamp the start time, even if we fail, so that it doesn't try to
  // start it repeatedly.

  time_t CurrentTime;
  time (&CurrentTime);
  localtime_r (&CurrentTime, &m_LastRunStartTime);

  // Update m_CommandLine with the latest value.

  ExpandTemplateIntoCommandLine ();

  // Create the pipe for transfering standard output data from the child.

  snprintf (ErrorTextPntr, ErrorTextSize,
    "RunCommand: Trouble setting up StdOut pipe for Command \"%s\"", Name ());

  if (m_PipeStdOutFromChild [0] != -1 || m_PipeStdOutFromChild [1] != -1)
    return B_NO_INIT; // Should have started at -1 set by the constructor.

  if (pipe (m_PipeStdOutFromChild) != 0) // Returns -1 if failed, 0 if okay.
    return errno; // Pipe creation failed.

  // From here on in (pipe allocated), errors go to ErrorExit to do resource
  // cleanup (close pipe, and anything else).

  const char *ArgumentArray [4];
  pid_t ChildThreadID = -1;
  status_t ErrorCode = B_OK;
  int NewStdIn = -1;
  int PipeFlags = 0;

  // Make the reading end [0] of StdOut non-blocking to the main program, so it
  // can continue running when there is no data available.  Writing end [1]
  // left unchanged so child will block when the pipe is full.  Hopefully the
  // program loop is running often enough to empty the pipe into the log file
  // and output text display, so that the child doesn't pause too often.

  PipeFlags = fcntl (m_PipeStdOutFromChild [0], F_GETFL);
  PipeFlags |= O_NONBLOCK;
  fcntl (m_PipeStdOutFromChild [0], F_SETFL, PipeFlags);

  // Read end of the pipe not used by the child, make it close on exec.

  PipeFlags = fcntl (m_PipeStdOutFromChild [0], F_GETFD);
  PipeFlags |= FD_CLOEXEC;
  fcntl (m_PipeStdOutFromChild [0], F_SETFD, PipeFlags);

  // Open a file handle to /dev/null for the child's standard input.

  NewStdIn = open ("/dev/null", O_RDONLY); // -1 if it fails.
  if (NewStdIn == -1)
  {
    ErrorCode = errno;
    snprintf (ErrorTextPntr, ErrorTextSize,
      "RunCommand: Failed to open /dev/null as new stdin for Command \"%s\"",
      Name ());
    goto ErrorExit;
  }

  // Set up standard files for the new child process to inherit.  Standard
  // output and standard error both write to the pipe, standard input reads
  // from /dev/null.

  if (!g_LockStdFilesPntr->Lock ())
  {
    ErrorCode = ENOLCK;
    snprintf (ErrorTextPntr, ErrorTextSize,
      "RunCommand: Failed to lock the StdInOut semaphore for Command \"%s\"",
      Name ());
    goto ErrorExit;
  }

  dup2 (NewStdIn, STDIN_FILENO);
  dup2 (m_PipeStdOutFromChild[1], STDOUT_FILENO);
  dup2 (m_PipeStdOutFromChild[1], STDERR_FILENO);

  // We shouldn't touch the write end of the StdOut pipe.  Close our copy of
  // the file descriptor (it gets duplicated as part of the load_image ()
  // operation), leaving the child as the only one using it.

  close (m_PipeStdOutFromChild[1]);
  m_PipeStdOutFromChild[1] = -1;

  close (NewStdIn); // Not feeding input, close now so no CLOEXEC needed.
  NewStdIn = -1;

  // Load the program we want to run, inheriting the current file handles.
  // Note that it doesn't do a fork (doesn't inherit the memory space of this
  // program and then throw it away), so it's faster, but trickier to set up.

  ArgumentArray [0] = "/bin/sh";
  ArgumentArray [1] = "-c"; // Option for bash to run a string as a command.
  ArgumentArray [2] = m_CommandLine.String (); // The string to run.
  ArgumentArray [3] = NULL;

  ChildThreadID = load_image (
    3 /* argument count */, ArgumentArray, const_cast<const char **>(environ));

  if (ChildThreadID < 0) // Get a negative ID if load failed.
  {
    ErrorCode = ChildThreadID;
    snprintf (ErrorTextPntr, ErrorTextSize,
      "RunCommand: Failed to load /bin/sh to run Command \"%s\"",
      Name ());
    goto ErrorExit;
  }

  // Switch the new program to a new process group (zero means actual group
  // number is copied from the thread ID, so it will be unique), so signals
  // sent the the main program (like control-C SIGTERM) don't also affect the
  // child.  Also makes the new child the head of a process group, so if
  // someone kills it, all the commands fired off by the child are killed too.
  // And start it running.

  setpgid (ChildThreadID, ChildThreadID);
  resume_thread (ChildThreadID); // Assume it works.
  m_ChildProcessID = ChildThreadID; // Main thread ID is also process ID.

  // Open the log file if requested (subsidiaries use their ancestor's log file
  // and settings).  Create log directory if needed.  Errors aren't fatal (just
  // run without a log file).

  m_LogFile.Unset ();
  if (!IsSubsidiary () && m_LogFileEnabled)
  {
    BString LogPathName;
    GetStockNamesAndPaths (NULL, &LogPathName, NULL, NULL);
    BPath LogDirPath (LogPathName.String ());
    LogDirPath.GetParent (&LogDirPath); // Self parenting is allowed.
    BDirectory LogDirectory;
    ErrorCode = LogDirectory.SetTo (LogDirPath.Path ());
    if (ErrorCode != B_OK)
    {
      // Create all directories down to the log directory.
      create_directory (LogDirPath.Path (), 0750 /* rwxr-x--- */);
      ErrorCode = LogDirectory.SetTo (LogDirPath.Path ());
    }
    if (ErrorCode != B_OK)
    {
      m_LogAccumulated << "\nUnable to access log directory of \"" <<
        LogDirPath.Path () << "\", error number " << ErrorCode << " (" <<
        strerror (ErrorCode) << ").  Logging to file won't be done for "
        "this run.\n";
    }
    else // The specific log directory now exists, make the log file.
    {
      ErrorCode = m_LogFile.SetTo (LogPathName.String (),
        B_WRITE_ONLY | B_CREATE_FILE | B_OPEN_AT_END |
        O_CLOEXEC /* Haiku's BFile implementation also adds this one */);
      if (ErrorCode != B_OK)
      {
        m_LogFile.Unset ();
        m_LogAccumulated << "\nUnable to write to log file \"" <<
          LogPathName << "\", error number " <<
          ErrorCode << " (" << strerror (ErrorCode) <<
          ").  Logging to file won't be done for this run.\n";
      }
      else // File is open.  Set mode to prevent global reading of it.
      {
        chmod (LogPathName.String (), 0640 /* rw-r----- */);
      }
    }
  }
  ErrorCode = B_OK;
  goto NormalExit;

ErrorExit: // Something went wrong, clean up more than normal.
  if (m_PipeStdOutFromChild [0] != -1)
    close (m_PipeStdOutFromChild [0]);
  m_PipeStdOutFromChild [0] = -1;

  if (m_PipeStdOutFromChild [1] != -1)
    close (m_PipeStdOutFromChild [1]);
  m_PipeStdOutFromChild [1] = -1;

NormalExit:
  if (NewStdIn != -1)
    close (NewStdIn);

  if (g_LockStdFilesPntr->IsLocked ())
  {
    dup2 (g_OriginalStdIn, STDIN_FILENO);
    dup2 (g_OriginalStdOut, STDOUT_FILENO);
    dup2 (g_OriginalStdErr, STDERR_FILENO);
    g_LockStdFilesPntr->Unlock ();
  }

  return ErrorCode;
}


void Command::SetStateRecursive (CommandState NewState)
{
  if (m_CommandState != NewState)
  {
    m_CommandState = NewState;
    ChangeHappened (false /* SettingsAffected */);
  }

  // Clear pending other Command has finished flag so that when changing from
  // editing or some other state to ready-to-run, it doesn't suddenly start
  // running.  It should only start up if the other Command finishes while this
  // Command is in the ready-to-run state.

  m_TriggerNamedCommandHasFinished = false;

  // Also set the state for subsidiary children.

  for (int i = 0; i < (int) m_CommandChildren.size (); i++)
  {
    Command *ChildPntr = m_CommandChildren [i];
    if (ChildPntr->IsSubsidiary ())
    {
      ChildPntr->SetStateRecursive (NewState);
    }
  }
}


status_t Command::StopCommand (char *ErrorTextPntr, int ErrorTextSize)
{
  status_t ErrorCode;

  // Make sure all subsidiary child Commands are stopped too.

  for (int i = 0; i < (int) m_CommandChildren.size (); i++)
  {
    Command *ChildPntr = m_CommandChildren [i];
    if (ChildPntr->IsSubsidiary ())
    {
      ErrorCode = ChildPntr->StopCommand (ErrorTextPntr, ErrorTextSize);
      if (ErrorCode != B_OK)
        return ErrorCode;
    }
  }

  snprintf (ErrorTextPntr, ErrorTextSize,
    "StopCommand: Unable to stop Command \"%s\".", Name ());

 if (m_ChildProcessID != -1)
 {
    // Use negative thread ID to kill the child process group (the /bin/sh
    // process is at the head of it) rather than just the /bin/sh thread,
    // otherwise children of the shell (generally everything invoked by the
    // command line) won't get killed.

    if (kill (-m_ChildProcessID, SIGTERM) == -1)
      return errno; // Failed to kill it.
  }

  return B_OK;
}


void Command::UpgradeFromNewerModel (BMessage *ModelPntr)
{
  status_t ErrorCode;

  // Copy over the Template from the Model.

  const char *NewTemplatePntr = NULL;
  if (B_OK != ModelPntr->FindString (g_TemplateName, &NewTemplatePntr))
    return; // Something really wrong, no Template!
  if (m_Template != NewTemplatePntr)
  {
    printf ("Upgrading \"%s\" template from \"%s\" to \"%s\".\n",
      Name (), m_Template.String (), NewTemplatePntr);
    m_Template = NewTemplatePntr;
  }

  // Update this Command's UserData.

  UserDataUpgradeFromNewerModel (ModelPntr);

  // Upgrade all the Fields inside this Command.  Make missing ones too.

  for (int32 iField = 0; true; iField++)
  {
    BMessage FieldMessage;
    ErrorCode = ModelPntr->FindMessage(g_FieldListName, iField, &FieldMessage);
    if (ErrorCode != B_OK)
      break; // Likely got to the end of the Field data in the Model.

    const char *FieldNamePntr = g_EmptyString;
    if (FieldMessage.FindString ("_name", &FieldNamePntr) != B_OK)
      continue; // Unable to read the name of the Field, skip it.

    FieldMap::iterator FieldIter;
    FieldIter = m_FieldMap.find (FieldNamePntr);
    if (FieldIter != m_FieldMap.end ())
    {
      // Field exists, update it with the Model's data.

      Field *FieldPntr = FieldIter->second;
      FieldPntr->UserDataUpgradeFromNewerModel (&FieldMessage);
    }
    else // Model Field doesn't exist in current Command, create it.
    {
      BArchivable *DearchivedPntr = instantiate_object (&FieldMessage);
      if (DearchivedPntr == NULL)
        continue; // Unable to find C++ class of object to rebuild it, etc.

      Field *FieldPntr = cast_as (DearchivedPntr, Field);
      if (FieldPntr != NULL)
      {
        printf ("Upgrading \"%s\" missing Field \"%s\" by making a new one.\n",
          Name (), FieldNamePntr);
        InternalAddField (FieldPntr);
      }
      else // Some other kind of object, someone's been hacking the settings?
        delete DearchivedPntr;
    }
  }
}


/* Normally used when adding a tree of subsidiary Commands to the global map of
active Commands.  Their parents may have been renamed as part of the addition,
so update the trigger to match. */

void Command::UseParentNameAsTrigger ()
{
  if (m_CommandParent == NULL)
    return;

  m_Trigger = m_CommandParent->Name ();
}


/******************************************************************************
 * Implementation of the Field class, constructor, destructor and the rest of
 * the member functions in mostly alphabetical order.
 */

Field::Field (const char *NewName) :
  BaseChangeScriptable (NewName),
  m_OwnerCommand (NULL)
{
}


Field::Field (BMessage *ArchiveMessagePntr) :
  BaseChangeScriptable (ArchiveMessagePntr),
  m_OwnerCommand (NULL)
{
  // Load our subclass specific data from the archived Field data.

  ArchiveMessagePntr->FindString (g_ValueName, &m_Value);
}


Field::~Field ()
{
}


/* Archives the Field object into a BMessage, for storage (usually in the
settings file). */

status_t Field::Archive (BMessage *ArchiveMessagePntr, bool Deep) const
{
  status_t ErrorCode;

  // Do the parent class archiving first.

  ErrorCode = BaseChangeScriptable::Archive (ArchiveMessagePntr, Deep);
  if (ErrorCode != B_OK)
    return ErrorCode;
/*
  ArchiveMessagePntr->AddString ("class", "Field");
    // BHandler's Archive function adds the class name of "Field" too (rather
    // than "BHandler"), so this isn't really needed, at least in BeOS R5.03.
*/
  // Add our subclass specific data to the archive.

  ArchiveMessagePntr->AddString (g_ValueName, m_Value);
  return ErrorCode;
}


/* This Field has been changed, increment the serial number and let the
parent object know about the change. */

void Field::ChangeHappened (bool SettingsAffected)
{
  BaseChangeScriptable::ChangeHappened (SettingsAffected);

  if (m_OwnerCommand != NULL)
    m_OwnerCommand->ChangeHappened (SettingsAffected);
}


/* Replaces all keywords related to this field in the given CommandLine string.
The result is in the modified CommandLine. */

void Field::ExpandMagicWords (BString &CommandLine)
{
  // The format is the Field's name followed by AsDir, AsName, AsDatedString,
  // AsDatedURL, AsNCFTP, AsSafe, or AsUserData(name) postfixes.  Process each
  // occurrence of the Field name separately, searching for as many postfix
  // keywords as possible and then processing them in a stock order.

  enum AsWordsEnum {
    AS_USERDATA = 0,
    AS_DIR,
    AS_NAME,
    AS_DATED_STRING,
    AS_DATED_URL,
    AS_NCFTP,
    AS_SAFE,
    AS_MAX};

  const char *AsWords [AS_MAX] =
  {
    "AsUserData(", // Bracket is included on purpose.
    "AsDir",
    "AsName",
    "AsDatedString",
    "AsDatedURL",
    "AsNCFTP",
    "AsSafe"
  };

  int32 FieldNameLength = strlen (Name ());

  // For each occurrence of the Field's name check for keywords following it.
  // Occasionally a Field has a single letter name and thus triggers an
  // infinite loop of expansion, so cut it off after a few hundred iterations.

  for (int IterationCount = 0; IterationCount < 200; IterationCount++)
  {
    int32 iFieldName = CommandLine.FindFirst (Name ());
    if (iFieldName < 0) // Negative position if FieldName not found.
      break;

    // Look for the various AsBlah keywords after the Field name, advancing
    // SourcePntr as each is discovered.  Keep on going until no more keywords
    // found.

    const char *FieldNameStartPntr = CommandLine.String () + iFieldName;
    const char *SourcePntr = FieldNameStartPntr + FieldNameLength;

    bool AKeywordFound = false;
    bool FoundKeywords[AS_MAX];
    memset (FoundKeywords, 0, sizeof (FoundKeywords));

    BString UserDataName; // The stuff inside the AsUserData(stuff) brackets.

    do
    {
      AKeywordFound = false;
      for (int iKeyword = 0; iKeyword < AS_MAX; iKeyword++)
      {
        const char *Keyword = AsWords [iKeyword];
        if (strncmp (SourcePntr, Keyword, strlen (Keyword)) != 0)
          continue;

        AKeywordFound = true;
        FoundKeywords[iKeyword] = true;
        SourcePntr += strlen (Keyword);

        // The AsUserData keyword is followed by the identifier of the
        // particular userdata, then an ending bracket.  Save the name.

        if (iKeyword == AS_USERDATA)
        {
          const char *NameStartPntr = SourcePntr;
          const char *EndingBracketPntr = strchr (NameStartPntr, ')');
          if (EndingBracketPntr == NULL)
            UserDataName = g_EmptyString;
          else
          {
            UserDataName.SetTo (NameStartPntr,
              EndingBracketPntr - NameStartPntr);
            SourcePntr = EndingBracketPntr + 1; // Next word after the bracket.
          }
        }
        break; // Restart the keyword list from the beginning for next word.
      }
    }
    while (AKeywordFound);

    int32 FieldNameAndPostfixesTotalLength = SourcePntr - FieldNameStartPntr;

    // Now do the As... actions in a preset order, acting on the text value of
    // the Field.

    BString ValueString (m_Value); // Start with the Field's value text.

    // AsUserData(name).

    if (FoundKeywords[AS_USERDATA])
      ValueString = UserDataGetString (UserDataName.String ());

    // AsDir.  Use the Value up to and including the last slash.  Empty string if
    // no slash.

    if (FoundKeywords[AS_DIR])
    {
      BString ValueAsDir;
      int iLastSlash = ValueString.FindLast ('/');
      if (iLastSlash >= 0)
        ValueString.CopyInto (ValueAsDir, 0, iLastSlash + 1);
      ValueString = ValueAsDir;
    }

    // AsName.  Use the text after the last slash.

    if (FoundKeywords[AS_NAME])
    {
      BString ValueAsName (ValueString);
      int iLastSlash = ValueString.FindLast ('/');
      if (iLastSlash >= 0)
        ValueString.CopyInto (ValueAsName, iLastSlash + 1,
          ValueString.Length () - iLastSlash - 1);
      ValueString = ValueAsName;
    }

    // AsDatedString.  If the first character of the text is "+" then interpret
    // it as a date format applied to the current time with an optional time
    // offset from UserData("DateOffset").

    if (FoundKeywords[AS_DATED_STRING])
    {
      const char *NamePntr = ValueString.String ();
      if (NamePntr [0] == '+')
      {
        BString NewName;
        ExpandDateFormat (NewName, NamePntr + 1 /* skip leading + sign */,
          UserDataGetString (g_KeyDateOffset));

        if (NewName.Length () > 0)
          ValueString = NewName;
      }
    }

    // AsDatedURL.  Convert value to a URL, if the first character of its name
    // is "+" then interpret it as a date format applied to the current time
    // with an optional time offset from UserData("DateOffset").  Similarly the
    // directory path is date converted if it starts with a plus sign.  Then
    // recombine the dated text back into URL format.

    if (FoundKeywords[AS_DATED_URL])
    {
      URLConverter ValueAsURL;
      bool DateExpansionDone = false;

      // Date expand the file name, if it starts with a plus sign.

      ValueAsURL.FromString (ValueString.String ());
      const char *NamePntr = ValueAsURL.GetValue (URLConverter::UET_FILENAME);
      if (NamePntr [0] == '+')
      {
        BString NewName;
        ExpandDateFormat (NewName, NamePntr + 1 /* skip leading + sign */,
          UserDataGetString (g_KeyDateOffset));

        if (NewName.Length () > 0)
        {
          ValueAsURL.SetValue (URLConverter::UET_FILENAME, NewName.String ());
          DateExpansionDone = true;
        }
      }

      // Date expand the directory path.

      NamePntr = ValueAsURL.GetValue (URLConverter::UET_PATH);
      if (NamePntr [0] == '+')
      {
        BString NewName;
        ExpandDateFormat (NewName, NamePntr + 1 /* skip leading + sign */,
          UserDataGetString (g_KeyDateOffset));

        if (NewName.Length () > 0)
        {
          ValueAsURL.SetValue (URLConverter::UET_PATH, NewName.String ());
          DateExpansionDone = true;
        }
      }

      // Convert the date expanded text back to a URL, only if needed.

      if (DateExpansionDone)
        ValueAsURL.ToString (ValueString);
    }

    // AsNCFTP.  Convert value to a URL, then break out the various components
    // and format them as arguments to the ncftpput command.

    if (FoundKeywords[AS_NCFTP])
    {
      URLConverter ValueAsURL;
      ValueAsURL.FromString (ValueString.String ());

      BString UserName (ValueAsURL.GetValue (URLConverter::UET_USERNAME));
      BString Password (ValueAsURL.GetValue (URLConverter::UET_PASSWORD));
      BString HostName (ValueAsURL.GetValue (URLConverter::UET_ADDRESS));
      BString RemotePath (ValueAsURL.GetValue (URLConverter::UET_PATH));

      SafeEscape (UserName);
      SafeEscape (Password);
      SafeEscape (HostName);
      SafeEscape (RemotePath);

      BString NCFTP_Arguments;
      NCFTP_Arguments << "-u \"" << UserName << "\" -p \"" << Password <<
        "\" \"" << HostName << "\" \"" << RemotePath << "\"";
      ValueString = NCFTP_Arguments;
    }

    // AsSafe.  Escape dangerous characters.

    if (FoundKeywords[AS_SAFE])
      SafeEscape (ValueString);

    // Replace the Field name and AsBlah keywords with the final modified Field
    // value.

    CommandLine.Remove (iFieldName, FieldNameAndPostfixesTotalLength);
    CommandLine.Insert (ValueString, iFieldName);
  }
}


/* Returns a list of the scripting Fields a Field understands. */

status_t Field::GetSupportedSuites (BMessage *MessagePntr)
{
  BPropertyInfo TempPropInfo (g_FieldScriptingPropertyList);
  MessagePntr->AddString ("suites", SUITE_BASE "field");
  MessagePntr->AddFlat ("messages", &TempPropInfo);
  return BaseChangeScriptable::GetSupportedSuites (MessagePntr);
}


/* Class global function for recreating objects from a BMessage archive.  Most
of the data loading work is done in the constructor, which takes an Archive
argument. */

BArchivable *Field::Instantiate (BMessage *ArchiveMessagePntr)
{
  if (validate_instantiation (ArchiveMessagePntr, "Field"))
    return new Field (ArchiveMessagePntr);
  return NULL;
}


/* Process mostly scripting messages. */

void Field::MessageReceived (BMessage *MessagePntr)
{
  const char           *ArgumentString = NULL;
  status_t              ErrorCode = B_ERROR;
  const char           *PropertyName;
  BMessage              ReplyMessage (B_REPLY);
  int32                 SpecifierIndex;
  int32                 SpecifierKind;
  BMessage              SpecifierMessage;
  char                  TempString [PATH_MAX + 1024];

  switch (MessagePntr->what)
  {
    case B_GET_PROPERTY:
    case B_SET_PROPERTY:
    case B_COUNT_PROPERTIES:
    case B_CREATE_PROPERTY:
    case B_DELETE_PROPERTY:
    case B_EXECUTE_PROPERTY:
      // Look at specifier[0] to find out what the scripting does.
      if (B_OK == MessagePntr->SetCurrentSpecifier (0))
        ErrorCode = MessagePntr->GetCurrentSpecifier (&SpecifierIndex,
        &SpecifierMessage, &SpecifierKind, &PropertyName);
      else // No specifiers, assume it is a direct operation on this Field.
      {
        PropertyName = g_PropertyNames[PN_FIELD];
        SpecifierKind = B_DIRECT_SPECIFIER;
        ErrorCode = B_OK;
      }

      if (ErrorCode == B_OK &&
      IsItOurScriptingCommand (g_FieldScriptingPropertyList,
      PropertyName, SpecifierKind))
        break; // It's one of ours, do it.
      // Else fall through, and punt the message to the parent class.

    default: // Not one of our messages, let the parent handle it.
      BaseChangeScriptable::MessageReceived (MessagePntr);
      return;
  }

  if (m_OwnerCommand == NULL)
  {
    snprintf (TempString, sizeof (TempString),
      "Can't do anything to Field named \"%s\" because it isn't part of a "
      "Command.  How did you even send a scripting message to it?", Name ());
    ErrorCode = B_ERROR;
    goto ErrorExit;
  }

  snprintf (TempString, sizeof (TempString),
    "Scripting operation code %c%c%c%c (get, set, count, etc) or access type "
    "%d (direct=1, index=2, reverse index=3, range=4, reverse range=5, "
    "name=6, ID=7) is unsupported for property \"%s\" of the Field "
    "object named \"%s\" which is part of Command named \"%s\"",
    (char) (MessagePntr->what >> 24),
    (char) (MessagePntr->what >> 16),
    (char) (MessagePntr->what >>  8),
    (char) (MessagePntr->what >>  0),
    (int) SpecifierKind, PropertyName, Name (), m_OwnerCommand->Name ());
  ErrorCode = B_BAD_SCRIPT_SYNTAX;

  // Handle the Name property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_NAME]) == 0)
  {
    // Get the current name.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, Name ());
      goto SuccessExit;
    }

    // Set the new Name.  This also means changing the Map of Fields inside the
    // parent Command since they are in alphabetical order.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Name property "
          "from.  Was for Field \"%s\" of Command \"%s\"",
          g_DataName, Name (), m_OwnerCommand->Name ());
        goto ErrorExit;
      }

      if (!m_OwnerCommand->IsEditing ())
      {
        snprintf (TempString, sizeof (TempString),
          "Can't change name of Field named \"%s\" to \"%s\" because the "
          "related Command named \"%s\" isn't in edit mode",
          Name (), ArgumentString, m_OwnerCommand->Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      if (m_OwnerCommand->DoesNamedFieldExist (ArgumentString))
      {
        snprintf (TempString, sizeof (TempString),
          "Field named \"%s\" already exists in Command \"%s\", can't "
          "rename \"%s\" to it",
          ArgumentString, m_OwnerCommand->Name (), Name ());
        ErrorCode = B_NAME_IN_USE;
        goto ErrorExit;
      }

      Command *OwnerCommandPntr = m_OwnerCommand;
        // Need a copy of the parent, since removing the field sets
        // m_OwnerCommand to NULL, and adding it sets it to the Command it
        // is added to.

      OwnerCommandPntr->InternalRemoveField (this);
      SetName (ArgumentString);
      OwnerCommandPntr->InternalAddField (this);
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Name property.
  }

  // Handle the Serial property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_SERIAL]) == 0)
  {
    // Get the current serial number.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, GetSerial ());
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Serial property.
  }

  // Handle the Field property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_FIELD]) == 0)
  {
    if (MessagePntr->what == B_DELETE_PROPERTY /* any kind of specifier OK */)
    {
      if (!m_OwnerCommand->IsEditing ())
      {
        snprintf (TempString, sizeof (TempString),
          "Can't delete Field named \"%s\" because the related "
          "Command named \"%s\" isn't in edit mode",
          Name (), m_OwnerCommand->Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      Command *OwnerCommandPntr = m_OwnerCommand; // Copy before NULLified
      OwnerCommandPntr->InternalRemoveField (this); // by the removal.
      OwnerCommandPntr->ChangeHappened ();
      delete this; // Careful, can't reference member variables after here.
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Field property.
  }

  // Handle the Value property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_VALUE]) == 0)
  {
    // Get the current Value.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, m_Value);
      goto SuccessExit;
    }

    // Set the new Value.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Value property "
          "from.  Was for Field \"%s\" of Command \"%s\"",
          g_DataName, Name (), m_OwnerCommand->Name ());
        goto ErrorExit;
      }

      if (!m_OwnerCommand->IsEditing ())
      {
        snprintf (TempString, sizeof (TempString),
          "Can't change Value of Field named \"%s\" to \"%s\" because the "
          "related Command named \"%s\" isn't in edit mode",
          Name (), ArgumentString, m_OwnerCommand->Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      m_Value.SetTo (ArgumentString);
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Value property.
  }

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return;

SuccessExit: /* Success. */
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
}


/* A Field is looking for the object to execute a scripting command. */

BHandler * Field::ResolveSpecifier (
  BMessage *MessagePntr,
  int32 Index,
  BMessage *SpecifierMsgPntr,
  int32 SpecificationKind,
  const char *PropertyNamePntr)
{
#if 0
printf ("Resolving Specifier in Field object, Message is: ");
MessagePntr->PrintToStream();
printf ("Specifier index %d, Specifier kind %d, Property \"%s\"\n",
  (int) Index, (int) SpecificationKind, PropertyNamePntr);
#endif

  if (Index == 0 && IsItOurScriptingCommand (g_FieldScriptingPropertyList,
  PropertyNamePntr, SpecificationKind))
  {
    return this; /* It's one of our properties, self will handle message. */
  }

  /* Handle an unrecognized scripting Field, let the parent figure it out. */

  return BaseChangeScriptable::ResolveSpecifier (
    MessagePntr, Index, SpecifierMsgPntr, SpecificationKind, PropertyNamePntr);
}


void Field::SetOwnerCommand (class Command *CommandPntr)
{
  m_OwnerCommand = CommandPntr;
}



/******************************************************************************
 * Implementation of the ScriptOCronApp class, constructor, destructor and the
 * rest of the member functions in mostly alphabetical order.
 */

ScriptOCronApp::ScriptOCronApp ()
: BApplication (g_AppSignature),
  m_ChangeSerial (GetNextSerialNumber ()),
  m_QuitRequestedStashedMessage (NULL),
  m_SerialWhenModelsLastChanged (-1),
  m_SerialWhenSettingsSaved (-1),
  m_TimeSettingsWereLastSaved (0)
{
  status_t    ErrorCode;

  /* Set up the pathname which identifies our settings directory.  Note that
  the actual settings are loaded later on (or set to defaults) by the main()
  function, before this BApplication starts running.  So we don't bother
  initialising the other setting related variables here. */

  ErrorCode = find_directory (B_USER_SETTINGS_DIRECTORY,
    &m_SettingsDirectoryPath, true /* Create if needed */);
  if (ErrorCode == B_OK)
    ErrorCode = m_SettingsDirectoryPath.Append (PROGRAM_NAME_WORD);
  if (ErrorCode != B_OK)
    m_SettingsDirectoryPath.SetTo (".");
}


ScriptOCronApp::~ScriptOCronApp ()
{
  // Save the settings if needed.  Also may save the Commands, so don't delete
  // them just yet!

  if (g_SettingsNeedSaving)
    LoadSaveSettings (false /* DoLoad */);

  DeleteAllCommands ();
  DeleteAllModels ();

  delete m_QuitRequestedStashedMessage;
  m_QuitRequestedStashedMessage = NULL;
}


/* Display a box showing information about this program. */

void ScriptOCronApp::AboutRequested ()
{
  BAlert *AboutAlertPntr;

  std::ostringstream TempStream;

  PrintUsage (TempStream, 0 /* WrapWidth */, false /* ListCommands */);
  TempStream.flush ();

  // Remove trailing linefeeds and spaces.

  BString AboutText (TempStream.str().c_str());
  while (isspace (AboutText.ByteAt (AboutText.Length () - 1)))
    AboutText.Truncate (AboutText.Length () - 1);

  AboutAlertPntr = new BAlert ("About", AboutText.String (), "Done");
  if (AboutAlertPntr != NULL)
  {
    AboutAlertPntr->SetShortcut (0, B_ESCAPE);

    // Do a synchronous alert so that the string still exists while the box is
    // up, otherwise it gets deleted and the BAlert sometimes crashes
    // (particularly in Haiku) while trying to reference it.  Though it does
    // mean all Command processing stops while the About is showing.

    AboutAlertPntr->Go ();
  }
}


/* The Application has been changed, increment the serial number.  There is no
parent to notify about the change so that's all that needs to be done.  Other
interested parties can use scripting to check the serial number for changes,
and then update themselves as needed. */

void ScriptOCronApp::ChangeHappened (bool SettingsAffected)
{
  m_ChangeSerial = GetNextSerialNumber ();

  if (SettingsAffected)
    g_SettingsNeedSaving = true;
}


/* Set the settings to the defaults.  Needed in case there isn't a settings
file or it is obsolete. */

void ScriptOCronApp::DefaultSettings ()
{
#if INCLUDE_GUI_CODE
  g_AdvancedMode = false; // Basic the first time.
  g_FontScaleCommandNames = 1.25f;
  g_FontScaleLogText = 1.0f;
  g_InitialWindowFrame = BRect (1, 2, 3, 4); // Meaning not initialised.
  g_ServerMode = false; // Shows the GUI window on the first run.
  g_SimpleWindowTitle = false;
#endif

  m_UserSpecifiedTemporaryDirectory = g_EmptyString;
  m_UserSpecifiedTemporaryDirectorySerial = m_ChangeSerial;

  // Load built-in Models from a resource named "Models", which contains a
  // flattened BMessage made from drag and dropping an AGMSScriptOCron
  // settings file into QuickRes, with a "deres" side trip to a text file
  // resource to give it a name which sticks (seems to be a bug in QuickRes).

  BMessage    ModelArchive;
  const char *ResourceData;
  size_t      ResourceSize;
  BResources *ResourcesPntr;

  if ((ResourcesPntr = AppResources ()) != NULL &&
  (ResourceData = (const char *) ResourcesPntr->LoadResource ('MSGG',
  "Models", &ResourceSize)) != NULL)
  {
    if (B_OK == ModelArchive.Unflatten (ResourceData))
      LoadModels (&ModelArchive, true /* BuiltIn */);
  }

  m_SerialWhenSettingsSaved = m_ChangeSerial;
}


/* Delete all the commands, taking the time to deallocate their memory too. */

void ScriptOCronApp::DeleteAllCommands ()
{
  CommandMap::iterator CommandIter;

  for (CommandIter = m_CommandMap.begin ();
  CommandIter != m_CommandMap.end (); CommandIter++)
  {
    // Just need to delete the value (Command object).  The key is the name of
    // the command as a const char *, which is a pointer into the Command
    // object's data (actually the BHandler's name field) so it implicitly gets
    // deleted when the Command is deleted.

    Command *CommandToBeDeletedPntr = CommandIter->second;
    CommandToBeDeletedPntr->AncestryReset (); // So children not deleted.
    delete CommandToBeDeletedPntr; // Will also remove it from Looper.
  }
  m_CommandMap.clear ();
}


void ScriptOCronApp::DeleteAllModels ()
{
  ModelMap::iterator ModelIter;

  for (ModelIter = m_ModelMap.begin ();
  ModelIter != m_ModelMap.end (); ModelIter++)
  {
    // Just need to delete the value (Model object).  The key is the name of
    // the Model as a const char *, which is a pointer into the Model's
    // BMessage data so it implicitly gets deleted when the Model is deleted.

    BMessage *ModelToBeDeletedPntr = ModelIter->second;
    delete ModelToBeDeletedPntr;
  }
  m_ModelMap.clear ();
}


/* Intercept B_QUIT_REQUESTED.  Otherwise this BLooper/BApplication may be
killed and a reply message automatically sent.  We do both of those ourselves
in an effort to delay death. */

void ScriptOCronApp::DispatchMessage (BMessage *MessagePntr, BHandler *Target)
{
  if (MessagePntr->what != B_QUIT_REQUESTED)
  {
    BApplication::DispatchMessage (MessagePntr, Target);
    return;
  }

  if (m_QuitRequestedStashedMessage != NULL)
  {
    // Reject an unexpected second request to quit.  Can happen if someone
    // closes the window while the operating system is shutting down.

    BMessage ReplyMessage (B_REPLY);
    ReplyMessage.AddBool (g_ResultName, false);
    int32 ThreadID = Thread (); // Haiku needs the thread ID too.
    ReplyMessage.AddInt32 ("thread", ThreadID);
    MessagePntr->SendReply (&ReplyMessage,
      (BHandler *) NULL /* where to send the reply to the reply */,
      500000 /* send timeout */);
    return;
  }

  m_QuitRequestedStashedMessage = DetachCurrentMessage ();

  if (g_QuitCountdown < 0)
  {
    g_QuitCountdown = 5; /* Start the countdown. */
    cerr << PROGRAM_NAME_WORD " shutdown sequence started.\n";

    // Ask all Commands to stop running.  Need the App message loop to
    // keep on running a bit longer so that the zombie processes can be
    // reaped, logs closed, etc.

    CommandMap::iterator CommandIter;
    for (CommandIter = m_CommandMap.begin ();
    CommandIter != m_CommandMap.end (); CommandIter++)
    {
      Command *CommandPntr = CommandIter->second;
      if (CommandPntr->IsRunningOrWaiting ())
      {
        cerr << "Stopping running Command \"" <<
          CommandPntr->Name () << "\".\n";

        BMessage StopScriptingMessage (B_DELETE_PROPERTY);
        StopScriptingMessage.AddSpecifier (g_PropertyNames[PN_RUN]);
        StopScriptingMessage.AddSpecifier (
          g_PropertyNames[PN_COMMAND], CommandPntr->Name ());
        PostMessage (&StopScriptingMessage);
      }
    }
  }
}


/* Tell other programs about the top Application level scripting commands we
support.  Try this command: "hey application/x-vnd.agmsmith.scriptocron
getsuites" to see it in action (this program has to be already running for it
to work).  Scripting commands for subobjects are handled in those objects. */

status_t ScriptOCronApp::GetSupportedSuites (BMessage *MessagePntr)
{
  BPropertyInfo TempPropInfo (g_AppScriptingPropertyList);
  MessagePntr->AddString ("suites", SUITE_BASE "application");
  MessagePntr->AddFlat ("messages", &TempPropInfo);
  return BApplication::GetSupportedSuites (MessagePntr);
}


const char *ScriptOCronApp::GetSettingsDirectory ()
{
  return m_SettingsDirectoryPath.Path ();
}


/* Add the given Command to the Map of all Commands, as well as any subsidiary
children linked to the Command via child pointers (rather than Trigger names,
and assumed to be a tree structure, guaranteed to be not circular).  Used by
the Command object when it is renaming itself, also as an undo system
(restoring an old copy of a Command).  Replaces any existing Command with the
same name.  Updates Loopers too.  Subsidiary children (they all should be
subsidiary) will be renamed to have the ancestor's name, a sequence number, and
the Model name of the child if it has one, before being added. */

void ScriptOCronApp::InternalAddCommandGuts (
  Command *CommandPntr, const char *AncestorName, int Depth)
{
  // Rename subsidiary Commands to match their ancestor.

  if (CommandPntr->IsSubsidiary () && AncestorName != NULL)
  {
    BString NewName;
    BString ModelName;
    ModelName = CommandPntr->UserDataGetString (g_KeyOriginalModel);
    for (int SequenceNo = Depth; SequenceNo < 1000; SequenceNo++)
    {
      char SequenceString [6];
      sprintf (SequenceString, "%03d", SequenceNo);
      NewName = AncestorName;
      NewName << "." << SequenceString;
      if (ModelName.Length () > 0)
        NewName << "." << ModelName;
      if (!DoesNamedCommandExist (NewName.String()))
      { // Found a unique new name.
        CommandPntr->SetName (NewName.String ());
        break;
      }
    }

    // Use the parent's name as our trigger string, in case the parent changed
    // its name.  Only applies to subsidiary Commands.

    CommandPntr->UseParentNameAsTrigger ();
  }

  // Now that the Command has a stable name, see if it exists.

  CommandMap::iterator CommandIter;
  CommandIter = m_CommandMap.find (CommandPntr->Name ());
  if (CommandIter != m_CommandMap.end ())
  {
    Command *OldCommandPntr = CommandIter->second;

    if (OldCommandPntr == CommandPntr)
      return; // It's already added to our Map of Commands.  Do nothing.

    m_CommandMap.erase (CommandIter);

    // We don't handle recursive child deletion, so disconnect the children
    // before deleting.  They'll get cleaned up by RecalculateAncestry ().

    OldCommandPntr->AncestryReset ();
    delete OldCommandPntr; // Will deLooper it too.
  }

  // Finally make this Command offically active by adding it to the global map
  // of all Commands.

  m_CommandMap[CommandPntr->Name ()] = CommandPntr;
  CommandPntr->AddToLooper (this);

  // Recursively do the Command's children.  Presumably they're all subsidiary.

  CommandPntr->InternalAddChildren (AncestorName, Depth + 1);
}


/* Look for the Command with the given name (case insensitive).  Returns NULL
if not found. */

Command * ScriptOCronApp::InternalFindCommand (
  const char *CommandNamePntr) const
{
  CommandMap::const_iterator CommandIter;
  CommandIter = m_CommandMap.find (CommandNamePntr);
  if (CommandIter == m_CommandMap.end ())
    return NULL;
  return CommandIter->second;
}


/* Remove the given Command from the map of all Commands (but don't delete it).
Used by the Command object when it is renaming itself.  Disconnects the Command
from the Looper too.  Recursively removes all subsidiary children too from the
global map, leaving them attached to their parent Command via parent and
children pointers.  Disconnects child/parent links between removed Commands and
regular (non-subsidiary) child Commands (since regular children are independent
of the original removed Command and remain active). */

void ScriptOCronApp::InternalRemoveCommandGuts (Command *CommandPntr)
{
  CommandMap::iterator CommandIter;
  CommandIter = m_CommandMap.find (CommandPntr->Name ());
  if (CommandIter == m_CommandMap.end ())
  {
    char TempString [1024];
    snprintf (TempString, sizeof (TempString),
      "Internal fault, trying to remove Command \"%s\" which isn't in "
      "the global list",
      (CommandPntr == NULL) ? "NULL": CommandPntr->Name ());
    DisplayErrorMessage (TempString, B_BAD_INDEX, "InternalRemoveCommandGuts");
    return; // Not found, shouldn't happen.
  }

  Command *OldCommandPntr = CommandIter->second;
  if (OldCommandPntr != CommandPntr)
  {
    char TempString [1024];
    snprintf (TempString, sizeof (TempString),
      "Internal fault, while removing Command \"%s\", the global list "
      "has some other Command with the same name",
      (CommandPntr == NULL) ? "NULL": CommandPntr->Name ());
    DisplayErrorMessage (TempString, B_BAD_INDEX, "InternalRemoveCommandGuts");
    return; // Duplicate found, shouldn't happen.
  }

  m_CommandMap.erase (CommandIter);
  OldCommandPntr->RemoveFromLooper ();
  OldCommandPntr->InternalRemoveChildren (); // Recursively do children.
}


/* Load the Models from an archive stored in a BMessage, which could be from
the settings file or from the built-in Models resource.  Skips ones which have
a duplicate name.  Marks them as built-in (so they don't get saved later) if
requested. */

void ScriptOCronApp::LoadModels (BMessage *ArchivePntr, bool BuiltIn)
{
  type_code TypeFound = 0;
  int32 CountFound = 0;
  if (B_OK == ArchivePntr->GetInfo(g_ModelListName, &TypeFound, &CountFound) &&
  TypeFound == B_MESSAGE_TYPE && CountFound > 0)
  {
    int32 iModel;
    for (iModel = 0; iModel < CountFound; iModel++)
    {
      // Load the next candidate Model.

      BMessage *ModelMessagePntr = new BMessage;
      if (B_OK != ArchivePntr->FindMessage (
      g_ModelListName, iModel, ModelMessagePntr))
      {
        delete ModelMessagePntr;
        continue; // Skip bad ones.
      }

      // What is the name of the Model?

      const char *ModelNamePntr = NULL;
      if (B_OK != ModelMessagePntr->FindString (g_InternalName,
      &ModelNamePntr))
      {
        delete ModelMessagePntr;
        continue; // It doesn't have a name, probably not a Model.
      }

      // Make sure a Model with the same name does not already exist.  If
      // another does exist, leave it alone, since it may be one of the
      // built-in Models, and we don't want the user to override them
      // (leads to support confusion).

      ModelMap::iterator ModelIter = m_ModelMap.find (ModelNamePntr);
      if (ModelIter != m_ModelMap.end ())
      {
        delete ModelMessagePntr;
        continue;
      }

      // Mark built-in ones with a flag so they don't get saved.  Flag only
      // exists in memory, will never be in a saved file (or leak into a
      // resource), so we don't have to worry about adding it too many times.

      if (BuiltIn)
        ModelMessagePntr->AddBool (g_NameBuiltIn, true);

      // Reload the name string.  Modifying the message by adding the bool for
      // BuiltIn may have changed its address (happens in Haiku, not BeOS).

      ModelMessagePntr->FindString (g_InternalName, &ModelNamePntr);

      // Have a valid enough Model, add it to the collection.

      m_ModelMap [ModelNamePntr] = ModelMessagePntr;
    }
  }
  m_SerialWhenModelsLastChanged = m_ChangeSerial;
}


/* Either load the Commands, Models and program settings (DoLoad is TRUE) from
the main settings file and optional add-on files or (DoLoad is FALSE) write the
user's Commands, Models and program settings to the main settings file.

The files all contain a flattened BMessage containing the various program
settings and Commands and Models in archived object format.

If the main settings file (and parent directories) doesn't exist then it will
be created when saving.  If it doesn't exist when loading, the program settings
will be set to default values.

Extra Models (usually puchased separately) are stored in files inside a
directory named ScriptOCron inside one of the add-on directories.  The models
are loaded in this order, with later duplicate Models being skipped: built-in,
user add-on non-packaged, user add-on packaged, system add-on non-packaged,
system add-on packaged, user settings file.  For BeOS those directories are
/boot/beos/system/add-ons/ScriptOCron/ & /boot/home/config/add-ons/ScriptOCron/
while for Haiku OS similar directories are used. */

status_t ScriptOCronApp::LoadSaveSettings (bool DoLoad)
{
  status_t    ErrorCode;
  BDirectory  SettingsDirectory;
  char        TempString [PATH_MAX + 1024];

  // Preset things to default values if loading, in case of an error or it's an
  // older version of the settings file which doesn't have every field defined.

  if (DoLoad)
  {
    DeleteAllModels ();
    DeleteAllCommands ();
    DefaultSettings (); // Also loads the built-in models from a resource.
  }

  // Look for our settings directory.  If it doesn't exist, try to create it.

  ErrorCode = SettingsDirectory.SetTo (m_SettingsDirectoryPath.Path ());
  if (ErrorCode != B_OK)
  {
    ErrorCode = create_directory (
      m_SettingsDirectoryPath.Path (), 0750 /* rwxr-x--- */);
    if (ErrorCode == B_OK)
      ErrorCode = SettingsDirectory.SetTo (m_SettingsDirectoryPath.Path ());
    if (ErrorCode != B_OK)
    {
      sprintf (TempString, "Can't create settings directory \"%s\"",
        m_SettingsDirectoryPath.Path ());
      goto ErrorExit;
    }
  }

  // Load the auxiliary Models.  Searches all files in a ScriptOCron
  // subdirectory of each of the user and system AddOn directories.

  if (DoLoad)
  {
    // Order of subdirectories to search gives preference to the user, and
    // non-packaged within that.

#if __HAIKU__
    BPath ModelPaths [8];
    find_directory (B_USER_NONPACKAGED_DATA_DIRECTORY, &ModelPaths[0]);
    find_directory (B_USER_DATA_DIRECTORY, &ModelPaths[1]);
    find_directory (B_SYSTEM_NONPACKAGED_DATA_DIRECTORY, &ModelPaths[2]);
    find_directory (B_SYSTEM_DATA_DIRECTORY, &ModelPaths[3]);
    find_directory (B_USER_NONPACKAGED_ADDONS_DIRECTORY, &ModelPaths[4]);
    find_directory (B_USER_ADDONS_DIRECTORY, &ModelPaths[5]);
    find_directory (B_SYSTEM_NONPACKAGED_ADDONS_DIRECTORY, &ModelPaths[6]);
    find_directory (B_SYSTEM_ADDONS_DIRECTORY, &ModelPaths[7]);
#else
    BPath ModelPaths [2];
    find_directory (B_USER_ADDONS_DIRECTORY, &ModelPaths[0]);
    find_directory (B_BEOS_ADDONS_DIRECTORY, &ModelPaths[1]);
#endif

    for (unsigned int i = 0;
    i < sizeof (ModelPaths) / sizeof (ModelPaths[0]); i++)
    {
      BPath *PathPntr = ModelPaths + i;
      if (PathPntr->InitCheck () != B_OK)
        continue;
      ErrorCode = PathPntr->Append (g_AddOnSubDirectory);
      if (ErrorCode != B_OK)
        continue;
      BDirectory ModelsSubDir (PathPntr->Path ());
      if (ModelsSubDir.InitCheck () != B_OK)
        continue;

      // Scan the ScriptOCron subdirectory for model files.

      while (true)
      {
        BEntry EntryOfModelFile;
        ErrorCode = ModelsSubDir.GetNextEntry (&EntryOfModelFile,
          true /* traverse symbolic links */);
        if (ErrorCode != B_OK)
          break; // End of list of files, or a real error.

        // Because of the traversal, it may be in a different directory.

        BDirectory DirOfModelFile;
        ErrorCode = EntryOfModelFile.GetParent (&DirOfModelFile);
        if (ErrorCode != B_OK || DirOfModelFile.InitCheck () != B_OK)
          continue;
        ErrorCode = EntryOfModelFile.GetName (TempString);
        if (ErrorCode != B_OK)
          continue;

        ErrorCode = LoadSaveSettingsFile (DoLoad, true /* JustModels */,
          DirOfModelFile, TempString);
        // Failure is an option, happens for non-model files.
      }
    }
  }

  // Do the main settings file.

  if (SettingsDirectory.InitCheck () == B_OK)
  {
    BString TemporarySettingsFileName (g_SettingsFileName);
    if (!DoLoad)
    {
      // When saving, save to a temporary file name in case we have trouble
      // (Haiku sometimes mangles files when overwriting), then rename the temp
      // file to be the more permanent settings file.

      TemporarySettingsFileName.Append (".new");
    }

    ErrorCode = LoadSaveSettingsFile (DoLoad, false /* JustModels */,
      SettingsDirectory, TemporarySettingsFileName.String ());
    if (DoLoad)
      RecalculateAncestry (); // Do before error test, in case partial load.

    if (DoLoad && ErrorCode == B_ENTRY_NOT_FOUND)
    { // No error message for expected error that happens on the first run.
      g_FirstTimeAppRun = true;
    }
    else if (ErrorCode != B_OK)
    {
      sprintf (TempString, "Can't %s settings file \"%s/%s\"",
        DoLoad ? "read" : "write",
        m_SettingsDirectoryPath.Path (), TemporarySettingsFileName.String ());
      goto ErrorExit;
    }
    else // Success
    {
      if (!DoLoad)
      {
        // Saving settings.  Rename the temporary settings file over the
        // permanent one.  If this is the first time in this run, save the
        // permanent one as a backup.

        static bool FirstTimeSave = true;

        if (FirstTimeSave)
        {
          // Rename the old settings to be a backup file.  Wipes out the
          // previous backup file.

          BString OldPath (m_SettingsDirectoryPath.Path ());
          OldPath.Append ("/");
          OldPath.Append (g_SettingsFileName);
          BString NewPath (OldPath);
          NewPath.Append (".backup");
          rename (OldPath.String (), NewPath.String ());
          FirstTimeSave = false;
        }

        // Rename the new settings file as the permanent settings.

        BString OldPath (m_SettingsDirectoryPath.Path ());
        OldPath.Append ("/");
        OldPath.Append (TemporarySettingsFileName);
        BString NewPath (m_SettingsDirectoryPath.Path ());
        NewPath.Append ("/");
        NewPath.Append (g_SettingsFileName);
        rename (OldPath.String (), NewPath.String ());
     }
    }
  }

  m_SerialWhenSettingsSaved = m_ChangeSerial;
  return B_OK;

ErrorExit: /* Error message in TempString, code in ErrorCode. */
  DisplayErrorMessage (TempString, ErrorCode, DoLoad ?
    "Loading Settings Error" : "Saving Settings Error");
  return ErrorCode;
}


/* Load or save a single settings file.  No error messages printed.  File is
closed before returning so it should be safe to move it etc. */

status_t ScriptOCronApp::LoadSaveSettingsFile (bool DoLoad, bool JustModels,
  BDirectory &Directory, const char *FileName)
{
  status_t    ErrorCode;
  const char *NamePntr;
  BFile       SettingsFile;
  BMessage    SettingsMessage;

  ErrorCode = SettingsFile.SetTo (&Directory, FileName,
    DoLoad ? B_READ_ONLY : B_READ_WRITE | B_CREATE_FILE | B_ERASE_FILE);
  if (ErrorCode != B_OK)
    return ErrorCode;

  if (DoLoad)
  {
    ErrorCode = SettingsMessage.Unflatten (&SettingsFile);
    if (ErrorCode != B_OK)
      return ErrorCode;
    if (SettingsMessage.what != g_SettingsWhatCode)
      return B_MISMATCHED_VALUES;
  }

  // Transfer the settings between the BMessage and our various global
  // variables.  For loading, if the setting isn't present, leave it at the
  // default value.  Note that loading and saving are intermingled here to make
  // code maintenance easier (less chance of forgetting to update it if load
  // and save were separate functions).  Will go through all the code here to
  // the end, even if an error happens.

  ErrorCode = B_OK; /* So that saving settings can record an error. */

#if INCLUDE_GUI_CODE
  NamePntr = "AdvancedMode";
  if (DoLoad)
  {
    if (!JustModels)
    {
      bool TempBool;
      if (SettingsMessage.FindBool (NamePntr, &TempBool) == B_OK)
        g_AdvancedMode = TempBool;
    }
  }
  else if (ErrorCode == B_OK)
    ErrorCode = SettingsMessage.AddBool (NamePntr, g_AdvancedMode);

  NamePntr = "FontScaleCommandNames";
  if (DoLoad)
  {
    if (!JustModels)
    {
      float TempFloat;
      if (SettingsMessage.FindFloat (NamePntr, &TempFloat) == B_OK)
        g_FontScaleCommandNames = TempFloat;
      if (g_FontScaleCommandNames < 0.5f)
        g_FontScaleCommandNames = 0.5f;
      if (g_FontScaleCommandNames > 4.0f)
        g_FontScaleCommandNames = 4.0f;
    }
  }
  else if (ErrorCode == B_OK)
    ErrorCode = SettingsMessage.AddFloat (NamePntr, g_FontScaleCommandNames);

  NamePntr = "FontScaleLogText";
  if (DoLoad)
  {
    if (!JustModels)
    {
      float TempFloat;
      if (SettingsMessage.FindFloat (NamePntr, &TempFloat) == B_OK)
        g_FontScaleLogText = TempFloat;
      if (g_FontScaleLogText < 0.5f)
        g_FontScaleLogText = 0.5f;
      if (g_FontScaleLogText > 4.0f)
        g_FontScaleLogText = 4.0f;
    }
  }
  else if (ErrorCode == B_OK)
    ErrorCode = SettingsMessage.AddFloat (NamePntr, g_FontScaleLogText);

  NamePntr = "InitialWindowFrame";
  if (DoLoad)
  {
    if (!JustModels)
    {
      BRect TempRect;
      if (SettingsMessage.FindRect (NamePntr, &TempRect) == B_OK)
        g_InitialWindowFrame = TempRect;
    }
  }
  else if (ErrorCode == B_OK)
    ErrorCode = SettingsMessage.AddRect (NamePntr, g_InitialWindowFrame);

  NamePntr = "ServerMode";
  if (DoLoad)
  {
    if (!JustModels)
    {
      bool TempBool;
      if (SettingsMessage.FindBool (NamePntr, &TempBool) == B_OK)
        g_ServerMode = TempBool;
    }
  }
  else if (ErrorCode == B_OK)
    ErrorCode = SettingsMessage.AddBool (NamePntr, g_ServerMode);

  NamePntr = "SimpleWindowTitle";
  if (DoLoad)
  {
    if (!JustModels)
    {
      bool TempBool;
      if (SettingsMessage.FindBool (NamePntr, &TempBool) == B_OK)
        g_SimpleWindowTitle = TempBool;
    }
  }
  else if (ErrorCode == B_OK)
    ErrorCode = SettingsMessage.AddBool (NamePntr, g_SimpleWindowTitle);
#endif // INCLUDE_GUI_CODE

  NamePntr = "UserSpecifiedTemporaryDirectory";
  if (DoLoad)
  {
    if (!JustModels)
    {
      const char *TempDirName = NULL;
      if (SettingsMessage.FindString (NamePntr, &TempDirName) == B_OK &&
      TempDirName != NULL)
      {
        m_UserSpecifiedTemporaryDirectory = TempDirName;
        m_UserSpecifiedTemporaryDirectorySerial = m_ChangeSerial;
      }
    }
  }
  else if (ErrorCode == B_OK)
    ErrorCode = SettingsMessage.AddString (
      NamePntr, m_UserSpecifiedTemporaryDirectory);

  // Load/save the Commands, which are stored as archived BMessages.

  NamePntr = g_CommandListName;
  if (DoLoad)
  {
    if (!JustModels)
    {
      type_code TypeFound = 0;
      int32 CountFound = 0;
      ErrorCode = SettingsMessage.GetInfo (NamePntr, &TypeFound, &CountFound);
      if (ErrorCode == B_OK && TypeFound == B_MESSAGE_TYPE && CountFound > 0)
      {
        int32 iCommand;
        for (iCommand = 0; iCommand < CountFound; iCommand++)
        {
          BMessage ArchiveMsg;
          ErrorCode =
            SettingsMessage.FindMessage (NamePntr, iCommand, &ArchiveMsg);
          if (ErrorCode != B_OK)
            continue; // Skip bad ones.
          BArchivable *DearchivedPntr = instantiate_object (&ArchiveMsg);
          if (DearchivedPntr == NULL)
            continue; // Unable to find C++ class of object to rebuild it, etc.

          Command *CommandPntr = cast_as (DearchivedPntr, Command);
          if (CommandPntr != NULL)
          {
            // Add the new Command (and children if any, though there shouldn't
            // be any) to the global list, but don't do RecalculateAncestry(),
            // save that for after they're all loaded.

            InternalAddCommandGuts (CommandPntr,
              NULL /* Don't do ancestor name renaming when loading */, 0);

            // Newly created Commands start in the editing state, which we
            // don't want after loading.
            CommandPntr->SetStateRecursive (CS_READY_TO_RUN);
          }
          else // Some other kind of object, someone hacking the settings?
            delete DearchivedPntr;
        }
      }
      ChangeHappened (false /* Settings file doesn't need saving */);
    }
  }
  else if (ErrorCode == B_OK) // Saving the Command list.
  {
    CommandMap::iterator CommandIter;
    for (CommandIter = m_CommandMap.begin ();
    CommandIter != m_CommandMap.end (); CommandIter++)
    {
      // Save each Command separately, child subsidiary Commands are stored by
      // themselves and will be relinked by name upon loading.

      BMessage ArchiveMsg (B_ARCHIVED_OBJECT);

      Command *CommandPntr = CommandIter->second;
      ErrorCode = CommandPntr->Archive (&ArchiveMsg, false /* Deep flag */);
      if (ErrorCode != B_OK)
        break;
      ErrorCode = SettingsMessage.AddMessage (NamePntr, &ArchiveMsg);
      if (ErrorCode != B_OK)
        break;
    }
  }

  // Load/save the user created Models (don't save the built-in ones), also
  // stored as archived objects in a BMessage.

  NamePntr = g_ModelListName;
  if (DoLoad)
  {
    LoadModels (&SettingsMessage, JustModels /* Mark as BuiltIn. */);
  }
  else if (ErrorCode == B_OK) // Saving the Model list.
  {
    ModelMap::iterator ModelIter;
    for (ModelIter = m_ModelMap.begin ();
    ModelIter != m_ModelMap.end (); ModelIter++)
    {
      BMessage *ModelPntr = ModelIter->second;
      bool IsBuiltIn = false;
      if (B_OK != ModelPntr->FindBool (g_NameBuiltIn, &IsBuiltIn))
        IsBuiltIn = false;
      if (!IsBuiltIn)
      {
        ErrorCode = SettingsMessage.AddMessage (NamePntr, ModelPntr);
        if (ErrorCode != B_OK)
          break;
      }
    }
  }

  // Loading/saving from/to the temporary BMessage is complete.  Did it work?

  if (ErrorCode != B_OK)
  {
    if (DoLoad)
      ErrorCode = B_OK; // Errors just skip data, no real error.
    else // Error while building the BMessage.  Nothing to save.
      return ErrorCode;
  }

  // Save the settings BMessage to the settings file.

  if (!DoLoad)
  {
    SettingsMessage.what = g_SettingsWhatCode;
    ErrorCode = SettingsMessage.Flatten (&SettingsFile);
  }

  return ErrorCode;
}


/* Process a message directed to the Application level. */

void ScriptOCronApp::MessageReceived (BMessage *MessagePntr)
{
  const char           *ArgumentString = NULL;
  status_t              ErrorCode = B_BAD_SCRIPT_SYNTAX;
  const char           *PropertyName;
  BMessage              ReplyMessage (B_REPLY);
  int32                 SpecifierIndex;
  int32                 SpecifierKind;
  BMessage              SpecifierMessage;
  char                  TempString [PATH_MAX + 1024];

  /* See if it is a scripting message that applies to one of the operations
  this program supports.  Pass on other kinds of messages to the inherited
  parent MessageReceived () function. */

  switch (MessagePntr->what)
  {
    case B_GET_PROPERTY:
    case B_SET_PROPERTY:
    case B_COUNT_PROPERTIES:
    case B_CREATE_PROPERTY:
    case B_DELETE_PROPERTY:
    case B_EXECUTE_PROPERTY:
      // Look at specifier[0] to find out what the scripting does.
      if (B_OK == MessagePntr->SetCurrentSpecifier (0))
        ErrorCode = MessagePntr->GetCurrentSpecifier (&SpecifierIndex,
        &SpecifierMessage, &SpecifierKind, &PropertyName);
      else // No specifiers, no default target for our BApplication.
        ErrorCode = B_ERROR;

      if (ErrorCode == B_OK &&
      IsItOurScriptingCommand (g_AppScriptingPropertyList,
      PropertyName, SpecifierKind))
        break; // It's one of ours, do it.
      // Else fall through, and punt the message to the parent class.

    default: // Not one of our messages, let the parent handle it.
      BApplication::MessageReceived (MessagePntr);
      return;
  }

  snprintf (TempString, sizeof (TempString),
    "Scripting operation code %c%c%c%c (get, set, count, etc) or access type "
    "%d (direct=1, index=2, reverse index=3, range=4, reverse range=5, "
    "name=6, ID=7) is unsupported for property \"%s\" of the "
    PROGRAM_NAME_WORD " Application level object",
    (char) (MessagePntr->what >> 24),
    (char) (MessagePntr->what >> 16),
    (char) (MessagePntr->what >>  8),
    (char) (MessagePntr->what >>  0),
    (int) SpecifierKind, PropertyName);
  ErrorCode = B_BAD_SCRIPT_SYNTAX;

  // Handle the Serial property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_SERIAL]) == 0)
  {
    // Get the current serial number.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, m_ChangeSerial);
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Serial property.
  }

  // Handle the Command property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_COMMAND]) == 0)
  {
    // Count the number of Commands.

    if (MessagePntr->what == B_COUNT_PROPERTIES &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, m_CommandMap.size ());
      goto SuccessExit;
    }

    // Create a new command, optionally with a name and using a Model.

    if (MessagePntr->what == B_CREATE_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK)
        ArgumentString = g_NameUnnamed;

      if (DoesNamedCommandExist (ArgumentString))
      {
        ErrorCode = B_NAME_IN_USE;
        snprintf (TempString, sizeof (TempString),
          "Can't create a Command named \"%s\" because it already exists",
          ArgumentString);
        goto ErrorExit;
      }

      const char *ModelNamePntr = NULL;
      ErrorCode = MessagePntr->FindString (g_ModelName, &ModelNamePntr);
      if (ErrorCode != B_OK)
        ModelNamePntr = NULL;

      Command *NewCommandPntr = NULL;
      if (ModelNamePntr == NULL)
        NewCommandPntr = new Command (ArgumentString);
      else // Make the new Command from a saved Model.
      {
        ModelMap::iterator ModelIter = m_ModelMap.find (ModelNamePntr);
        if (ModelIter == m_ModelMap.end ())
        {
          ErrorCode = B_NAME_NOT_FOUND;
          snprintf (TempString, sizeof (TempString),
            "Can't create a Command named \"%s\" because the Model "
            "named \"%s\" doesn't exist", ArgumentString, ModelNamePntr);
          goto ErrorExit;
        }

        BArchivable *DearchivedPntr = instantiate_object (ModelIter->second);
        if (DearchivedPntr == NULL ||
        (NewCommandPntr = cast_as (DearchivedPntr, Command)) == NULL)
        {
          delete DearchivedPntr; // NULL or wrong kind of object.
          snprintf (TempString, sizeof (TempString), "Something went wrong "
            "when instantiating a Command named \"%s\" from a Model named "
            "\"%s\"", ArgumentString, ModelNamePntr);
          ErrorCode = B_ERROR;
          goto ErrorExit;
        }

        NewCommandPntr->SetName (ArgumentString);
      }

      InternalAddCommand (NewCommandPntr);
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on a Command property.
  }

  // Handle the Model property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_MODEL]) == 0)
  {
    // Get a list of all Model names, returned as an array of strings.  Also
    // include the corresponding input and output data types, and finally the
    // current change serial number.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ModelMap::iterator ModelIter;
      for (ModelIter = m_ModelMap.begin ();
      ModelIter != m_ModelMap.end (); ModelIter++)
      {
        const char *ModelName = ModelIter->first;
        BMessage *ModelAsBMessagePntr = ModelIter->second;
        ReplyMessage.AddString (g_ResultName, ModelName);

        // Look for the input and output data type names, in the UserData.
        // Will be an empty string if they aren't found.

        int32 CountFound = 0;
        BString DataTypeIn;
        BString DataTypeOut;
        type_code TypeFound = 0;

        ErrorCode = ModelAsBMessagePntr->GetInfo (g_UserDataName,
          &TypeFound, &CountFound);
        if (ErrorCode == B_OK && TypeFound == B_MESSAGE_TYPE && CountFound > 0)
        {
          int32 iUserData;
          for (iUserData = 0; iUserData < CountFound; iUserData++)
          {
            BMessage UserDataMessage;
            ErrorCode = ModelAsBMessagePntr->
              FindMessage (g_UserDataName, iUserData, &UserDataMessage);
            if (ErrorCode != B_OK)
              continue; // Skip bad ones.
            const char *KeyPntr = NULL;
            if (UserDataMessage.FindString (g_NameKey, &KeyPntr) != B_OK)
              continue; // Always need a key.
            const char *ValuePntr = NULL;
            if (UserDataMessage.FindString (g_ResultName, &ValuePntr) != B_OK)
              continue;
            if (ValuePntr == NULL)
              continue; // Ignore missing strings.

            if (strcasecmp (KeyPntr, g_DataTypeInName) == 0)
              DataTypeIn = ValuePntr;
            else if (strcasecmp (KeyPntr, g_DataTypeOutName) == 0)
              DataTypeOut = ValuePntr;
          }
        }

        ReplyMessage.AddString (g_DataTypeInName, DataTypeIn);
        ReplyMessage.AddString (g_DataTypeOutName, DataTypeOut);
      }

      // Finally add that serial number.

      ReplyMessage.AddInt32 (g_SerialName, m_SerialWhenModelsLastChanged);
      goto SuccessExit;
    }

    // Just get the change serial number for the Models.

    if (MessagePntr->what == B_COUNT_PROPERTIES &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, m_SerialWhenModelsLastChanged);
      goto SuccessExit;
    }

    // Create a new Model from a named Command.

    if (MessagePntr->what == B_CREATE_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ArgumentString = NULL;
      ErrorCode = MessagePntr->FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK || ArgumentString == NULL ||
      ArgumentString[0] == 0)
      {
        snprintf (TempString, sizeof (TempString),
        "Need an additional \"%s\" string to specify the name of the "
        "Command to use for the new Model", g_NameName);
        goto ErrorExit;
      }

      CommandMap::iterator CommandIter;
      CommandIter = m_CommandMap.find (ArgumentString);
      if (CommandIter == m_CommandMap.end ())
      {
        ErrorCode = B_NAME_NOT_FOUND;
        snprintf (TempString, sizeof (TempString), "Unable to create a Model "
          "for Command named \"%s\" because no Command with that name exists",
          ArgumentString);
        goto ErrorExit;
      }

      ModelMap::iterator ModelIter;
      ModelIter = m_ModelMap.find (ArgumentString);
      if (ModelIter != m_ModelMap.end ())
      {
        ErrorCode = B_NAME_IN_USE;
        snprintf (TempString, sizeof (TempString), "Unable to create a Model "
          "for Command named \"%s\" because a Model with that name already "
          "exists.  Delete it first if you want to replace it",
          ArgumentString);
        goto ErrorExit;
      }

      Command *CommandPntr = CommandIter->second;
      BMessage *ModelMessagePntr = new BMessage (B_ARCHIVED_OBJECT);
      ErrorCode = CommandPntr->Archive (ModelMessagePntr, true /* deep */);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Failed to make an Archive copy of Command \"%s\"", ArgumentString);
        delete ModelMessagePntr;
        goto ErrorExit;
      }

      // Set up the "Original Model" UserData inside the Model to use the new
      // name of the Model, so that you can tell where Commands cloned from it
      // originally came from.  Add a version stamp there too, so you can tell
      // if a Command cloned from it needs updating.  First delete any existing
      // UserData["Original Model"] entries.

      int32 iUserData;
      for (iUserData = 0; true; iUserData++)
      {
        BMessage UserDataMessage;
        ErrorCode = ModelMessagePntr->
          FindMessage (g_UserDataName, iUserData, &UserDataMessage);
        if (ErrorCode != B_OK)
          break; // Probably hit the end of the array of UserData.

        const char *KeyPntr = NULL;
        if (UserDataMessage.FindString (g_NameKey, &KeyPntr) != B_OK)
          continue; // Always need a key.
        if (KeyPntr == NULL || strcmp (KeyPntr, g_KeyOriginalModel) != 0)
          continue; // Some other UserData, not about the Model, skip it.

        if (B_OK == ModelMessagePntr->RemoveData (g_UserDataName, iUserData))
          iUserData--; // Successfully deleted it, adjust next item index.
      }

      // Make the new "Original Model" UserData, with name and date values.

      BMessage OriginalModelMessage (B_SIMPLE_DATA);
      OriginalModelMessage.AddString (g_NameKey, g_KeyOriginalModel);
      OriginalModelMessage.AddString (g_ResultName, CommandPntr->Name ());
      BString FormattedDate;
      ExpandDateFormat (FormattedDate, "%c");
      OriginalModelMessage.AddString (g_ResultName, FormattedDate);
      ModelMessagePntr->AddMessage (g_UserDataName, &OriginalModelMessage);

      // Save the new Model in our global collection of Models.

      const char *PersistentNameString;
      ErrorCode = ModelMessagePntr->FindString (g_InternalName,
        &PersistentNameString); // Don't modify ModelMessagePntr past here!
      if (ErrorCode != B_OK ||
      strcmp (PersistentNameString, ArgumentString) != 0)
      {
        snprintf (TempString, sizeof (TempString),
          "Something wrong with internal name in Archived copy of "
          "Command \"%s\".  Doesn't exist or doesn't match Command name",
          ArgumentString);
        delete ModelMessagePntr;
        goto ErrorExit;
      }

      m_ModelMap [PersistentNameString] = ModelMessagePntr;

      ChangeHappened ();
      m_SerialWhenModelsLastChanged = m_ChangeSerial;
      goto SuccessExit;
    }

    // Delete a Model, specified by a name.

    if (MessagePntr->what == B_DELETE_PROPERTY &&
    SpecifierKind == B_NAME_SPECIFIER)
    {
      ArgumentString = NULL;
      ErrorCode = SpecifierMessage.FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK || ArgumentString == NULL ||
      ArgumentString[0] == 0)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify the name of the Model to be deleted");
        goto ErrorExit;
      }

      ModelMap::iterator ModelIter;
      ModelIter = m_ModelMap.find (ArgumentString);
      if (ModelIter == m_ModelMap.end ())
      {
        ErrorCode = B_NAME_NOT_FOUND;
        snprintf (TempString, sizeof (TempString),
          "Unable to delete Model named \"%s\" because it doesn't exist",
          ArgumentString);
        goto ErrorExit;
      }

      BMessage *DeletedModelPntr = ModelIter->second;
      m_ModelMap.erase (ModelIter);
      delete DeletedModelPntr;

      ChangeHappened ();
      m_SerialWhenModelsLastChanged = m_ChangeSerial;
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on a Model property.
  }

  // Handle the UserTempDir property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_USER_TEMP_DIR]) == 0)
  {
    // Get the current user specified temporary directory.  To see the actual
    // directory used, use GetStockNamesAndPaths() or the related PN_PATHS
    // scripting property.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName,
        m_UserSpecifiedTemporaryDirectory.String ());
      ReplyMessage.AddInt32 (g_SerialName,
        m_UserSpecifiedTemporaryDirectorySerial);
      goto SuccessExit;
    }

    // Set or delete the temporary directory.  An empty string (which you can't
    // type into "Hey") or delete operation means use the system default
    // temporary directory.

    if (SpecifierKind == B_DIRECT_SPECIFIER &&
    (MessagePntr->what == B_SET_PROPERTY ||
    MessagePntr->what == B_DELETE_PROPERTY))
    {
      const char *NewName = g_EmptyString;
      if (MessagePntr->what == B_SET_PROPERTY)
      {
        ErrorCode = MessagePntr->FindString (g_DataName, &NewName);
        if (ErrorCode != B_OK)
        {
          snprintf (TempString, sizeof (TempString),
            "Need to specify a \"%s\" string value to set the \"%s\" "
            "property.",
            g_DataName, PropertyName);
          goto ErrorExit;
        }
        if (NewName [0] != '/' && NewName [0] != 0 /* Empty string OK too */)
        {
          snprintf (TempString, sizeof (TempString),
            "The \"%s\" property needs to start with a slash \"/\" since it "
            "has to be an absolute path.  You specified \"%s\"",
            PropertyName, NewName);
          ErrorCode = B_BAD_VALUE;
          goto ErrorExit;
        }
      }

      // Make sure that nothing is currently running.

      CommandMap::iterator CommandIter;
      for (CommandIter = m_CommandMap.begin ();
      CommandIter != m_CommandMap.end (); CommandIter++)
      {
        Command *CommandPntr = CommandIter->second;
        if (CommandPntr->IsRunningOrWaiting ())
        {
          snprintf (TempString, sizeof (TempString),
            "Can't change \"%s\" to \"%s\" while Command \"%s\" "
            "is currently running, try again later",
            PropertyName, NewName, CommandPntr->Name ());
          ErrorCode = B_BUSY;
          goto ErrorExit;
        }
      }

      // Use the specified directory, and remove trailing slashes, which cause
      // BPath to fail.  Note use of ByteAt which returns 0 for out of bounds
      // string accesses, in case of an empty string.

      m_UserSpecifiedTemporaryDirectory = NewName;

      while (m_UserSpecifiedTemporaryDirectory.ByteAt (
      m_UserSpecifiedTemporaryDirectory.Length () - 1) == '/')
      {
        m_UserSpecifiedTemporaryDirectory.Remove (
          m_UserSpecifiedTemporaryDirectory.Length () - 1, 1);
      }

      ChangeHappened ();
      m_UserSpecifiedTemporaryDirectorySerial = m_ChangeSerial;
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on UserTempDir property.
  }

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return;

SuccessExit: /* Success. */
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
}


/* Do periodic polling, mostly checking for output from the running Commands.
Since quitting stops the program before the results of a script command are
received, we use a time delay to do the quit.  Also, we have a quit countdown
which can be interrupted by an incoming scripting message in case one client
tells us to quit while another one is still using us. */

void ScriptOCronApp::Pulse ()
{
  if (g_QuitCountdown == 0)
  {
    cerr << PROGRAM_NAME_WORD " quitting time.\n";

    // There's a "_shutdown_" boolean field set to TRUE in the original
    // B_QUIT_REQUESTED message if the system is shutting down, it's in BeOS
    // too.

    bool OSIsShuttingDown = false;
    if (m_QuitRequestedStashedMessage != NULL &&
    m_QuitRequestedStashedMessage->FindBool ("_shutdown_", &OSIsShuttingDown)
    != B_OK)
      OSIsShuttingDown = false;

    // Quitting time.  First try closing windows, and ask each window if it's
    // okay to quit.  QuitRequested returns TRUE if all windows closed.  Skip
    // if shutting down, windows will be closed in app destructor instead.

    bool QuitOK = true;
    if (!OSIsShuttingDown)
      QuitOK = BApplication::QuitRequested ();

    // Reply to a delayed quit request (usually from the system when it is
    // trying to shut down).

    if (m_QuitRequestedStashedMessage != NULL)
    {
      BMessage ReplyMessage (B_REPLY);
      ReplyMessage.AddBool (g_ResultName, QuitOK);
      // Haiku's BApplication quit message handling code returns a "thread"
      // field with the BApplication's thread identifier, presumably the
      // BRoster requesting the shutdown needs it to figure out which pending
      // shutdown has failed or succeeded.
      int32 ThreadID = Thread ();
      ReplyMessage.AddInt32 ("thread", ThreadID);
      m_QuitRequestedStashedMessage->SendReply (&ReplyMessage,
        (BHandler *) NULL /* where to send the reply to the reply */,
        500000 /* send timeout */);

      delete m_QuitRequestedStashedMessage;
      m_QuitRequestedStashedMessage = NULL;
    }

    if (QuitOK)
    {
      // Shut down our message processing loop.  Last few messages will still
      // be processed (after this function returns), then Run() will return.

      Quit ();
    }
    else // Go back to normal running.
      g_QuitCountdown = -1;
  }
  else if (g_QuitCountdown > 0)
  {
    cerr << PROGRAM_NAME_WORD " quitting in " << g_QuitCountdown << ".\n";
    g_QuitCountdown--;
  }

  // Poll Commands.  Collects and stores output of running commands and checks
  // for commands that have finished and ones which should start running.  We
  // can assume no Command will be deleted during this polling (any delete
  // requests would have to go through the message system, which won't do
  // anything until this function has returned).

  time_t CurrentTime;
  time (&CurrentTime);
  struct tm CurrentBrokenUpTime;
  localtime_r (&CurrentTime, &CurrentBrokenUpTime);

  CommandMap::iterator CommandIter;
  for (CommandIter = m_CommandMap.begin ();
  CommandIter != m_CommandMap.end (); CommandIter++)
  {
    Command *CommandPntr = CommandIter->second;
    CommandPntr->PollAndUpdate (&CurrentBrokenUpTime);
  }

  // Save settings once in a while.  Don't want to save it on every change
  // while the user is editing things.

  if (g_SettingsNeedSaving && system_time () - m_TimeSettingsWereLastSaved >
  60 * 1000000 /* microsecond units */)
  {
    LoadSaveSettings (false /* DoLoad */);
    g_SettingsNeedSaving = false;
    m_TimeSettingsWereLastSaved = system_time ();
  }
}


void ScriptOCronApp::ReadyToRun ()
{
#if INCLUDE_GUI_CODE
  SetSizesAndOpenWindow ();
#endif

  SetPulseRate (200000); // Do after GUI windows opened, to get our rate.
}


/* Set the ancestry fields in all Commands to reflect their current
relationships.  Will disconnect circular links in subsidiary Commands (so that
the subsidiary Commands always form a tree under a regular Command).  Will
delete orphaned subsidiary Commands too, so watch out if you call this from a
Command's method, it may get deleted! */

void ScriptOCronApp::RecalculateAncestry ()
{
  CommandMap::iterator Iter;
  CommandMap::iterator NextIter;

  // Reset all the Commands to empty child lists and no parent.

  for (Iter = m_CommandMap.begin (); Iter != m_CommandMap.end (); Iter++)
    Iter->second->AncestryReset ();

  // Fill in the parent and child lists.

  for (Iter = m_CommandMap.begin (); Iter != m_CommandMap.end (); Iter++)
    Iter->second->AncestryUpdateParent ();

  // Determine the ancestor for each subsidiary Command.  Note that sometimes
  // we restart the loop from the beginning after deleting a bad Command, while
  // maintaining valid parent and child links so we don't need to recompute
  // those.

  for (Iter = m_CommandMap.begin (); Iter != m_CommandMap.end ();
  Iter = NextIter)
  {
    NextIter = Iter;
    NextIter++;
    Command *CommandPntr = Iter->second;

    if (!CommandPntr->IsSubsidiary ())
      continue; // Regular Commands don't have ancestors.

    if (!CommandPntr->AncestryUpdateAncestry (m_CommandMap.size ()))
    {
      // Failed to find an ancestor, perhaps because the named parent doesn't
      // exist any more, or maybe a circular loop of subsidiary Commands was
      // detected.  Related Commands will have their parent/child lists updated
      // by AncestryUpdateAncestry() so we just need to delete this Command and
      // start over without having to reset parents and children.

      char TempString [200];
      snprintf (TempString, sizeof (TempString),
        "Deleting unexpectedly orphaned subsidiary Command named \"%s\"",
        CommandPntr->Name ());
      DisplayErrorMessage (TempString, B_MISMATCHED_VALUES,
        "RecalculateAncestry");

      // Can't call general purpose InternalRemoveCommand, it calls us, so do
      // the equivalent operation.

      InternalRemoveCommandGuts (CommandPntr);

      delete CommandPntr;
      ChangeHappened ();

      NextIter = m_CommandMap.begin ();
    }
  }
}


/* A scripting command is looking for some object to execute it. */

BHandler * ScriptOCronApp::ResolveSpecifier (
  BMessage *MessagePntr,
  int32 Index,
  BMessage *SpecifierMsgPntr,
  int32 SpecificationKind,
  const char *PropertyNamePntr)
{
  status_t ErrorCode = B_OK;
  char TempString [PATH_MAX + 1024];

#if 0
printf ("Resolving Specifier in ScriptOCronApp, Message is: ");
MessagePntr->PrintToStream();
printf ("Specifier index %d, Specifier kind %d, Property \"%s\"\n",
  (int) Index, (int) SpecificationKind, PropertyNamePntr);
#endif

  /* See if it is one of our sub-object selection properties. */

  if (strcasecmp (PropertyNamePntr, g_PropertyNames[PN_COMMAND]) == 0)
  {
    // Looking for a named Command?

    if (SpecificationKind == B_NAME_SPECIFIER)
    {
      const char *NamePntr = NULL;
      ErrorCode = SpecifierMsgPntr->FindString (g_NameName, &NamePntr);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString),
          "\"%s\" string field is missing in the scripting specifier for "
          "the \"%s\" property", g_NameName, PropertyNamePntr);
        goto ErrorExit;
      }

      CommandMap::iterator CommandIter;
      CommandIter = m_CommandMap.find (NamePntr);
      if (CommandIter != m_CommandMap.end ())
      {
        // Found the named Command, pass on the rest of the scripting command
        // to the Command object.

        MessagePntr->PopSpecifier ();
        return CommandIter->second;
      }
      ErrorCode = B_NAME_NOT_FOUND;
      snprintf (TempString, sizeof (TempString),
        "Property \"%s\" sub-object specified by name \"%s\" not found",
        PropertyNamePntr, NamePntr);
      goto ErrorExit;
    }
    else if (SpecificationKind == B_INDEX_SPECIFIER)
    {
      // Look for the numbered (indexed) Command, the slow way.

      int32 DesiredIndex = 0;
      ErrorCode = SpecifierMsgPntr->FindInt32 (g_NameIndex, &DesiredIndex);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field is "
          "missing in the scripting specifier for the \"%s\" property",
          g_NameIndex, PropertyNamePntr);
        goto ErrorExit;
      }

      if (DesiredIndex < 0 || DesiredIndex >= (int) m_CommandMap.size ())
      {
        ErrorCode = B_BAD_INDEX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field for "
          "property \"%s\" is %d, which is not between 0 and %d",
          g_NameIndex, PropertyNamePntr, (int) DesiredIndex,
          (int) m_CommandMap.size() - 1);
        goto ErrorExit;
      }

      CommandMap::iterator CommandIter;
      CommandIter = m_CommandMap.begin ();
      int32 i = 0;
      while (i < DesiredIndex && CommandIter != m_CommandMap.end ())
      {
        CommandIter++;
        i++;
      }
      if (CommandIter != m_CommandMap.end ())
      {
        MessagePntr->PopSpecifier ();
        return CommandIter->second;
      }
      ErrorCode = B_ERROR;
      strcpy (TempString, "ScriptOCronApp::ResolveSpecifier shouldn't happen");
      goto ErrorExit;
    }
  }

  if (0 == Index && IsItOurScriptingCommand (g_AppScriptingPropertyList,
  PropertyNamePntr, SpecificationKind))
  {
    return this; /* It's one of our properties, self will handle message. */
  }

  /* Handle an unrecognized scripting command, let the parent figure it out. */

  return BApplication::ResolveSpecifier (
    MessagePntr, Index, SpecifierMsgPntr, SpecificationKind, PropertyNamePntr);

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  BMessage ReplyMessage (/* B_REPLY or maybe */ B_MESSAGE_NOT_UNDERSTOOD);
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return NULL; // Stop trying to resolve the message.
}


/* Go through all Commands and see that each one is up to date with the
original Model it was cloned from. */

void ScriptOCronApp::UpdateWithNewerModels ()
{
  CommandMap::iterator CommandIter;
  status_t ErrorCode = B_OK;

  for (CommandIter = m_CommandMap.begin (); CommandIter != m_CommandMap.end ();
  CommandIter++)
  {
    Command *CommandPntr = CommandIter->second;

    // Look for a Model with the same name as the "Original Model" of the
    // Command.  If none (user edited Commands), an empty string is used.

    BString ModelName (CommandPntr->UserDataGetString (g_KeyOriginalModel));

    ModelMap::iterator ModelIter = m_ModelMap.find (ModelName.String ());
    if (ModelIter != m_ModelMap.end ())
    {
      // Compare the version strings of both.  Empty strings if they don't
      // exist (happens for Commands saved by an older version of this
      // program).

      BString CommandVersion (CommandPntr->UserDataGetString (
        g_KeyOriginalModel, NULL, 1 /* second array element is version */));

      BString ModelVersion (g_EmptyString);
      int32 iUserData;
      for (iUserData = 0; true; iUserData++)
      {
        BMessage UserDataMessage;
        ErrorCode = ModelIter->second->FindMessage (
          g_UserDataName, iUserData, &UserDataMessage);
        if (ErrorCode != B_OK)
          break; // Probably hit the end of the array of UserData.

        const char *KeyPntr = NULL;
        if (UserDataMessage.FindString (g_NameKey, &KeyPntr) != B_OK)
          continue; // Always need a key.
        if (KeyPntr == NULL || strcmp (KeyPntr, g_KeyOriginalModel) != 0)
          continue; // Some other UserData, not about the Model, skip it.

        if (UserDataMessage.FindString (g_ResultName, 1 /* index */,
        &ModelVersion) != B_OK)
          ModelVersion = g_EmptyString;

        break; // Found the Original Model entry, should only be one.
      }

      if (strcmp (CommandVersion.String (), ModelVersion.String ()) != 0)
      {
        // Different versions between Command and Model, update the Command.

#if INCLUDE_GUI_CODE && 0 /* Hmmm, stops unattended startups, leave out. */
        BString Description ("Download \"");
        Description << CommandPntr->Name () << "\" uses Template \"";
        Description << ModelName << "\", which has changed.  Upgrading from "
          "version \"";
        Description << CommandVersion << "\" to \"" << ModelVersion << "\".";

        DisplayErrorMessage (Description.String (), 0, "Upgrading a Download");
#endif
        printf ("Upgrading Command \"%s\" using Model \"%s\".  "
          "Moved from version \"%s\" to \"%s\".\n",
          CommandPntr->Name (), ModelName.String (),
          CommandVersion.String (), ModelVersion.String ());

        CommandPntr->UpgradeFromNewerModel (ModelIter->second);

        // Store the new Model version number in the Command.

        CommandPntr->UserDataDelete (g_KeyOriginalModel);
        CommandPntr->UserDataAddString (g_KeyOriginalModel,
          ModelName.String ());
        CommandPntr->UserDataAddString (g_KeyOriginalModel,
          ModelVersion.String ());

        g_SettingsNeedSaving = true;
      }
    }
  }
}



/******************************************************************************
 * Finally, the main program which drives it all.
 */

int main (int argc, char **argv)
{
  // Save the file handles to standard input, output, error so we can restore
  // them whenever we change them so that a child process inherits different
  // file handles for its standard input, etc.

  int PipeFlags = 0;

  g_OriginalStdErr = dup (STDERR_FILENO);
  PipeFlags = fcntl (g_OriginalStdErr, F_GETFD);
  PipeFlags |= FD_CLOEXEC;
  fcntl (g_OriginalStdErr, F_SETFD, PipeFlags);

  g_OriginalStdIn = dup (STDIN_FILENO);
  PipeFlags = fcntl (g_OriginalStdIn, F_GETFD);
  PipeFlags |= FD_CLOEXEC;
  fcntl (g_OriginalStdIn, F_SETFD, PipeFlags);

  g_OriginalStdOut = dup (STDOUT_FILENO);
  PipeFlags = fcntl (g_OriginalStdOut, F_GETFD);
  PipeFlags |= FD_CLOEXEC;
  fcntl (g_OriginalStdOut, F_SETFD, PipeFlags);

  g_LockStdFilesPntr = new BLocker ("Std-In-Out-Err Changed Lock");

  cout << endl << PrintUsage;

#if TEST_CRON_BITSET
  TestCronBitset ();
#endif

#if TEST_URL_CONVERTER
  TestURLConverter ();
#endif

#if INCLUDE_GUI_CODE
  { // Magic command line argument of "meta" turns on meta mode.
    int iArg;
    for (iArg = 1; iArg < argc; iArg++)
    {
      if (strcasecmp (argv[iArg], "meta") == 0)
        g_MetaMode = true;
    }
    cerr << "Meta-advanced editing (use \"meta\" on the command line) is o" <<
      (g_MetaMode ? "n." : "ff.") << endl;
  }
#endif // INCLUDE_GUI_CODE

  {
    ScriptOCronApp MyApp;

    if (MyApp.InitCheck () == B_OK)
    {
      g_App = &MyApp;
      MyApp.LoadSaveSettings (true /* DoLoad */);
      MyApp.UpdateWithNewerModels ();
      MyApp.Run ();
    }
    cerr << PROGRAM_NAME_WORD " shutting down." << endl;
  } // Implicitly destroys MyApp at the end of this block.
  g_App = NULL;

  delete g_LockStdFilesPntr;

  close (g_OriginalStdErr);
  close (g_OriginalStdIn);
  close (g_OriginalStdOut);

  return 0;
}


/******************************************************************************
 * $Log: ScriptOCron.cpp,v $
 * Revision 2.91  2018/02/19 00:17:20  agmsmith
 * More help text.
 *
 * Revision 2.90  2018/02/17 21:32:56  agmsmith
 * Typo.
 *
 * Revision 2.89  2018/02/15 20:59:43  agmsmith
 * Don't upgrade user data set by the user; add rename GUI choices to
 * a list of UserData that doesn't get upgraded.
 *
 * Revision 2.88  2018/02/14 20:10:57  agmsmith
 * Added LocalPath display type and unified code back to one class to handle
 * LocalDir and LocalFile too.  Also add a slash after directory names, though
 * the user can edit that.  Also show parent directory window in Tracker for
 * files, and change the Tracker selection to highlight the file (rather than
 * having it open the file based on its file type).
 *
 * Revision 2.87  2018/01/25 15:40:27  agmsmith
 * Bump copyright year, changes still being made to code.
 *
 * Revision 2.86  2018/01/24 20:26:18  agmsmith
 * Added a menu item to switch to a simple window title (just the program name)
 * for stack & tile convenience.  Corresponding entry added to the settings
 * file.
 *
 * Revision 2.85  2018/01/22 16:35:08  agmsmith
 * For AsDatedURL also expand dates in the directory path, if it starts with a
 * plus sign.
 *
 * Revision 2.84  2018/01/02 21:51:00  agmsmith
 * Bump to version 2.84, for new (December 2017) Haiku.
 *
 * Revision 2.83  2017/11/14 15:39:53  agmsmith
 * No change, just bumped version number.
 *
 * Revision 2.82  2017/11/13 22:52:42  agmsmith
 * Change the LocalDir display and the Copy to Destination Directory processing
 * step to use paths relative to the Station directory, wherever it may be.
 * That includes converting absolute paths to relative, if they are in the
 * Station directory.  So "/boot/Station/Music" becomes "Music".
 *
 * Revision 2.81  2017/11/11 16:32:25  agmsmith
 * Bumped to version 2.81.  New updater code which updates everything
 * from a password protected repository, not just one package.
 *
 * Revision 2.80  2017/11/10 23:48:58  agmsmith
 * Bumped to version 2.80, Fetch Just New Files now defaults to
 * downloading all files on the first run.
 *
 * Revision 2.79  2017/11/08 21:45:13  agmsmith
 * No change, just a version bump.
 *
 * Revision 2.78  2017/11/05 00:07:45  agmsmith
 * Added the Run External Program Command and Processing Step, also the related
 * LocalFile display type, and rearranged the order of GUI labels for it and
 * LocalDir.
 *
 * Revision 2.77  2017/11/03 16:57:45  agmsmith
 * Reduce size of log text buffers down to 2MB in both command line and GUI
 * sides of the program.  So max 4MB per download, times 100 is 400MB, which
 * should fit in the 800MB of free RAM per program, accounting for string
 * fragmentation etc.
 *
 * Revision 2.76  2017/11/02 21:47:51  agmsmith
 * Rename the log file and work directory when renaming a Command.  Also delete
 * the work directory as well as the log file when the Command is deleted.
 *
 * Revision 2.75  2017/11/02 20:40:28  agmsmith
 * Add a user setting for changing the font size of the Command names and
 * another for the log text size.  Menu items cycle through the sizes.
 *
 * Revision 2.74  2017/11/01 23:41:43  agmsmith
 * Fix Fetch Just New Files to skip directories which the server lists as:
 * 12-09-15  05:47PM       <DIR>          Answers With Ken Ham Cue Sheets
 * otherwise it tries to download them as files.
 *
 * Revision 2.73  2017/10/31 20:58:06  agmsmith
 * Update comment about settings file being safely closed before renaming.
 *
 * Revision 2.72  2017/10/27 21:59:44  agmsmith
 * Use a temporary name when saving the settings file, and keep around a backup
 * of the file as it was before the program started running.
 *
 * Revision 2.71  2017/10/10 19:35:26  agmsmith
 * Bumped to version 2.71, which fixes a bug with & or \ in file names.
 *
 * Revision 2.70  2017/10/09 20:47:14  agmsmith
 * Version bump for Send-it GUI changes, no actual changes here.
 *
 * Revision 2.69  2017/09/07 21:10:30  agmsmith
 * Implement the AsNCFTP Field value modifying operation.
 *
 * Revision 2.68  2017/09/07 17:22:54  agmsmith
 * Turn off log to file by default.  Some systems were crashing with too much
 * logging activity (flakey Haiku OS problems).
 *
 * Revision 2.67  2017/08/02 00:18:44  agmsmith
 * Don't allow changes to the temporary directory while things are running.
 *
 * Revision 2.66  2017/08/01 23:53:36  agmsmith
 * Added a serial number to the user specified temp directory.
 *
 * Revision 2.65  2017/07/31 20:21:40  agmsmith
 * Add the user specified temporary directory setting.
 *
 * Revision 2.64  2017/07/30 20:52:12  agmsmith
 * Added PN_PATHS scripting operation to get the paths to the work directory,
 * temp directory etc, for a particular Command.  Also rearrange existing code
 * to use the internal function that gets those names, so they're all
 * calculated in one spot without duplication.
 *
 * Add a GUI (two buttons) for Fetch Just New Files to reset the list of files
 * known to be already downloaded, to force redownloading or skip all of them
 * currently on the server.  Use the new PN_PATHS to find the file to
 * manipulate.
 *
 * Revision 2.63  2017/07/23 20:10:07  agmsmith
 * Adding a new JustNewOps display type, also restrict URLDirOnly to be FTP.
 *
 * Revision 2.62  2017/07/15 14:16:51  agmsmith
 * Remove the time delay before opening the log file, crash still happens,
 * seemingly due to a corrupt log file on disk.
 *
 * Revision 2.61  2017/07/14 16:40:32  agmsmith
 * Add a half second delay before opening the log file, to maybe avoid a Haiku
 * page fault in their file caching code.
 *
 * Revision 2.60  2017/06/23 14:59:36  agmsmith
 * For Just New Files downloads, skip lines in the FTP listing that start with
 * "total" since they're aren't file names.  Happens in Haiku FTP server
 * listings.
 *
 * Revision 2.59  2017/06/22 22:50:26  agmsmith
 * No changes, just to bump the version number.
 *
 * Revision 2.58  2017/06/16 18:24:22  agmsmith
 * Added branch for Haiku2013 to log.
 *
 * Revision 2.57.1.1  2017/06/16 18:16:17  agmsmith
 * Special build for Haiku 2013 (no packages, no updater).
 *
 * Revision 2.57  2017/05/10 20:57:36  agmsmith
 * Validate Command names when renaming, don't want slashes or other characters
 * which may break things when the name is used as a file name or directory
 * name.
 *
 * Revision 2.56  2017/05/03 21:50:12  agmsmith
 * Also search Haiku's new DATA directory locations, as well as the ADDONS
 * directories.  Cédric pointed out that the files in ADDONS are normally just
 * executables.
 *
 * Revision 2.55  2017/04/25 13:48:01  agmsmith
 * Better upgrade debug text formatting.
 *
 * Revision 2.54  2017/04/18 22:56:19  agmsmith
 * Added documentation for MultipleChoiceWithText field display type.
 *
 * Revision 2.53  2017/04/07 21:34:31  agmsmith
 * Added CheckBox Field display type and related class.
 *
 * Revision 2.52  2017/04/05 19:01:20  agmsmith
 * Docs for VisbileModes changed to included Fields too.
 *
 * Revision 2.51  2017/03/28 22:49:48  agmsmith
 * Add SOCLogFile and explain how to e-mail the log file and summary text.
 *
 * Revision 2.50  2017/03/28 17:24:31  agmsmith
 * Add VisibleModes flag to hide subsidiary Commands in various GUI display
 * modes.
 *
 * Revision 2.49  2017/03/18 20:07:36  agmsmith
 * Avoid an infinite loop when doing substitutions with a Field name that is a
 * simple letter, which shows up in the substitution.
 *
 * Revision 2.48  2017/03/16 21:08:46  agmsmith
 * Bump copyright year.
 *
 * Revision 2.47  2017/03/08 15:30:05  agmsmith
 * Fix stop button, need to set process group explicitly, can't use zero like
 * we did in the fork technique, since parent process is the wrong one.
 *
 * Revision 2.46  2017/03/08 01:50:43  agmsmith
 * Moved around some string constants.  Added functions for setting UserData
 * from contents of a BMessage.  Fleshed out the upgrade from a Model mechanism
 * to also do UserData and Fields.
 *
 * Revision 2.45  2017/03/05 00:21:40  agmsmith
 * Updated GroupBox UserData documentation.
 *
 * Revision 2.44  2017/02/10 22:54:01  agmsmith
 * Made g_NameIndex an exported global variable.
 *
 * Revision 2.43  2017/02/02 01:11:06  agmsmith
 * Converted running of Commands from fork() to load_image(), making it faster
 * and less likely to crash on Haiku.  Also some mutual exclusion between
 * running Commands and running generic CLI from the window thread.
 *
 * Revision 2.42  2017/01/29 23:38:09  agmsmith
 * Make RunCLIString use load_image, so no forking etc, and faster.
 *
 * Revision 2.41  2017/01/29 01:40:59  agmsmith
 * Moved code for getting output from a piped command line to a separate
 * function.
 *
 * Revision 2.40  2017/01/17 16:08:06  agmsmith
 * Bumped version number.
 *
 * Revision 2.39  2016/12/09 23:02:13  agmsmith
 * Added AsDatedString postfix keyword to help with renaming.
 *
 * Revision 2.38  2016/12/05 18:09:35  agmsmith
 * Added DisplayType of Renamer.
 *
 * Revision 2.37  2016/11/02 19:08:05  agmsmith
 * Added HelpText for per-Field GUI status line text shown when mouse over
 * Field.
 *
 * Revision 2.36  2016/10/25 21:07:55  agmsmith
 * Added GroupBox UserData entry and key string.
 *
 * Revision 2.35  2016/10/18 21:09:54  agmsmith
 * Add URLDirOnly display option, hides the filename "what" input boxes.
 *
 * Revision 2.34  2016/10/06 22:48:59  agmsmith
 * Adding e-mail Model, and needed SOCName to show ancestor's name for
 * subsidiary Commands.
 *
 * Revision 2.33  2016/09/17 22:51:38  agmsmith
 * Save a version number (date stamp) in "Original Model" user data in
 * Commands, so we can tell if a changed (could be older or newer) Model is
 * available.  Also update Commands with new Models upon loading.  Add a
 * UserData delete utility function, and allow for arrays of strings when
 * getting a UserData string.
 *
 * Revision 2.32  2016/09/16 21:20:49  agmsmith
 * Change "Original Model" to be generated when the Model is created rather
 * than added to a new Command when it is cloned from the Model.  Also add a
 * version number (date stamp) to the original model data, so we can tell if it
 * needs an update or already matches the current Model when loading a Command.
 *
 * Revision 2.31  2016/09/14 21:57:41  agmsmith
 * Moved code into functions SafeEscape and ExpandDateFormat so GUI version can
 * use them too.
 *
 * Revision 2.30  2016/09/09 19:58:45  agmsmith
 * Add AsDatedURL option to incorporate a date in a file name.
 *
 * Revision 2.29  2016/09/09 15:50:05  agmsmith
 * Fix a bug with handling UserData where there are no "result" fields.
 *
 * Revision 2.28  2016/09/08 22:00:51  agmsmith
 * Added AsSafe postfix and redid substitution of magic words in Field values.
 *
 * Revision 2.27  2016/08/23 00:47:10  agmsmith
 * Multiple choice GUI now working, pop-up menu and custom GUI titles etc.
 *
 * Revision 2.26  2016/08/22 00:50:12  agmsmith
 * Add g_NameUnknown for "Unknown" default string, mostly for convenience.
 *
 * Revision 2.25  2016/07/27 20:28:18  agmsmith
 * Include key in returned UserData scripting BMessage so you can see which key
 * you got if accessing it via array indices (also lets you list the keys).
 * Clarify "Original Model" comments to have a space in the term.  Add the
 * g_KeyOriginalModel string to the exported variables.
 *
 * Revision 2.24  2016/04/30 21:21:06  agmsmith
 * Bad bugs and crashes happen when renaming subsidiary Commands, so don't
 * allow it.
 *
 * Revision 2.23  2016/04/26 12:28:08  agmsmith
 * Docs.
 *
 * Revision 2.22  2016/04/21 02:31:44  agmsmith
 * More readable version of the "Datatype of Output" key string.
 *
 * Revision 2.21  2016/04/20 20:44:05  agmsmith
 * Added more info to the Get Model scripting operation.  Now tracks a change
 * serial number for Models, lets you get it.  Also can now find out the data
 * types of the Model's input and output.
 *
 * Revision 2.20  2016/04/20 15:42:24  agmsmith
 * Better usage text for scripting operations, divide by category.
 *
 * Revision 2.19  2016/04/20 14:57:49  agmsmith
 * Pass changes up to the Ancestor Command too.  Make SOCWorkDir and other
 * directory names use the Ancestor Command's name for the directory.
 *
 * Revision 2.18  2016/04/19 21:34:54  agmsmith
 * Typo.
 *
 * Revision 2.17  2016/04/19 21:20:31  agmsmith
 * Added descriptions of data types for Command inputs and outputs.
 *
 * Revision 2.16  2016/04/10 21:29:36  agmsmith
 * Add ancestor and parent to the Children scripting operation.
 *
 * Revision 2.15  2016/04/08 18:24:54  agmsmith
 * Make maximum in-memory log length visible in the documentation.
 *
 * Revision 2.14  2016/03/14 18:04:51  agmsmith
 * Editing state of subsidiary Commands now controlled by ancestor.
 *
 * Revision 2.13  2016/03/05 22:57:52  agmsmith
 * Add a scripting command to get the list of chilren of a Command.
 *
 * Revision 2.12  2016/03/05 22:28:14  agmsmith
 * Hide subsidiary Commands except in Meta mode.
 *
 * Revision 2.11  2016/03/05 22:08:32  agmsmith
 * Fix saving and loading of tree structured Commands so Duplicate now works.
 *
 * Revision 2.10  2016/03/01 23:57:34  agmsmith
 * Dump command line text direct to log, can be too large for temp buffer.
 *
 * Revision 2.9  2016/03/01 22:44:26  agmsmith
 * Haiku has a different name for the temporary files directory.
 *
 * Revision 2.8  2016/02/29 02:09:19  agmsmith
 * No trailing slash on SOCTempDir and SOCWorkDir, update docs.
 *
 * Revision 2.7  2016/02/28 21:18:21  agmsmith
 * Add a few magic words for Work and Temp directories, sanitize names better.
 *
 * Revision 2.6  2016/02/24 19:57:01  agmsmith
 * Make it so disabled subsidiary Commands still run their children.
 *
 * Revision 2.5  2016/02/24 17:50:21  agmsmith
 * Collect error codes of children only after they've all finished, otherwise
 * may pick up an uninitialised error from the previous run.
 *
 * Revision 2.4  2016/02/24 15:50:40  agmsmith
 * Fix logging bugs, pass on child error codes to the ancestor Command when
 * running and print them to the log only after all the child Commands have
 * finished.
 *
 * Revision 2.3  2016/02/24 14:06:36  agmsmith
 * Remove RecursiveFlag infrastructure, don't need it now that we guarantee
 * non-circularity of subsidiary Command parent/child relationships.
 *
 * Revision 2.2  2016/02/24 00:17:34  agmsmith
 * Fix bug with detecting circular loops in ancestry, set trigger of Child
 * commands to the new name of their parent.
 *
 * Revision 2.1  2016/02/21 21:21:44  agmsmith
 * Fixing some recursive problems with loading.  Also might as well make this
 * the start of version 2 (child operations being the main feature), though
 * should have done that earlier.
 *
 * Revision 1.143  2016/02/20 23:50:38  agmsmith
 * Work in progress, still have to remove RecursiveFlags.
 *
 * Revision 1.142  2016/02/16 19:02:18  agmsmith
 * Work in progress on InternalAddCommand and InternalRemoveCommand, adding
 * ancestory and subsidiary child handling.
 *
 * Revision 1.141  2016/02/13 15:52:10  agmsmith
 * Adding Ancestry, moved around App functions to be in alphabetical order.
 *
 * Revision 1.140  2016/02/13 14:53:26  agmsmith
 * Some renaming and starting to add ancestry variables.
 *
 * Revision 1.139  2016/02/11 23:00:25  agmsmith
 * Added m_Subsidiary and scripting operations, alphabetical PN_ enum.
 *
 * Revision 1.138  2016/02/11 19:18:54  agmsmith
 * Work in progress, adding wait for children state, only enum done so far.
 *
 * Revision 1.137  2016/02/10 03:33:14  agmsmith
 * Moved disable functionality to a separate flag with its own scripting
 * commands.  Old @Disabled in the trigger string now removed when loading
 * settings.
 *
 * Revision 1.136  2016/02/02 21:30:56  agmsmith
 * Don't return an error code in reply BMessages if no error happened, so
 * command line users aren't startled by seeing an error code when they use
 * "hey".
 *
 * Revision 1.135  2016/02/02 21:02:21  agmsmith
 * Always create the settings directory, otherwise first time startup detection
 * will fail.
 *
 * Revision 1.134  2016/01/23 20:59:16  agmsmith
 * Work in progress, adding multiple model file reading.
 *
 * Revision 1.133  2015/12/30 22:23:12  agmsmith
 * Some scripting operations usage wording improved.
 *
 * Revision 1.132  2015/12/26 17:14:09  agmsmith
 * Add the close on exec O_CLOEXEC flag to log files, don't want to leave
 * zillions of file handles open in child processes.
 *
 * Revision 1.131  2015/12/21 00:22:53  agmsmith
 * Move log files of deleted Commands to the system trash directory.
 *
 * Revision 1.130  2015/12/20 17:30:50  agmsmith
 * Bump version number.
 *
 * Revision 1.129  2015/12/19 22:22:30  agmsmith
 * Bump version number.
 *
 * Revision 1.128  2015/12/19 04:41:06  agmsmith
 * Change @Reboot to be @Launch, and the button label to be Run at Startup.
 *
 * Revision 1.127  2015/12/17 20:02:52  agmsmith
 * Added magic word ThisCommandName, which is what it says.
 *
 * Revision 1.126  2015/12/17 18:48:22  agmsmith
 * Use the canonical name of the Model in the Original Model saved data.
 *
 * Revision 1.125  2015/12/16 16:18:29  agmsmith
 * Downloads renamed, now "FTP Download" is the default for new ones in Basic
 * mode.  Now saves and loads the initial window size, so your last window size
 * and position is remembered.  Goes back to default size if you make the
 * window too small or move it off screen.
 *
 * Revision 1.124  2015/12/15 22:46:55  agmsmith
 * Bumped version.
 *
 * Revision 1.123  2015/12/13 20:13:56  agmsmith
 * Added Duplicate command and menu items, reworked new name generation too.
 *
 * Revision 1.122  2015/12/12 20:24:01  agmsmith
 * Added Meta mode and command line argument to trigger it.
 *
 * Revision 1.121  2015/12/11 19:09:53  agmsmith
 * Copyright notice shortened.
 *
 * Revision 1.120  2015/12/11 18:56:15  agmsmith
 * About box usage text vastly reduced for GUI version.
 *
 * Revision 1.119  2015/12/11 01:59:17  agmsmith
 * Bump version number to reflect changes in other files.
 *
 * Revision 1.118  2015/12/09 22:46:05  agmsmith
 * Export the CopyBMessageField function.
 *
 * Revision 1.117  2015/12/09 16:26:09  agmsmith
 * Remove welcome error message on first run (no settings file error).
 *
 * Revision 1.116  2015/12/09 03:34:45  agmsmith
 * Wording different for first time Fetchit vs AGMSScriptOCron.
 *
 * Revision 1.115  2015/12/08 21:44:01  agmsmith
 * By default, enable the log file on new Commands.  Make start in Windowed
 * mode the default, and rephrase the welcome text error message.
 *
 * Revision 1.114  2015/12/08 21:09:22  agmsmith
 * Back to launch with window showing as the initial default.
 *
 * Revision 1.113  2015/12/08 11:42:28  agmsmith
 * Don't confirm quit with user if it's a system shutdown kind of quit.
 *
 * Revision 1.112  2015/12/08 00:16:39  agmsmith
 * Resume running if the quit was cancelled.
 *
 * Revision 1.111  2015/12/07 14:58:38  agmsmith
 * Added some more UserData key descriptions.  Functions to access string type
 * UserData conveniently.  Add the OriginalModel UserData to new Commands
 * created from a Model, and remove it if the Template is edited by the user.
 *
 * Revision 1.110  2015/12/06 00:12:40  agmsmith
 * Fix QuitRequested behaviour so it delays replying to the B_QUIT_REQUESTED
 * message until a few updates have gone by (time to close threads and fetch
 * last log file entries).
 *
 * Revision 1.109  2015/12/04 22:33:15  agmsmith
 * Default to server mode on for new installs, as Dane requested.
 *
 * Revision 1.108  2015/12/04 22:25:10  agmsmith
 * Check for changes that affect settings, and save once a minute.
 *
 * Revision 1.107  2015/12/04 21:18:01  agmsmith
 * Added the ability to detect the first use of the program.
 *
 * Revision 1.106  2015/12/03 22:23:51  agmsmith
 * Log file GUI now working, enable/disable plus view log file button.
 *
 * Revision 1.105  2015/12/03 20:50:35  agmsmith
 * Also return the log path name when asked for log info.
 *
 * Revision 1.104  2015/12/03 19:14:34  agmsmith
 * Logging to a file feature added, no GUI yet.
 *
 * Revision 1.103  2015/12/02 00:03:05  agmsmith
 * Tune log receiving buffer and speed, Haiku uses bigger blocks too.
 *
 * Revision 1.102  2015/12/01 00:38:48  agmsmith
 * Move the proprietary GUI code into a separate file, change program name for
 * GUI mode to Fetchit!TM.
 *
 * Revision 1.101  2015/11/29 17:37:07  agmsmith
 * File requestor and local directory text input now working.
 *
 * Revision 1.100  2015/11/29 14:26:55  agmsmith
 * Work in progress adding FieldViewLocalDir, also rearranged base class
 * FieldViewGeneric and fixed some enabling bugs in FieldViewURL and some width
 * spacing everywhere.
 *
 * Revision 1.99  2015/11/28 17:27:13  agmsmith
 * Wording changes from Dane.
 *
 * Revision 1.98  2015/11/28 03:10:53  agmsmith
 * Trigger time classes reorganised under a new base class.  Now auto-edit-mode
 * works for the full trigger time editor GUI.  Added Disabled trigger time
 * checkbox.  Template text now has auto-edit.
 *
 * Revision 1.97  2015/11/27 02:35:41  agmsmith
 * Added basic mode time display and editing.  Still need to backport the
 * disabled feature and auto-edit to the full time edit view.
 *
 * Revision 1.96  2015/11/26 01:11:21  agmsmith
 * Process user input from the URL field GUI.  Now fully working.
 *
 * Revision 1.95  2015/11/25 22:49:58  agmsmith
 * URL parsing now displayed, but the reverse direction (editing URL) isn't
 * implemented yet.  Auto-save in basic mode when Enter is typed.  Change short
 * instruction text to reflect current situation.
 *
 * Revision 1.94  2015/11/25 00:36:04  agmsmith
 * Working on the URL display, added URLConverter class and test code.
 *
 * Revision 1.93  2015/11/23 23:04:12  agmsmith
 * Work in progress - moved URL form into a FieldViewURL class, no longer built
 * into the Basic view (so can be used in Advanced too).
 *
 * Revision 1.92  2015/11/23 21:53:19  agmsmith
 * Renaming FieldView classes, getting ready to work on FieldViewURL.
 *
 * Revision 1.91  2015/11/23 20:49:40  agmsmith
 * Make a common parent BView class for the advanced and basic Command details
 * views, so they have a common virtual polling function and can then be used
 * interchangably.  Turns out we didn't need it, can force an extra poll by
 * using the SelectionChanged flag when switching between advanced and basic.
 *
 * Revision 1.90  2015/11/22 22:42:13  agmsmith
 * Split FieldGenericView into a base abstract class and one for text Fields.
 *
 * Revision 1.89  2015/11/22 02:40:05  agmsmith
 * Added the UserData facility, now we can store stuff in Commands and Fields
 * without having to add more scripting code.
 *
 * Revision 1.88  2015/11/20 20:31:47  agmsmith
 * Fix bug with detecting if a name is used by a Model.
 *
 * Revision 1.87  2015/11/20 20:03:22  agmsmith
 * Added "What" elements for picking a file name.  Also don't start Commands
 * when shutting down.
 *
 * Revision 1.86  2015/11/20 18:27:41  agmsmith
 * Removed g_ServerMode from non-GUI build and from scripting system (though
 * still saved in settings), renamed non-server as Windowed mode.  Added
 * g_AdvancedMode and a menu to switch between basic and advanced mode Command
 * details user interfaces.  Also New Command button now only creates file
 * download Commands in basic mode (no pop-up menu).
 *
 * Revision 1.85  2015/11/19 21:51:00  agmsmith
 * Got an initial Basic command details view working, just with name and a
 * radio button.
 *
 * Revision 1.84  2015/11/18 22:57:06  agmsmith
 * Now working with advanced stuff in a separate sub-view.
 *
 * Revision 1.83  2015/11/17 22:51:45  agmsmith
 * Splitting out code into AdvancedCommandDetailsView so that we can have both
 * basic and advanced modes.  Work in progress - have to hook up messaging
 * again.
 *
 * Revision 1.82  2015/11/17 19:02:25  agmsmith
 * x-vnd rather than vnd, for app signature.
 *
 * Revision 1.81  2015/11/17 19:00:48  agmsmith
 * Move some GUI variables to their own section, make signature
 * usage more consistently lower case for scripting sigs etc.
 *
 * Revision 1.80  2015/11/16 22:42:22  agmsmith
 * Added the company logo badge graphic, which pulls up the docs too.
 *
 * Revision 1.79  2015/11/12 19:57:22  agmsmith
 * Added a menu bar and moved the About and Server Mode gadgets there.
 *
 * Revision 1.78  2015/11/11 23:07:47  agmsmith
 * Fix a bug in loading models, name pointer changes when manipulating a
 * BMessage.
 *
 * Revision 1.77  2015/11/05 23:02:35  agmsmith
 * Now loads built-in Models from a resource, and only saves user created
 * Models.
 *
 * Revision 1.76  2015/11/05 21:01:49  agmsmith
 * Use Model name when making new Commands in the GUI, also add a sequence
 * number at the end so you can add a lot of Commands in a batch.
 *
 * Revision 1.75  2015/11/04 22:58:50  agmsmith
 * Add a text box so you can edit the Template of a Command.
 *
 * Revision 1.74  2015/11/04 21:11:10  agmsmith
 * Create new Commands in Edit state, so they don't start running immediately.
 * Also only allow delete of a Command in edit state.  Punctuation of error
 * messages removed from the end (the display adds a comma and the error
 * codes).
 *
 * Revision 1.73  2015/11/03 22:38:48  agmsmith
 * Show a pop-up menu listing available Models when New button used.
 *
 * Revision 1.72  2015/11/03 20:54:51  agmsmith
 * Added buttons for creating and deleting Models.
 *
 * Revision 1.71  2015/11/01 23:03:48  agmsmith
 * Finished Model functionality, can now create, list, delete and use Models.
 *
 * Revision 1.70  2015/10/31 21:42:24  agmsmith
 * Now loading and saving the Models.
 *
 * Revision 1.69  2015/10/31 19:11:46  agmsmith
 * Fix up a few small things, start on adding Models.
 *
 * Revision 1.68  2015/10/30 21:34:26  agmsmith
 * Move the CommandState into a global space enum, also fix space for text
 * labels to be a bit bigger so it shows up unclipped in Haiku.
 *
 * Revision 1.67  2015/10/29 17:01:16  agmsmith
 * Added the radio buttons for setting Trigger type.
 *
 * Revision 1.66  2015/10/28 19:14:58  agmsmith
 * Added the trigger string text box and connected up the checkboxes so that
 * they change the trigger time too.
 *
 * Revision 1.65  2015/10/27 21:56:47  agmsmith
 * Now updates the time checkboxes with data from the Command's trigger string.
 *
 * Revision 1.64  2015/10/26 17:00:35  agmsmith
 * Added a scroll bar to the Start Times tab, also moved the scrolling stuff to
 * an internal class so it can be reused for subclasses.
 *
 * Revision 1.63  2015/10/25 22:49:48  agmsmith
 * Add the All or None checkbox after the other cron time ones.
 *
 * Revision 1.62  2015/10/25 21:42:19  agmsmith
 * Reverse the order of crontab fields to be more intuitive, with minutes last.
 *
 * Revision 1.61  2015/10/24 21:09:14  agmsmith
 * Working on TriggerTimeView to let you edit the cron string, currently shows
 * the checkboxes but they aren't hooked up to data yet.
 *
 * Revision 1.60  2015/10/21 23:23:29  agmsmith
 * Fixed tabbed view resizing so that when switching back to a tab, it gets
 * resized to the current size of the parent view, compensating for missed
 * resize messages.  Also move creation of child GUI elements out of
 * AttachedToWindow so you don't get duplicates every time a tab is selected
 * (tabs work by detaching the unused tabs, and then attaching the visible tab
 * to the main view).
 *
 * Revision 1.59.1.1  2015/10/21 19:53:36  agmsmith
 * Tried fixing missing size updates by using a custom BScrollView that resizes
 * whenever it is reattached to window.  Unfortunately, parent view is an
 * internal one in the BTab that is the wrong size.
 *
 * Revision 1.59  2015/10/21 18:51:16  agmsmith
 * Move Command Line display to the miscellaneous tab.
 *
 * Revision 1.58  2015/10/19 19:57:31  agmsmith
 * Moved various GUI elements into a BTabView, with tabs for Fields, Log and
 * Miscellaneous.  Mostly working, but with some redraw problems in Fields.
 *
 * Revision 1.57  2015/10/15 18:52:13  agmsmith
 * List of Field data entry fields is now scrollable, with a scroll bar that
 * responds to the size of the window, etc.  Also have the log and Field
 * display splitting the available space in half, dynamically updated as the
 * window is resized.
 *
 * Revision 1.56  2015/10/05 21:44:31  agmsmith
 * Started moving the display of Fields into its own BView so that we can later
 * do scrolling of the whole thing.
 *
 * Revision 1.55  2015/10/04 20:13:10  agmsmith
 * Now with working Cron type command triggers, and @ReBoot works too.
 *
 * Revision 1.54  2015/10/02 03:03:22  agmsmith
 * Work in progress on trigger functionality.  Works for named Command, still
 * have to do Cron and reboot.
 *
 * Revision 1.53  2015/10/01 23:25:45  agmsmith
 * Added Trigger string to the Command scripting, doesn't yet do anything.
 *
 * Revision 1.52  2015/09/28 22:05:55  agmsmith
 * Work in progress on CronBitset.
 *
 * Revision 1.51  2015/08/24 23:16:22  agmsmith
 * Now keeps log text separately for each Command in the GUI, so selecting
 * commands will change the log text to the accumulation for that Command.
 *
 * Revision 1.50  2015/08/24 21:09:36  agmsmith
 * Moved change log to the end of the file, was getting too long.
 *
 * Revision 1.49  2015/08/21 22:35:06  agmsmith
 * Added a text box to display the log output.  Not quite right, need to
 * preserve logs for each Command separately, and show the appropriate one when
 * selecting different Commands.
 *
 * Revision 1.48  2015/08/21 18:05:31  agmsmith
 * Added polling to fetch standard output and store it in a log string, added a
 * scripting operation to read the log string, figured out how to kill the
 * child shell and its children using process groups.
 *
 * Revision 1.47  2015/08/20 19:22:32  agmsmith
 * Can now run a command, using fork() and execve().  Can't see the output yet.
 *
 * Revision 1.46  2015/08/13 18:25:39  agmsmith
 * Now does keyword substitution to generate the command line.
 *
 * Revision 1.45  2015/08/11 21:54:46  agmsmith
 * Added scripting to get the current command line, GUI to display it, just
 * need some code to calculate it.
 *
 * Revision 1.44  2015/08/10 21:47:51  agmsmith
 * Added the Template string to Commands, including scripting and archiving.
 *
 * Revision 1.43  2015/08/10 16:06:56  agmsmith
 * Too many buttons, make two rows.  Also Save button now forces text updates
 * for text boxes so that the last bit of editing by the user gets transfered
 * too.
 *
 * Revision 1.42  2015/08/10 15:28:59  agmsmith
 * Added a Stop button, scripting for Run command, enable and disable buttons
 * depending on state.
 *
 * Revision 1.41  2015/08/10 00:08:21  agmsmith
 * Also select the new Command when using the New button.
 *
 * Revision 1.40  2015/08/09 23:48:12  agmsmith
 * Switch to using messages to change the selected Command after a rename,
 * finally figured out how to send a message to self with a BInvoker
 * (PostMessage doesn't work for some reason).
 *
 * Revision 1.39  2015/08/09 21:53:38  agmsmith
 * Added Delete button, reselection after name changed (using global vars).
 *
 * Revision 1.38  2015/07/28 21:50:47  agmsmith
 * Can now edit the name of a Command in the GUI.
 *
 * Revision 1.37  2015/07/28 19:25:12  agmsmith
 * Field BViews now are inside the Command Details box.
 *
 * Revision 1.36  2015/07/28 15:27:34  agmsmith
 * Trying to stuff the scrolling list view inside a BBox, but the coordinates
 * of the box's interior space aren't quite right.
 *
 * Revision 1.35  2015/07/24 19:50:32  agmsmith
 * Can now see and edit the Field values.
 *
 * Revision 1.34  2015/07/24 18:56:06  agmsmith
 * Consistently format class parts, using // comments after public/private.
 *
 * Revision 1.33  2015/07/22 21:27:04  agmsmith
 * Added buttons for Edit, Save, Cancel (all now working) and an unimplemented
 * Run button.
 *
 * Revision 1.32  2015/07/15 22:26:58  agmsmith
 * Can now set and get Field's Value, also stored in settings.
 *
 * Revision 1.31  2015/07/15 15:07:22  agmsmith
 * Switch to BTextControl for displaying field values.
 *
 * Revision 1.30  2015/07/08 01:41:23  agmsmith
 * Field value as a BTextView.  Not too good, should try BTextControl instead.
 *
 * Revision 1.29  2015/07/05 00:24:50  agmsmith
 * Made fonts global, rather than having a copy for each displayed thing.
 * Display the Field names (temporary code using DrawString) and update them as
 * the fields change.
 *
 * Revision 1.28  2015/06/15 21:59:01  agmsmith
 * Adding FieldGenericView, also put implementation of classes in alphabetical
 * order, so Command stuff comes before Field stuff.
 *
 * Revision 1.27  2015/06/12 19:39:16  agmsmith
 * Can now create, delete, rename, count Field objects.
 *
 * Revision 1.26  2015/06/11 17:19:09  agmsmith
 * In the middle of adding Field objects.
 *
 * Revision 1.25  2015/06/10 01:08:54  agmsmith
 * Finished the Edit property implementation, changed the list
 * items to update their state if the Command changed.
 *
 * Revision 1.24  2015/06/09 20:34:08  agmsmith
 * Working on Edit property of Commands.
 *
 * Revision 1.23  2015/05/26 02:04:34  agmsmith
 * List items now have their own large font, and the progress bar
 * is now being drawn.
 *
 * Revision 1.22  2015/05/25 22:00:03  agmsmith
 * More colourful list items, with proper antialiasing of text too.
 *
 * Revision 1.21  2015/05/22 21:57:08  agmsmith
 * Switch to using a custom ListItem to draw our Commands.
 *
 * Revision 1.20  2015/05/22 16:19:12  agmsmith
 * Force use of Specifier[0] when examining scripting messages, and other
 * changes to get Scripting also working in Haiku.
 *
 * Revision 1.19  2015/05/21 21:08:30  agmsmith
 * Now saves and loads Commands, also use modification detection system to
 * trigger saving of the settings file.
 *
 * Revision 1.18  2015/05/21 19:23:37  agmsmith
 * Use the change detection mechanism to avoid some updates.
 *
 * Revision 1.17  2015/05/21 18:39:37  agmsmith
 * Window list of commands now tracks the server's list of commands.
 *
 * Revision 1.16  2015/05/18 20:49:12  agmsmith
 * Work in progress in doing scripting and displaying commands.
 *
 * Revision 1.15  2015/05/15 21:57:26  agmsmith
 * Now with a scrolling list box that will later contain task names.
 *
 * Revision 1.14  2015/05/14 03:01:52  agmsmith
 * Added change serial numbers, and the ability to delete a Command.
 *
 * Revision 1.13  2015/05/13 00:18:34  agmsmith
 * Can now change the name of a Command, and read it, and the scripting now
 * gets down to the Command object using the specifier system.
 *
 * Revision 1.12  2015/05/12 22:52:54  agmsmith
 * Shuffling order of classes and implementations, more alphabetical now.
 *
 * Revision 1.11  2015/05/12 21:59:44  agmsmith
 * Working on adding a Command operation, also removed useless busy cursor and
 * added some locking utility methods.
 *
 * Revision 1.10  2015/05/12 15:50:17  agmsmith
 * Add scripting for an edit mode for Commands, with the ability to cancel an
 * edit and restore the Command to the original state.  Also usage wording
 * updated.
 *
 * Revision 1.9  2015/05/12 02:15:10  agmsmith
 * Better error messages for resolving specifiers.
 *
 * Revision 1.8  2015/05/11 21:45:45  agmsmith
 * Can now create Commands using scripting.
 *
 * Revision 1.7  2015/05/08 21:52:48  agmsmith
 * Resolving scripting specifiers still in progress, but closer now!
 *
 * Revision 1.6  2015/05/07 21:58:48  agmsmith
 * Work in progress on scripting.
 *
 * Revision 1.5  2015/05/07 18:48:03  agmsmith
 * Remove CommanderLooper since we no longer take command line arguments, also
 * update the scripting documentation and usage.
 *
 * Revision 1.4  2015/05/07 15:53:48  agmsmith
 * Mostly cosmetics - program name as a macro.
 *
 * Revision 1.3  2015/05/01 21:58:43  agmsmith
 * Use the same About text in the printed usage and in the about box, which
 * means being able to turn off word wrap, and capturing the usage output to an
 * in-memory string.
 *
 * Revision 1.2  2015/05/01 21:04:39  agmsmith
 * Got ServerMode set, get, save, load, display working.
 *
 * Revision 1.1  2015/05/01 19:46:46  agmsmith
 * Initial revision.  Main program for Script-o-Cron, initially a stripped down
 * copy of AGMSBaysianSpamServer.
 */
